/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include "MathFunc.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"

/*============================================================================*/
/* FUNCTION: Log_2 ()                                                         */
/*----------------------------------------------------------------------------*/
/* PURPOSE:  Approximation of log in base 2                                   */
/*----------------------------------------------------------------------------*/
/* INPUT ARGUMENTS:                                                           */
/*          X_INT32 x :  argument for logarithm                               */
/*----------------------------------------------------------------------------*/
/* OUTPUT ARGUMENTS:                                                          */
/*----------------------------------------------------------------------------*/
/* INPUT/OUTPUT ARGUMENTS:                                                    */
/*----------------------------------------------------------------------------*/
/* RETURN ARGUMENTS:  X_INT16                                                 */
/*           - approximation of  1024 * ln(|x|) / ln(2)                       */
/*           - 0 for x = 0                                                    */
/*           - 31 x 1024 for x = 0x80000000L                                  */
/*============================================================================*/
#define A0  1
#define A1  2943
#define A2  (-1374)
#define A3  640
#define A4  (-161)
#define A0_xx16 (65536L*A0) 


X_INT16 Log_2(
              X_INT32 x     /* (i) argument for logarithm */
    ) 
{
    X_INT16 exp ;
    X_INT16 mant ;
    X_INT32 log2 ;

    x = L_abs(x);
    test();
    if( x == 0 ) { 
        return (X_INT16)  0 ; 
    } 
    /* exponent */
    test();
    if(L_sub(x,0x8000L) < 0) {
        X_INT16 xs = extract_l(x) ;
        exp = norm_s(xs);
        mant = shl(xs,exp);
        exp=sub(14,exp);
    } else {
        X_INT16 xh = extract_h(x);
        test();
        if(xh == 0) {
            exp = 15 ; move16();
        } else {
            exp = norm_s(xh);
        }
        mant = extract_h(L_shl(x,exp));
        exp = sub(30,exp);
    }
    mant = shl((X_INT16) ( mant & (X_INT16) 0x3fff) , (X_INT16) 1 ) ; logic16();
    {   /* polynomial approximation */
        X_INT16 m2 = mult(mant,mant) ;
        X_INT16 m3 = mult(mant,m2);
        X_INT16 m4 = mult(m2,m2) ;
        log2 = L_mac(L_mac(L_mac(L_mac(A0_xx16,A1,mant),A2,m2),A3,m3),A4,m4);
        log2 = add(extract_h(L_shr(log2,1)),shl(exp,10));

    }

    return (X_INT16) log2 ;
}


static const X_INT16 tbInt0[48] = {
    12415, 12655, 12882, 13096, 13298, 13489, 13670, 13841, 14003, 14156,
    14302, 14440, 14571, 14695, 14813, 14925, 15031, 15132, 15227, 15317,
    15403, 15483, 15560, 15631, 15699, 15763, 15823, 15879, 15932, 15982,
    16027, 16070, 16110, 16146, 16180, 16211, 16239, 16264, 16287, 16307,
    16325, 16340, 16353, 16364, 16372, 16378, 16382, 16384
} ;
static const X_INT16 *tbInt = tbInt0 - 16 ;



/*============================================================================
 FUNCTION: Sqrt_2 ()                                                        
----------------------------------------------------------------------------
 PURPOSE:  compute x= round(sqrt(X)) i.e.  x  as abs(X - x*x) is minimum                                          
----------------------------------------------------------------------------
 INPUT ARGUMENTS:                                                           
          X_INT32 X :  argument for sqrt 
----------------------------------------------------------------------------
 OUTPUT ARGUMENTS:                                                          
----------------------------------------------------------------------------
 INPUT/OUTPUT ARGUMENTS:                                                    
----------------------------------------------------------------------------
 RETURN ARGUMENTS:   X_INT32                                                
           - x  as abs(X - x*x) is minimum   
           So :  if X < 0 x = 0 
                 for X=6 x = 2 
                 for X=7 x = 3 
                 if X >= 0x40000000 x >= 0x8000 
============================================================================*/

X_INT32 Sqrt_2(X_INT32 X) {
    X_INT32 D ;
    X_INT16 x ;
    X_INT16 h = round(X) ;
    test();
    if(sub(h,0x4000) < 0) {
        return L_deposit_l(Sqrt16_2(X));
    } else {
        h = shr(h,1);
        x = add(tbInt[shr(h,9)],shr(h,1)) ;
        x =  round(L_mac(L_mult(x,0x4000),0x4000,div_s(h,x))) ;
		X = X & 0xFFFFFFFC ; logic32(); /* only for backward compatibility */
        D = L_msu(L_msu(X,x,x),x,x);
        test();
        if(L_msu(D,x,1) > 0) {
            test(); sub(0,2);
            if(L_msu(D,x,3) > 2) {
                return L_mac(2,x,1) ;
            } else {
                return L_mac(1,x,1) ;
            }
        } else {
            test();
            if(L_mac(D,x,1) <= 0) {
				test(); sub(0,3);
                if(L_mac(D,x,3)< 3) {
                    return L_mac(-2,x,1) ;
                } else {
                    return L_mac(-1,x,1) ;
                }
            } else {
                return L_mult(1,x);
            }
        }
    }
}


/*============================================================================
 FUNCTION: Sqrt16_2 ()                                                        
----------------------------------------------------------------------------
 PURPOSE:  compute x= round(sqrt(X)) i.e.  x  as abs(X - x*x) is minimum 
           for X < 0x40000000                                         
----------------------------------------------------------------------------
 INPUT ARGUMENTS:                                                           
          X_INT32 X :  argument for sqrt16_2 
          Must be < 0x40000000
----------------------------------------------------------------------------
 OUTPUT ARGUMENTS:                                                          
----------------------------------------------------------------------------
 INPUT/OUTPUT ARGUMENTS:                                                    
----------------------------------------------------------------------------
 RETURN ARGUMENTS:  X_INT16                                                 
           - x  as abs(X - x*x) is minimum for X < 0x40000000
           So :  if X < 0 x = 0 
                 for X=6 x = 2 
                 for X=7 x = 3 
                 if X >= 03fff0001 x = 0x7fff (saturation)
============================================================================*/

X_INT16 Sqrt16_2(X_INT32 X) {
    X_INT32 D ;
    X_INT16  x ;
    X_INT16 h;

    h = extract_h(X) ;
    test();
    if(h<=0) {
        test();
        if(X<0) return 0 ;
        h = extract_h(L_shl(X,14)) ;
        test();
        if(sub(h,0x400)>= 0) {
            test();
            if(sub(h,0x1000) >=0) {
                h = round(L_shl(X,15)) ;
                x = add(tbInt[shr(h,9)],shr(h,1)) ;
                x =  round(L_mac(L_mult(x,0x80),0x80,div_s(h,x))) ;
            } else {
                h =shl(extract_l(X),1);
                x = add(tbInt[shr(h,9)],shr(h,1)) ;
                x =  round(L_mac(L_mult(x,0x40),0x40,div_s(h,x))) ;
            }
            D = L_msu(L_shl(X,1),x,x) ;
            test();
            if(L_mac(D,x,1) <= 0) {
                x = sub(x,1) ;
            }
        } else {
            test();
            if(sub(h,0x40) >= 0) {
                test();
                if(sub(h,0x100) >= 0) {
                    h =shl(extract_l(X),3);
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x20),0x20,div_s(h,x))) ;
                } else {
                    h =shl(extract_l(X),5);
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x10),0x10,div_s(h,x))) ;
                }
                D = L_msu(L_shl(X,1),x,x) ;
                test();
                if(L_mac(D,x,1) <= 0) {
                    x = sub(x,1) ;
                }
            } else {
                h = extract_l(X) ;
                x = add(2,shr_r(h,4));
                D = L_mult(1,h);
                test();
                if(L_msu(D,x,x) <= 0) {
                    X_INT16 x1 = sub(x,1);
                    test();
                    if(L_msu(D,x,x1) <= 0) {
                        x = sub(x1,1) ;
                        test();
                        if(L_msu(D,x,x1) > 0) {
                            x = x1 ; move16();
                        }
                    }
                } else {
                    X_INT16 x1 = add(x,1);
                    test();
                    if(L_msu(D,x,x1) > 0) {
                        x = add(x1,1);
                        test();
                        if(L_msu(D,x,x1) <= 0) {
                            x = x1 ; move16();
                        }
                    }
                }
            }
        }
    } else {
        test();
        if(sub(h,0x100) >= 0) {
            test();
            if(sub(h,0x1000) >= 0) {
                h= round(L_shl(X,1));
                x = add(tbInt[shr(h,9)],shr(h,1)) ;
                x =  round(L_mac(L_mult(x,0x4000),0x4000,div_s(h,x))) ;
            } else {
                test();
                if(sub(h,0x400) >= 0) {
                    h= round(L_shl(X,3));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x2000),0x2000,div_s(h,x))) ;
                } else {
                    h= round(L_shl(X,5));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x1000),0x1000,div_s(h,x))) ;
                }
            }
        } else {
            test();
            if(sub(h,0x10) >= 0) {
                test();
                if(sub(h,0x40) >= 0) {
                    h= round(L_shl(X,7));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x800),0x800,div_s(h,x))) ;
                } else {
                    h= round(L_shl(X,9));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x400),0x400,div_s(h,x))) ;
                }
            } else {
                test();
                if(sub(h,4) >= 0) {
                    h= round(L_shl(X,11));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x200),0x200,div_s(h,x))) ;
                } else {
                    h= round(L_shl(X,13));
                    x = add(tbInt[shr(h,9)],shr(h,1)) ;
                    x =  round(L_mac(L_mult(x,0x100),0x100,div_s(h,x))) ;
                }
            }
        }
        D = L_msu(L_shl(X,1),x,x) ;
        test();
        if(L_msu(D,x,1) > 0) {
            x = add(x,1);
        } else {
            test();
            if(L_mac(D,x,1) <= 0) {
                x = sub(x,1) ;
            }
        }
    }

    return x; 
}


        
