/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "rvc_pitch_api_B.h"
#include "rvc_pitch_B.h"
#include "computePCorr_B.h"
#include "ParmType.h"

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#include "qsort_B.h"
#include "fix_mathlib.h"


// ===============================================================
//
//                 Private Functions
//
// ===============================================================

/*
***********************************************************************
*
*   FUNCTION NAME: Mpy_lw_lw
*
*   PURPOSE: To multiply two longwords (32-bit words)
*
*   INPUTS:
*
*     lwIn1 - Input # 1 (longword)
*
*     lwIn2 - Input # 2 (longword)
*
*   OUTPUTS: None
*
*   RETURN VALUE: The product is returned as a longword
*
**********************************************************************
*/

Longword Mpy_lw_lw(Longword lwIn1, Longword lwIn2)
  {
  Shortword swIn1H, swIn1L;
  Shortword swIn2H, swIn2L;
  Longword lwTmp;

  swIn1H = extract_h(lwIn1);
  swIn1L = extract_l(lwIn1);
  swIn1L = shr(swIn1L,1);
  swIn1L = swIn1L & 0x7FFF; logic16();

  swIn2H = extract_h(lwIn2);
  swIn2L = extract_l(lwIn2);
  swIn2L = shr(swIn2L,1);
  swIn2L = swIn2L & 0x7FFF; logic16();

  lwTmp = L_mult(swIn1H, swIn2L);
  lwTmp = L_mac(lwTmp, swIn1L, swIn2H);
  lwTmp = L_shr(lwTmp, 15);
  lwTmp = L_mac(lwTmp, swIn1H, swIn2H);



  return(lwTmp);
  }



/*
***********************************************************************
*
*   FUNCTION NAME: Mpy_lw_sw
*
*   PURPOSE: To multiply a longword (32-bit) and a shortword (16-bit)
*
*   INPUTS:
*
*     lwIn1 - Input # 1 (longword)
*
*     swIn2 - Input # 2 (shortword)
*
*   OUTPUTS: None
*
*   RETURN VALUE: The product is returned as a longword
*
**********************************************************************
*/

Longword Mpy_lw_sw(Longword lwIn1, Shortword swIn2)
  {
  Shortword swIn1H, swIn1L;
  Longword lwTmp;
  Shortword swTmp;
  swIn1H = extract_h(lwIn1);
  swIn1L = extract_l(lwIn1);
  swIn1L = shr(swIn1L,1);
  swIn1L = swIn1L & 0x7FFF; logic16();
  lwTmp = L_mult(swIn1H, swIn2);
  swTmp = mult(swIn1L, swIn2);
  lwTmp = L_mac(lwTmp, swTmp, 0x0001);
  return(lwTmp);
  }

/*
***********************************************************************
*
*   FUNCTION NAME: Mpy_lw_sw_Add
*
*   PURPOSE: To multiply a longword (32-bit) and a shortword (16-bit)
*            and add the result to a longword (32-bit)
*
*   INPUTS:
*
*     lwInA - Input # 1 (longword to add)
*
*     lwIn1 - Input # 2 (longword to multiply)
*
*     swIn2 - Input # 3 (shortword)
*
*   OUTPUTS: None
*
*   RETURN VALUE: The product is returned as a longword ( lwInA + Mpy_lw_sw(lwIn1,swIn2)
*
**********************************************************************
*/

Longword Mpy_lw_sw_Add(Longword lwInA,Longword lwIn1, Shortword swIn2) { 
  Shortword swIn1H, swIn1L; 
  Longword lwTmp; 
  Shortword swTmp;
  swIn1H = extract_h(lwIn1);
  swIn1L = extract_l(lwIn1);  
  swIn1L = shr(swIn1L,1);  
  swIn1L = 0x7fff & swIn1L ; move16() ;
  lwTmp = L_mac(lwInA,swIn1H, swIn2);  
  swTmp = mult(swIn1L, swIn2); 
  lwTmp = L_mac(lwTmp, swTmp, 0x0001);  
  return(lwTmp);
}
/*
***********************************************************************
*
*   FUNCTION NAME: L_Extract
*
*   PURPOSE: To extract the appropriate components from a longword
*            (32-bit) ready for multiplication with a shortword
*            (16-bit) [See function Mpy_32_16() below]
*
*   INPUTS:
*
*     L_32 - Input longword (32-bit)
*
*   OUTPUTS:
*
*     *hi - Output shortword # 1
*
*     *lo - Output shortword # 2
*
*   RETURN VALUE: None
*
**********************************************************************
*/

void L_Extract (Word32 L_32, Word16 *hi, Word16 *lo)
  {
  *hi = extract_h (L_32);
  *lo = 0x7FFF & shr (extract_l (L_32), 1); logic16();
  return;
  }



/*
***********************************************************************
*
*   FUNCTION NAME: Mpy_32_16
*
*   PURPOSE: To multiply a longword (32-bit) and a shortword (16-bit)
*            after the longword has been split into two components by
*            the function L_Extract() above
*
*   INPUTS:
*
*     hi - First shortword component of the longword
*
*     lo - Second shortword component of the longword
*
*     n  - Shortword
*
*   OUTPUTS: None
*
*   RETURN VALUE: The product is returned as a longword
*
**********************************************************************
*/

Word32 Mpy_32_16 (Word16 hi, Word16 lo, Word16 n)
  {
  Word32 L_32;

  L_32 = L_mult (hi, n);
  L_32 = L_mac (L_32, mult (lo, n), 1);

  return (L_32);
  }



/*
***********************************************************************
*   FUNCTION NAME: UtilityFunctionAtGivenPitchFreq
*
*   PURPOSE: To calculate the utility function value at given pitch
*            frequency
*
*   INPUTS:
*
*     pstSpectralPeaks - Array of spectral peaks
*
*     iNoOfSpectralPeaks - Number of spectral peaks
*
*     lwPitchFreq - Pitch frequency
*
*   OUTPUTS: None
*
*   RETURN VALUE:
*
*     U(lwPitchFreq) - Value of the utility function at the given
*                      pitch frequency is returned
*
***********************************************************************
*/

static Shortword UtilityFunctionAtGivenPitchFreq_be(
					    const POINT_be *pstSpectralPeaks,
                                            const X_INT16 iNoOfSpectralPeaks,
                                            const Longword lwPitchFreq)
{

  X_INT16 i;
  X_INT16 iShift;

  Shortword swAmp;
  Longword lwTemp;
  Longword lwTemp1;
  Longword lwTemp2;

  Shortword swInvPitchFreq;
  Shortword swTemp;


  /*
   * Compute the inverse of the pitch frequency
   */

  lwTemp = lwPitchFreq;		move32();
  iShift = 0;			move16();
  test(); logic32();
  while (lwTemp & 0xFFFF0000)
  {
    iShift = add(iShift,1);
    lwTemp = L_shr(lwTemp,1);
    test(); logic32();
  }
  swTemp = extract_l(L_shr(L_add(lwTemp,1),1));
  test();
  if (sub(swTemp,(Shortword)0x4000) >= 0)
  {
    swInvPitchFreq = div_s((Shortword)0x4000,swTemp);
  }
  else
  {
    iShift = add(iShift,1);
    swInvPitchFreq = 0x7FFF;			move16();
  }
  iShift = sub(iShift,1);


  /*
   * Calculate the utility function in a loop
   */

  swAmp = 0;						move16();
  for (i = 0; i< iNoOfSpectralPeaks; i++)
  {
    /* Multiply the given frequency by the inverse pitch frequency */
    lwTemp = pstSpectralPeaks[i].lwFreq;		move32();
    lwTemp1 = L_mult(extract_h(lwTemp),swInvPitchFreq);
    lwTemp2 = L_mult(extract_l((L_shr(lwTemp,1) & 0x7FFF)),
                        swInvPitchFreq);		logic32();
    lwTemp = L_add(lwTemp1,L_shr(lwTemp2,16));
    lwTemp = L_shr(L_add(lwTemp,L_shl(0x1,iShift)),add(iShift,1));

    /* Compute ABS((Fpeak/F0) - ROUND(Fpeak/F0)) */
    swTemp = extract_l((lwTemp & 0x7FFF));		logic32();
    test();
    if (sub(swTemp,0x4000) > 0)
    {
      swTemp = negate(add(0x8000,swTemp));
    }

    /* Compute contribution to utility function */
    test(); 
    if (sub(swTemp,swUDIST1) < 0)
    {
      swAmp = add(swAmp,pstSpectralPeaks[i].swAmp);
    }
    else 
    {
        test();
        if (sub(swTemp,swUDIST2) < 0)
        {
            swAmp = add(swAmp,L_shr(pstSpectralPeaks[i].swAmp,1));
        }
    }

  }

  return (swAmp);

}



/*
***********************************************************************
*   FUNCTION NAME: ComparePitchFreqAscending_be
*
*   PURPOSE: To help sort an array of PITCH_be elements by the associated
*            frequency values in an ascending order
*
*   INPUTS:
*
*     pvValue1 - Pointer to PITCH_be data structure # 1
*
*     pvValue2 - Pointer to PITCH_be data structure # 2
*
*   OUTPUTS: None
*
*   RETURN VALUE: +1, -1, 0
*      
***********************************************************************
*/

static X_INT16 ComparePitchFreqAscending_be(const void *pvValue1, const void *pvValue2)
{
  // No move()s required - the input arguments can be used directly
  const Longword lwFreq1 = ((PITCH_be *) pvValue1)->lwFreq;
  const Longword lwFreq2 = ((PITCH_be *) pvValue2)->lwFreq;
  Longword tmp;

  tmp = L_sub(lwFreq1,lwFreq2);

  test(); 
  if (tmp < 0)
  {
    return(-1);
  }
  else 
  {
      test();
      if (tmp > 0)
      {
          return(+1);
      }
      else
      {
          return(0);
      }
  }

}



/*
***********************************************************************
*   FUNCTION NAME: ComparePitchFreqDescending_be
*
*   PURPOSE: To help sort an array of PITCH_be elements by the associated
*            frequency values in a descending order
*
*   INPUTS:
*
*     pvValue1 - Pointer to PITCH_be data structure # 1
*
*     pvValue2 - Pointer to PITCH_be data structure # 2
*
*   OUTPUTS: None
*
*   RETURN VALUE: +1, -1, 0
*      
***********************************************************************
*/

static X_INT16 ComparePitchFreqDescending_be(const void *pvValue1, const void *pvValue2)
{
  // No move()s required - the input arguments can be used directly
  const Longword lwFreq1 = ((PITCH_be *) pvValue1)->lwFreq;
  const Longword lwFreq2 = ((PITCH_be *) pvValue2)->lwFreq;

  Longword tmp;

  tmp = L_sub(lwFreq1,lwFreq2);

  test(); 
  if ( tmp > 0)
  {
      return(-1);
  }
  else 
  {
      test();
      if (tmp < 0)
      {
          return(+1);
      }
      else
      {
          return(0);
      }
  }
}



/*
***********************************************************************
*   FUNCTION NAME: CompareIpointAmp_be
*
*   PURPOSE: To help sort an array of IPOINT_be elements by the associated
*            amplitude values
*
*   INPUTS:
*
*     pvValue1 - Pointer to IPOINT_be data structure # 1
*
*     pvValue2 - Pointer to IPOINT_be data structure # 2
*
*   OUTPUTS: None
*
*   RETURN VALUE: +1, -1, 0
*      
***********************************************************************
*/
static X_INT16 CompareIpointAmp_be(const void *pvValue1, const void *pvValue2)
{
  // No move()s required - the input arguments can be used directly
  const Shortword swAmp1 = ((IPOINT_be *) pvValue1)->swAmp;
  const Shortword swAmp2 = ((IPOINT_be *) pvValue2)->swAmp;

  Shortword tmp;
  tmp = sub(swAmp1,swAmp2);

  test();
  if (tmp > 0)
  {
    return(-1);
  }
  else 
  {
      test();
      if (tmp < 0)
      {
          return(+1);
      }
      else
      {
          return(0);
      }
  }

}



/*
***********************************************************************
*   FUNCTION NAME: Compare_ARRAY_OF_XPOINTS_be
*
*   PURPOSE: To help sort an array of ARRAY_OF_XPOINTS_be elements by
*            the associated lengths
*
*   INPUTS:
*
*     pvValue1 - Pointer to ARRAY_OF_XPOINTS_be data structure # 1
*
*     pvValue2 - Pointer to ARRAY_OF_XPOINTS_be data structure # 2
*
*   OUTPUTS: None
*
*   RETURN VALUE: +1, -1, 0
*      
***********************************************************************
*/

static X_INT16 Compare_ARRAY_OF_XPOINTS_be(const void *pvValue1, const void *pvValue2)
{
  // No move()s required - the input arguments can be used directly
  const X_INT16 iLen1 = ((ARRAY_OF_XPOINTS_be*)pvValue1)->length;
  const X_INT16 iLen2 = ((ARRAY_OF_XPOINTS_be*)pvValue2)->length;

  X_INT16 tmp;

  tmp = sub(iLen1,iLen2);

  test();
  if ( tmp < 0)
  {
    return(-1);
  }
  else 
  {
      test();
      if (tmp > 0)
      {
          return(1);
      }
      else
      {
          return(0);
      }
  }
}



/*
***********************************************************************
*   FUNCTION NAME: Final_ScaleDownAmpsOfHighFreqPeaks_be
*
*   PURPOSE: To scale down the amplitudes of spectral peaks in the
*            middle and upper bands if they exceed a predefined
*            threshold relative to the peak value in the lower band
*
*   INPUTS:
*
*     pstPeaks - Array of spectral peaks represented by POINT_be structure
*
*     iNoOfPeaks - Number of peaks
*
*   OUTPUTS: None
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

void Final_ScaleDownAmpsOfHighFreqPeaks_be(
    POINT_be *pstPeaks,
    const X_INT16 iNoOfPeaks)
  {

  X_INT16 i;
  X_INT16 iStart2;
  X_INT16 iStart3;
  X_INT16 iShift;

  Shortword swTemp1;
  Shortword swTemp2;
  Shortword swScaleFactor;

  Shortword swMax1;
  Shortword swMax2;
  Shortword swMax3;
  Shortword swMaxAmpAllowed;
  Shortword swTemp;


  /*  Find maximum spectral value in the first third of the spectrum */
  swMax1 = 0;					move16();
  test(); L_sub(0L,0L);
  for (i = 0; (i < iNoOfPeaks) && (pstPeaks[i].lwFreq <= lwTHIRD_REF_BANDWIDTH); i++)
  {
    test();
    if (sub(pstPeaks[i].swAmp,swMax1) > 0)
    {
      swMax1 = pstPeaks[i].swAmp;		move16();
    }
    test(); L_sub(0L,0L);
  }


  /* Return if all spectral peaks are in the first third */
  test(); test();
  if ((sub(iNoOfPeaks,i) == 0) || (swMax1 == 0))
  {
    return;
  }


  /* Find maximum spectral value in the second third of the spectrum */
  swMax2 = 0;					move16();
  iStart2 = i;					move16();
  test(); L_sub(0L,0L);
  for (i= iStart2 ; (i < iNoOfPeaks) && (pstPeaks[i].lwFreq <= lwTWO_THIRDS_REF_BANDWIDTH); i++)
  {
    test();
    if (sub(pstPeaks[i].swAmp,swMax2) > 0)
    {
      swMax2 = pstPeaks[i].swAmp;		move16();
    }
    test(); L_sub(0L,0L);
  }


  /* Find maximum spectral value the last third of the spectrum */
  swMax3 = 0;					move16();
  iStart3 = i;					move16();
  for (i= iStart3; i < iNoOfPeaks; i++)
  {
    test();
    if (sub(pstPeaks[i].swAmp,swMax3) > 0)
    {
      swMax3 = pstPeaks[i].swAmp;		move16();
    }
  }


  /*
   * Scale down spectral peaks in the [1/3,2/3] bandwidth range, which are
   * above the allowed amplitude. The allowed amplitude in this region
   * is defined as a fraction of the maximum amplitude in the [0,1/3]
   * bandwidth frequency range.
   */

  swMaxAmpAllowed = mult_r(swMax1, swAMP_SCALE_DOWN1);

  test();
  if (sub(swMax2,swMaxAmpAllowed) > 0)
  {
    iShift = norm_s(swMax2);
    swTemp1 = shl(swMax2,iShift);
    swTemp2 = shl(swMaxAmpAllowed,iShift);
    swScaleFactor = div_s(swTemp2,swTemp1);

    for (i = iStart2; i < iStart3; i++)
    {
      test();
      if (sub(pstPeaks[i].swAmp,swMaxAmpAllowed) > 0)
      {
        swTemp = mult_r(pstPeaks[i].swAmp, swScaleFactor);
        pstPeaks[i].swAmp = swTemp;			move16();
      }
    }
  }


  /*
   * Scale down spectral peaks in the [2/3,1] bandwidth range, which are
   * above the allowed amplitude
   */

  swMaxAmpAllowed = shr_r(mult(swMax1, swAMP_SCALE_DOWN2b),1);

  test();
  if (sub(swMax3,swMaxAmpAllowed) > 0)
  {
    iShift = norm_s(swMax3);
    swTemp1 = shl(swMax3,iShift);
    swTemp2 = shl(swMaxAmpAllowed,iShift);
    swScaleFactor = div_s(swTemp2,swTemp1);

    for (i = iStart3; i < iNoOfPeaks; i++)
    {
      test();
      if (sub(pstPeaks[i].swAmp,swMaxAmpAllowed) > 0)
      {
        swTemp = mult_r(pstPeaks[i].swAmp, swScaleFactor);
        pstPeaks[i].swAmp = swTemp;			move16();
      }
    }
  }

  return;

  }



/*
***********************************************************************
*   FUNCTION NAME: Prelim_ScaleDownAmpsOfHighFreqPeaks_be
*
*   PURPOSE: Similar to Final_ScaleDownAmpsOfHighFreqPeaks_be() function
*            but works on peaks represented by IPOINT_be structure
*
*   INPUTS:
*
*     pstPeaks - Array of spectral peaks represented by IPOINT_be structure
*
*     iNoOfPeaks - Number of peaks
*
*   OUTPUTS: None
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

#define		swTHIRD_REF_BANDWIDTH		85
#define		swTWO_THIRDS_REF_BANDWIDTH	171

void Prelim_ScaleDownAmpsOfHighFreqPeaks_be(
				IPOINT_be *pstPeaks,
				const X_INT16    iNoOfPeaks
                        )
  {

  X_INT16 iStart2, iStart3;
  X_INT16 i;

  Shortword swMaxScale;
  Shortword swTmp;
  Shortword swMax1, swMax2, swMax3;
  Shortword swMax1Norm, swMax2Norm, swMax3Norm;
  Shortword swMaxAmpAllowed, swMaxAmpAllowedNorm;

  //
  // Find maximum spectral value in the first third of the spectrum.
  //
  swMax1 = 0x0000;			move16();
  test(); sub(0,0);
  for (i=0; (i < iNoOfPeaks) && (pstPeaks[i].ind <= swTHIRD_REF_BANDWIDTH); i++)
    {
    test();
    if (sub(pstPeaks[i].swAmp,swMax1) > 0)
      {
      swMax1 = pstPeaks[i].swAmp;	move16();
      }
    test(); sub(0,0);
    }

  // Return if all spectral peaks are in the first third
  test(); test();
  if ((sub(iNoOfPeaks,i) == 0) || (swMax1==0)) 
    {
    return;
    }

  //
  // Find maximum spectral value in the second third of the spectrum.
  //
  swMax2 =0x0000;				move16();
  iStart2 = i;					move16();
  test(); sub(0,0);
  for (i = iStart2; (i < iNoOfPeaks) && (pstPeaks[i].ind <= swTWO_THIRDS_REF_BANDWIDTH); i++)
    {
    test();
    if (sub(pstPeaks[i].swAmp,swMax2) > 0)
      {
      swMax2 = pstPeaks[i].swAmp;		move16();
      }
    test(); sub(0,0);
    }

  //
  // Find maximum spectral value the last third of the spectrum.
  //
  swMax3 = 0x0000;				move16();
  iStart3 = i;					move16();
  for (i= iStart3; i < iNoOfPeaks; i++)
    {
    test();
    if (sub(pstPeaks[i].swAmp,swMax3) > 0)
      {
      swMax3 = pstPeaks[i].swAmp;		move16();
      }
    }


  //
  // Scale down spectral peaks in the [1/3,2/3] bandwidth range, which are
  // above the allowed amplitude. The allowed amplitude in this region
  // is defined as a fraction of the maximum amplitude in the [0,1/3] 
  // bandwidth frequency range
  //
  // Normalize max values here to simplify divisions later on.
  swTmp = swMax1;				move16();
  test();
  if (sub(swTmp,swMax2) < 0)
    {
    swTmp = swMax2;				move16();
    }
  test();
  if (sub(swTmp,swMax3) < 0)
    {
    swTmp = swMax3;				move16();
    }
  swMaxScale = norm_s(swTmp);
  swMax1Norm = shl(swMax1, swMaxScale);
  swMax2Norm = shl(swMax2, swMaxScale);
  swMax3Norm = shl(swMax3, swMaxScale);

  swTmp = 13844;
  swMaxAmpAllowed = mult_r(swTmp, swMax1);
  swMaxAmpAllowedNorm = mult_r(swTmp, swMax1Norm);

  test();
  if (sub(swMax2Norm,swMaxAmpAllowedNorm) > 0)
    {
    Shortword swScaleFactor;

    swScaleFactor = div_s(swMaxAmpAllowedNorm, swMax2Norm);

    for (i=iStart2; i< iStart3; i++)
      {             
      test();
      if (sub(pstPeaks[i].swAmp,swMaxAmpAllowed) > 0)
	{
	pstPeaks[i].swAmp = mult_r(pstPeaks[i].swAmp, swScaleFactor);
						  move16();
	}
      }
    }

  //
  // Scale down spectral peaks in the [2/3,1] bandwidth range, which are
  // above the allowed amplitude
  //
  swTmp = 6636;
  swMaxAmpAllowed = mult_r(swTmp, swMax1);
  swMaxAmpAllowedNorm = mult_r(swTmp, swMax1Norm);

  test();
  if (sub(swMax3Norm,swMaxAmpAllowedNorm) > 0)
    {
    Shortword swScaleFactor;

    swScaleFactor = div_s(swMaxAmpAllowedNorm, swMax3Norm);

    for (i=iStart3; i< iNoOfPeaks; ++i)
      {             
      test();
      if (sub(pstPeaks[i].swAmp,swMaxAmpAllowed) > 0)
	{
	pstPeaks[i].swAmp = mult_r(pstPeaks[i].swAmp, swScaleFactor);
						move16();
	}
      }
    }

  }



/*
***********************************************************************
*   FUNCTION NAME: CalculateDoubleWindowDft_be
*
*   PURPOSE: To calculate the DFT of the combined (double) window given
*            the interpolated DFTs of two successive (single) windows
*            and a proper table holding  the complex exponent values
*
*   INPUTS:
*
*     pstFirstWinDft_be - DFT of the first single window 
*
*     iFirstWinShift - Shift value for the above DFT
*
*     pstSecondWinDft_be - DFT of the second single window 
*
*     iSecondWinShift - Shift value for the above DFT
*
*     iDftLen - DFT length
*
*     pstWindowShiftTable_be - Table holding complex exponent values
*                              corresponding to the time shift
*
*   OUTPUTS:
*
*     pstDoubleWinDft_be - DFT of the double window
*
*     *piDoubleWinShift - Shift value for the above DFT
*
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

static void CalculateDoubleWindowDft_be(
		 const CMPLXsw_be *pstFirstWinDft_be,
		 X_INT16 iFirstWinShift,
		 const CMPLXsw_be *pstSecondWinDft_be,
		 X_INT16 iSecondWinShift,
		 CMPLXsw_be *pstDoubleWinDft_be,
		 X_INT16 *piDoubleWinShift,
		 const X_INT16    iDftLen,
		 const CMPLXsw_be *pstWindowShiftTable_be
					 )
  {
  X_INT16 i;
  CMPLX_be stTemp_be;
  X_INT16 iTmp;
  X_INT16 iShift;
  Longword lwTmp1, lwTmp2;
  Longword lwTmp, lwMax;
  Shortword swTmp1, swTmp2;


  lwMax = 0;						move32();
  iTmp = sub(iFirstWinShift, iSecondWinShift);
  test();
  if (iTmp < 0)
    {
    iTmp = negate(iTmp);
    iShift = iFirstWinShift;				move16();
    for (i=0; i< iDftLen; i++)
      {	  
      // 
      // Time shift == multiplication with complex exponents
      //
      lwTmp1 = L_mult(pstSecondWinDft_be[i].swReal,
	  pstWindowShiftTable_be[i].swReal);
      lwTmp2 = L_mult(pstSecondWinDft_be[i].swImag,
	  pstWindowShiftTable_be[i].swImag);
      stTemp_be.lwReal = L_sub(lwTmp1, lwTmp2);

      lwTmp1 = L_mult(pstSecondWinDft_be[i].swReal,
	  pstWindowShiftTable_be[i].swImag);
      lwTmp2 = L_mult(pstSecondWinDft_be[i].swImag,
	  pstWindowShiftTable_be[i].swReal);
      stTemp_be.lwImag = L_add(lwTmp1, lwTmp2);

      //
      // Add result to the DFT of the first frame to get the double-window DFT
      //
      lwTmp1 = L_shr(stTemp_be.lwReal, iTmp);
      lwTmp1 = L_msu(lwTmp1, pstFirstWinDft_be[i].swReal, 0x8000);
      pstDoubleWinDft_be[i].swReal = round(lwTmp1);		move16();
      lwTmp1 = L_abs(lwTmp1);
      lwTmp = L_sub(lwTmp1, lwMax);
      test();
      if (lwTmp > 0)
	{
	lwMax = lwTmp1;			move32();
	}

      lwTmp2 = L_shr(stTemp_be.lwImag, iTmp);
      lwTmp2 = L_msu(lwTmp2, pstFirstWinDft_be[i].swImag, 0x8000);
      pstDoubleWinDft_be[i].swImag = round(lwTmp2);		move16();
      lwTmp2 = L_abs(lwTmp2);
      lwTmp = L_sub(lwTmp2, lwMax);
      test();
      if (lwTmp > 0)
	{
	lwMax = lwTmp2;			move32();
	}
      }
    }
  else
    {
    iShift = iSecondWinShift;				move16();
    for (i=0; sub(i,iDftLen) < 0; i++)
      {
      // 
      // Time shift == multiplication with complex exponents
      //
      lwTmp1 = L_mult(pstSecondWinDft_be[i].swReal,
	  pstWindowShiftTable_be[i].swReal);
      lwTmp2 = L_mult(pstSecondWinDft_be[i].swImag,
	  pstWindowShiftTable_be[i].swImag);
      stTemp_be.lwReal = L_sub(lwTmp1, lwTmp2);

      lwTmp1 = L_mult(pstSecondWinDft_be[i].swReal,
	  pstWindowShiftTable_be[i].swImag);
      lwTmp2 = L_mult(pstSecondWinDft_be[i].swImag,
	  pstWindowShiftTable_be[i].swReal);
      stTemp_be.lwImag = L_add(lwTmp1, lwTmp2);

      //
      // Add result to the DFT of the first frame to get the double-window DFT
      //
      swTmp1 = shr(pstDoubleWinDft_be[i].swReal, iTmp);
      lwTmp1 = L_msu(stTemp_be.lwReal, swTmp1, 0x8000);
      pstDoubleWinDft_be[i].swReal = round(lwTmp1);		move16();
      lwTmp1 = L_abs(lwTmp1);
      lwTmp = L_sub(lwTmp1, lwMax);
      test();
      if (lwTmp > 0)
	{
	lwMax = lwTmp1;			move32();
	}

      swTmp2 = shr(pstDoubleWinDft_be[i].swImag, iTmp);
      lwTmp2 = L_msu(stTemp_be.lwImag, swTmp2, 0x8000);
      pstDoubleWinDft_be[i].swImag = round(lwTmp2);		move16();

      lwTmp2 = L_abs(lwTmp2);
      lwTmp = L_sub(lwTmp2, lwMax);
      test();
      if (lwTmp > 0)
	{
	lwMax = lwTmp2;			move32();
	}
      }
    }
  // Scale
  iTmp = norm_l(lwMax);
  iTmp = sub(iTmp, 1);
  for (i=0; i< iDftLen; i++)
    {
    swTmp1 = shl(pstDoubleWinDft_be[i].swReal, iTmp);
    swTmp2 = shl(pstDoubleWinDft_be[i].swImag, iTmp);
    pstDoubleWinDft_be[i].swReal = swTmp1;		move16();
    pstDoubleWinDft_be[i].swImag = swTmp2;		move16();
    }

  *piDoubleWinShift = add(iShift, iTmp);

  }



/*
***********************************************************************
*   FUNCTION NAME: NormalizeAmplitudes_be
*
*   PURPOSE: Normalizes spectral peaks such that their sum equal 1
*
*   INPUTS:
*
*     iNoOfPeaks - Number of peaks
*
*   INPUT/OUTPUT:
*
*     pstPeaks - Array of spectral peaks (to be normalized)
*
*   OUTPUTS: None
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

static void NormalizeAmplitudes_be(POINT_be *pstPeaks,
                                const X_INT16 iNoOfPeaks)
  {
  X_INT16 i;
  Shortword swSum;
  Longword lwSum;
  Shortword swNorm;
  X_INT16 iShift;
  Shortword swTmp;


  lwSum = 0;		move32();
  for(i=0;  i< iNoOfPeaks; i++)
    {
        lwSum = L_mac(lwSum, pstPeaks[i].swAmp, 0x0001);
    }
  lwSum = L_shr(lwSum,1);

  test();
  if (L_sub(lwSum,0x00007FFF) > 0)
  {
    iShift = 0;
    test(); logic32();
    while (lwSum & 0x7FFF8000)
    {
      lwSum = L_shr(lwSum,1);
      iShift = sub(iShift,1);
      test(); logic32();
    }
    swSum = extract_l(lwSum);
  }
  else
  {
    swSum = extract_l(lwSum);
    iShift = norm_s(swSum);
    swSum = shl(swSum,iShift);
  }
  swNorm = div_s(0x4000,swSum);  
  iShift = add(iShift,1);

  // Normalize all peaks 
  for(i=0; i< iNoOfPeaks; i++)
    {
    swTmp = mult_r(pstPeaks[i].swAmp, swNorm);
    pstPeaks[i].swAmp = shl(swTmp, iShift);	move16();
    }

  return;

  }




/*
***********************************************************************
*   FUNCTION NAME: SelectTopPitchCandidates_be
*
*   PURPOSE: Select a predefined number of the best elements (most likely
*            pitch candidates) out of given array of pitch candidates.
*            Pitch candidate quality measure is combined from the swAmp
*            value and closeness to the previous frame's pitch frequency
*            while some preference is given to higher pitch frequencies
*
*   INPUTS:
*
*     pstPitchCand_be - Array of pitch candidates sorted in ascending
*                       order of pitch frequency
*
*     iNoOfPitchCand - Number of pitch candidates
*
*     lwPrevPitchFreq - Pitch frequency estimate obtained for the
*                       previous frame, or 0 indicating unvoiced frame
*
*     iNoOfOutPitchCand - Maximal number of best candidates to find
*
*
*   OUTPUTS: pstPitch_be - Array of best pitch candidates
*
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

static void SelectTopPitchCandidates_be(
			    const PITCH_be *pstPitchCand_be,
                            const X_INT16  iNoOfPitchCand,
			    const Longword lwPrevPitchFreq,
			    const X_INT16  iNoOfOutPitchCand,
                            PITCH_be       *pstPitch_be
				    )
  {
  X_INT16 i, j, k;
  X_INT16 iIndexBest[MAX_BEST_CANDS];
  BOOL bCond1, bCond2;
  Shortword swAmpBest[MAX_BEST_CANDS];
  Longword lwFreqBest[MAX_BEST_CANDS];
  Shortword swTmp;
  Longword lwTmp;
  Longword lwTmp2;


  for (i=0; i< iNoOfOutPitchCand; i++)
    {
    pstPitch_be[i].lwFreq       = UNVOICED ; 		move32();
    pstPitch_be[i].swAmp        = 0 ;			move16();
    /* use (-(Shortword)swAMP_MARGIN1) instead of negate(swAMP_MARGIN1) */
    swAmpBest[i] = (-(Shortword)swAMP_MARGIN1);		move16();
    lwFreqBest[i] = 0;					move32();
    iIndexBest[i] = -1;					move16();
    }

  test();
  if (0==iNoOfPitchCand)
    {
    return;
    }


  // Select best pitch candidate from maxima list
  // Starting from the highest frequency to the lowest,
  // one of two conditions has to be true in order to
  // select a new pitch
  //
  sub(iNoOfPitchCand,1);
  for(i=iNoOfPitchCand-1; i>=0; i--) {

    for (k=0; k < iNoOfOutPitchCand ; k++)
      {
      swTmp = add(swAmpBest[k], swAMP_MARGIN1);
      swTmp = sub(pstPitchCand_be[i].swAmp, swTmp);
      test();
      if (swTmp > 0)
	{
	bCond1 = 1;		move16();
	}
      else
	{
	bCond1 = 0;		move16();
	}
      swTmp = sub(pstPitchCand_be[i].swAmp, swAmpBest[k]);
      test();
      if(swTmp > 0)
	{
	lwTmp = Mpy_lw_sw(pstPitchCand_be[i].lwFreq, swFREQ_MARGIN1);
	//Compensate for scaling in swFREQ_MARGIN1
	lwTmp = L_shl(lwTmp, 1);
	lwTmp = L_sub(lwTmp, lwFreqBest[k]);
	test();
	if (lwTmp > 0)
	  {
	  bCond2 = 1;		move16();
	  }
	else
	  {
	  bCond2 = 0;		move16();
	  }
	}
      else
	{
	bCond2 = 0;		move16();
	}

      test(); test();
      if (bCond1 || bCond2) break;
      }

    test();
    if (sub(k, iNoOfOutPitchCand) < 0)
      {
      // The pstPitchCand[i] should be inserted at k-th position
      sub(iNoOfOutPitchCand,1);
      for (j= iNoOfOutPitchCand-1; j > k; j--)
	{
//	m = sub(j,1); /* inline [j-1] instead of processing sub(j,1) */
	iIndexBest[j] = iIndexBest[j-1];			move16();
	swAmpBest[j] = swAmpBest[j-1];			move16();
	lwFreqBest[j] = lwFreqBest[j-1];			move32();
	}
      iIndexBest[k] = i;				move16();
      swAmpBest[k] = pstPitchCand_be[i].swAmp;		move16();
      lwFreqBest[k] = pstPitchCand_be[i].lwFreq;	move32();
      }
  }

  //
  // If this frame is part of a stable pitch track,
  // check the possibility to replace the pitch that
  // was selected above by another pitch which is more
  // close the pitch of the previous frame. This is how
  // we make sure that we continue the pitch track -- by
  // giving a second chance to one of the maxima points
  // to be selected.
  //
  test();
  if (lwPrevPitchFreq != 0)
    {

    // Define a frequency region aroud the pitch of the 
    // previous frame
    Longword lwLowFreq;
    Longword lwHighFreq;

    //
    // Go back to the local maxima list, and find the maximum
    // point in this list with frequency that falls in the
    // range defined above
    //	

    X_INT16 iIndexAlt = -1;
    Shortword swAmpAlt = 0;
    move16(); move16();

    lwLowFreq = Mpy_lw_sw(lwPrevPitchFreq, swSTABLE_FREQ_LOWER_MARGIN);
    lwHighFreq = Mpy_lw_sw(lwPrevPitchFreq, swSTABLE_FREQ_UPPER_MARGIN);
    lwHighFreq = L_shl(lwHighFreq, 1);

    lwTmp2 = L_sub(pstPitchCand_be[0].lwFreq, lwHighFreq);
    test();
    for (i=0; (i < iNoOfPitchCand) && (lwTmp2 < 0); i++)
      {
      lwTmp = L_sub(pstPitchCand_be[i].lwFreq, lwLowFreq);
      test();
      if (lwTmp > 0)
	{
	swTmp = sub(pstPitchCand_be[i].swAmp, swAmpAlt);
	test();
	if (swTmp > 0)
	  {
	  swAmpAlt = pstPitchCand_be[i].swAmp;     	move16();
	  iIndexAlt       = i;				move16();
	  }
	}
      lwTmp2 = L_sub(pstPitchCand_be[i].lwFreq, lwHighFreq);
      test();
      }


    test();
    if (iIndexAlt >=0)
      {
      // Check if the alternative candidate already appears in the list
      for (k=0; k< iNoOfOutPitchCand; k++)
	{
	test();
	if (sub(iIndexBest[k],iIndexAlt) == 0)
	  {
	  break;
	  }
	}
      test();
      if (sub(k,iNoOfOutPitchCand) < 0)
	{ // Appears - do not insert it twice!
	iIndexAlt = -1;					move16();
	}
      }


    test();
    if (iIndexAlt >=0)
      {
      // Check if the alternative candidate is good enough
      for (k=0; k < iNoOfOutPitchCand; k++)
	{
	swTmp = add(swAmpAlt, swAMP_MARGIN2);
	swTmp = sub(swTmp, swAmpBest[k]);
	test();
	if (swTmp > 0)
	  {
	  break;
	  }
	}
      test();
      if (sub(k,iNoOfOutPitchCand) < 0)
	{
	// Insert the alternative candidate
        sub(0,0);
	for (j=iNoOfOutPitchCand-1; j > k; j--)
	  {
//	  m = sub(j,1); /* inline [j-1] instead of processing sub(j,1) */
	  iIndexBest[j] = iIndexBest[j-1];		move16();
	  }
	iIndexBest[k] = iIndexAlt;			move16();
	} // endif insert alternative pitch
      } // endif alternative pitch found
    } // endif stable pitch contour


  // Copy the best candidates
  test();
  for (i=0; (i < iNoOfOutPitchCand) && (iIndexBest[i] >= 0); i++)
    {
    pstPitch_be[i].lwFreq = pstPitchCand_be[iIndexBest[i]].lwFreq;  move32();
    pstPitch_be[i].swAmp = pstPitchCand_be[iIndexBest[i]].swAmp;    move16();
    pstPitch_be[i].swCorr = pstPitchCand_be[iIndexBest[i]].swCorr;  move16();
    test();
    }

  return;

  }



/*
***********************************************************************
*   FUNCTION NAME: FindDominantLocalMaximaInUtilityFunction_be
*
*   PURPOSE: To determine MAX_PRELIM_CANDS "best" local maxima on the
*            (piece-wise constant) utility function.  The best maxima
*            are either the highest ones or those that are high enough
*            and located in the vicinity of pre-defined pitch frequency
*            if it is specified
*
*   INPUTS:
*
*     pstPoints_be - Array of nodes defining the piecewise constant
*                    utility function
*
*     iNoOfPoints - Number of nodes
*
*     lwStableFreq - Pre-defined pitch frequency or 0 (for unvoiced)
*
*   OUTPUTS:
*
*     pstPitchCand_be - The best maxima found
*
*     *piNoOfCand - Number of best maxima found
*
*   RETURN VALUE: None
*      
***********************************************************************
*/

static void FindDominantLocalMaximaInUtilityFunction_be(
		const POINT_be *pstPoints_be,
                const X_INT16  iNoOfPoints,
		const Longword lwStableFreq,
                PITCH_be       *pstPitchCand_be,
                X_INT16        *piNoOfCand
				)
  {
  X_INT16 i, k, j, m;
  X_INT16 iNoOfCand;
  // The following arrays can be "allocated" on scratch memory
  X_INT16 aiIndBest[MAX_PRELIM_CANDS];
  Longword lwTmp;
  Shortword aswAmpBest[MAX_PRELIM_CANDS];



  iNoOfCand = 0;				move16();

  //
  // If the utility function is empty then
  // there is no point in searching for pitch candidate. The frame
  // is unvoiced
  test();
  if (0==iNoOfPoints) {
    *piNoOfCand = 0;				move16();
    return;
  }


  //
  // First, we identify all local maxima in the utility function
  // Since the amplitude of the function is assumed constant until
  // the next edge-point, we do not know exactly where the maxima
  // are. So we take a maxima frequency which is in the middle 
  // between the two edge points.

  // Take care of first point
  test();
  if (sub(pstPoints_be[0].swAmp,pstPoints_be[1].swAmp) > 0)
    {
    aswAmpBest[0] = pstPoints_be[0].swAmp;	move16();
    aiIndBest[0] = 0;				move16();
    iNoOfCand = add(iNoOfCand, 1);
    }


  // Check for local maxima in amplitude
  sub(iNoOfPoints,1);
  for (i=1; i < iNoOfPoints-1; i++)
    {
    test(); test();
    if ((sub(pstPoints_be[i].swAmp,pstPoints_be[i-1].swAmp) >= 0) &&
	(sub(pstPoints_be[i].swAmp,pstPoints_be[i+1].swAmp) > 0))
      {
      // This is a local maximum - compare to the stored ones
      for (k=0; k < iNoOfCand; k++) 
	{
	test();
	if (sub(pstPoints_be[i].swAmp,aswAmpBest[k]) >= 0)
	  {
	  break;
	  }
	}
      test();
      if (sub(k,*piNoOfCand) == 0)
	{
	// Too small amplitude - it would not fall in the
	// list of maximal length
	continue;
	}
      // Insert/add at k-th position
      test();
      if (sub(iNoOfCand,*piNoOfCand) < 0)
	{
	j = iNoOfCand;			move16();
	iNoOfCand = add(iNoOfCand, 1);
	}
      else
	{
	j = sub(iNoOfCand, 1);
	}

      sub(j,1);
      for (m=j-1; j > k; j--,m--)
	{
	aswAmpBest[j] = aswAmpBest[m];		move16();
	aiIndBest[j] = aiIndBest[m];		move16();
	}
      aswAmpBest[k] = pstPoints_be[i].swAmp;	move16();
      aiIndBest[k] = i;				move16();

      i++; // the next point cannot be a local maximum
      } //endif local maximum
    } // end for (i=1; i<iNoOfPoints-1; i++)

  i = sub(iNoOfPoints, 1);

  // Take care of last point. The frequency here is always
  // MAX_PITCH_FREQ, and its amplitude applies to this frequency
  // alone
  test();
  if (sub(pstPoints_be[i].swAmp,pstPoints_be[i-1].swAmp) > 0)
    {
    // This is a local maximum - compare to the stored ones
    for (k=0; k < iNoOfCand; k++) 
      {
      test();
      if (sub(pstPoints_be[i].swAmp,aswAmpBest[k]) >= 0)
	{
	break;
	}
      }
    test();
    if (sub(k,*piNoOfCand) < 0)
      {
      test();
      if (sub(iNoOfCand,*piNoOfCand) < 0)
	{
	j = iNoOfCand;				move16();
	iNoOfCand = add(iNoOfCand, 1);
	}
      else
	{
	j = sub(iNoOfCand, 1);
	}
      // Insert/add at k-th position
      sub(j,1);
      for (m= j-1; j > k; j--,m--)
	{
	aswAmpBest[j] = aswAmpBest[m];		move16();
	aiIndBest[j] = aiIndBest[m];		move16();
	}
      aswAmpBest[k] = pstPoints_be[i].swAmp;	move16();
      aiIndBest[k] = i;				move16();
      } //endif local maximum
    }


  //
  // Try to find a local maximum in vicinity of the stable track pitch
  //
  test();
  if (lwStableFreq != UNVOICED)
    {
    Longword lwLowFreq, lwHighFreq;
    Shortword swAmp0;
    Longword lwTmp;
    Shortword swTmp, swTmp2;
    X_INT16   iInd0;

    lwLowFreq = Mpy_lw_sw(lwStableFreq, swSTABLE_FREQ_LOWER_MARGIN);
    lwHighFreq = Mpy_lw_sw(lwStableFreq, swSTABLE_FREQ_UPPER_MARGIN);
    lwHighFreq = L_shl(lwHighFreq, 1);
    swAmp0 = 0;		move16();


    // Find the best local maximum in the vicinity of the stable pitch
    sub(iNoOfPoints,1);
    for (i=1; i < iNoOfPoints-1; i++)
      {
      lwTmp = L_sub(pstPoints_be[i].lwFreq, lwLowFreq);
      test();
      if (lwTmp < 0)
	{
	continue;
	}
      lwTmp = L_sub(pstPoints_be[i].lwFreq, lwHighFreq);
      test();
      if (lwTmp > 0)
	{
	break;
	}
      swTmp = sub(pstPoints_be[i].swAmp, pstPoints_be[i-1].swAmp);
      swTmp2 = sub(pstPoints_be[i].swAmp,  pstPoints_be[i+1].swAmp);
      test(); test();
      if (swTmp >= 0 && swTmp2 >  0)
	{
	// Local maximum
	swTmp = sub(pstPoints_be[i].swAmp, swAmp0);
	test();
	if (swTmp > 0)
	  {
	  swAmp0 = pstPoints_be[i].swAmp;		move16(); 
	  iInd0 = i;					move16();
	  }
	}
      }

    test();
    if (swAmp0 > 0)
      {
      // Check if the local maximum found is already stored
      for (k=0; k < iNoOfCand; k++)
	{
	test();
	if (sub(iInd0,aiIndBest[k]) == 0)
	  {
	  break;
	  }
	}
      test();
      if (sub(k,iNoOfCand) < 0) // it is stored - nothing to be done
	{
	swAmp0 = 0;					move16();
	}
      }

    // Compare to the stored candidates
    test();
    if (swAmp0 > 0)
      {
      for (k=0; k < iNoOfCand; k++) 
	{
	test();
	if (sub(swAmp0, aswAmpBest[k]) >= 0)
	  {
	  break;
	  }
	}
      test();
      if (sub(k,*piNoOfCand) < 0)
	{ 				
	// Insert/add at k-th position
	test();
	if (sub(iNoOfCand,*piNoOfCand) < 0)
	  {
	  j = iNoOfCand;			move16();
	  iNoOfCand = add(iNoOfCand, 1);
	  }
	else
	  {
	  j = sub(iNoOfCand, 1);
	  }
       sub(j,1);
	for (m= j-1; j > k; j--,m--)
	  {
	  aswAmpBest[j] = aswAmpBest[m];		move16();
	  aiIndBest[j] = aiIndBest[m];			move16();
	  }
	aswAmpBest[k] = pstPoints_be[iInd0].swAmp;	move16();
	aiIndBest[k] = iInd0;				move16();
	} // endif insert
      } // endif a local maximum found

    } // endif there is a stable pitch to look around it



  //
  //   Fill in the pitch candidate structures
  //

  sub(iNoOfCand,1);
  for (k=0; k< iNoOfCand-1; k++)
    {
    i = aiIndBest[k];					move16();
    lwTmp = L_add(pstPoints_be[i+1].lwFreq, pstPoints_be[i].lwFreq);
    pstPitchCand_be[k].swAmp = pstPoints_be[i].swAmp;	move16();
    pstPitchCand_be[k].lwFreq = L_shr_r(lwTmp, 1);	move32();
    }

  // The last candidate if obtained from the last point requires special
  // care
  i = aiIndBest[k];					move16();
  pstPitchCand_be[k].swAmp = pstPoints_be[i].swAmp;	move16();
  test();
  if (sub(i,sub(iNoOfPoints,1)) == 0)
    {
    pstPitchCand_be[k].lwFreq = pstPoints_be[i].lwFreq;	move32();
    }
  else
    {
    lwTmp = L_add(pstPoints_be[i+1].lwFreq, pstPoints_be[i].lwFreq);
    pstPitchCand_be[k].lwFreq = L_shr_r(lwTmp, 1);	move32();
    }

  (*piNoOfCand) = iNoOfCand;				move16();

  return;

  }



/*
***********************************************************************
*   FUNCTION NAME: CreatePieceWiseConstantFunction_be
*
*   PURPOSE: To create differential utility function component generated
*            by single given spectral peak within a given frequency band.
*            The differential utility function is piecewise linear and
*            therefore it is defined by an array of its nodes
*
*   INPUTS:
*
*     stSpecPeak_be - Spectral peak
*
*     lwMinFreq - Lower limit of the frequency band
*
*     lwMaxFreq - Upper limit of the frequency band
*
*     pstFrac_be -  FRACTIONS table
*
*     iCreatePieceWiseFunc_LoopLimit - Complexity limitation threshold
*
*   INPUT/OUTPUT:
*
*     *piCreatePieceWiseFunc_LoopCount - Number of the iterations over the
*                           inner loop (equal to the number of utility func.
*                           shapes) performed since the beginning of the
*                           current sub-range processing
*
*   OUTPUTS:
*
*     pstPoint_be - Array of the differential utility function nodes
*
*     *piNoOfPoints - Number of nodes
*
*     *piOutOffset - Offset of the data origin in the output array pstPoint_be
*
*   RETURN VALUE:
*
*     0 - the utility function has been created
*     1 - piCreatePieceWiseFunc_LoopCount reached the limit specified by
*         iCreatePieceWiseFunc_LoopLimit. The utility function has not
*         been calculated
*
***********************************************************************
*/

static X_INT16 CreatePieceWiseConstantFunction_be(
                       const POINT_be stSpecPeak_be, /* not a pointer ?? */
                        const Longword lwMinFreq,
			const Longword lwMaxFreq,
                        const FRACTIONS_be *pstFrac_be,
			XPOINT_be *pstPoint_be,
			X_INT16       *piNoOfPoints,
			X_INT16       *piCreatePieceWiseFunc_LoopCount,
			const X_INT16 iCreatePieceWiseFunc_LoopLimit,
			X_INT16       *piOutOffset
			)
{
	X_INT16 i, j, iStartIndx, iLastIndx;
	X_INT16 iIndx, iFrom, iTo, iN;
	BOOL bZerothShapeFallsInside;
	Longword lwTmp;
	Longword lwF0, lwF1;
	Shortword swALow, swAHigh;
	Shortword swDaHighPos;
	Shortword swDaLowPos;
	Shortword swDaHighNeg;
	Shortword swDaLowNeg;
	Shortword swTmp;
	Longword lwTmp2;
	Shortword swSpecPeakFreq_hi, swSpecPeakFreq_lo;
	Shortword iTo2;


	// Amplitudes of the platos
	swALow = mult_r(stSpecPeak_be.swAmp, swUSTEP);
	swAHigh = stSpecPeak_be.swAmp;			move16();

	// Differential amplitudes
	swDaLowPos = swALow;				move16();
	swDaLowNeg = negate(swALow);
	swDaHighPos = sub(swAHigh, swALow);
	swDaHighNeg = negate(swDaHighPos);
		
	// Nmax
	lwTmp = L_shl(lwMinFreq, 6);
	swTmp = round(lwTmp);
	lwTmp = L_mult(swTmp, swUDIST2);
	lwTmp = L_shr_r(lwTmp, 6);
	lwTmp = L_add(lwTmp, stSpecPeak_be.lwFreq);

	iFrom = 0;					move16();
	lwTmp2 = L_sub(lwTmp, lwMinFreq);
	test();
	while (lwTmp2 > 0)
	  {
	  iFrom = add(iFrom, 1);
	  lwTmp2 = L_sub(lwTmp2, lwMinFreq);
	  test();
	  }
	test();
	if (lwTmp2 == 0)
	  {
	  iFrom = sub(iFrom, 1);
	  }
	// Nmin
	lwTmp = L_shl(lwMaxFreq, 6);
	swTmp = round(lwTmp);
	lwTmp = L_mult(swTmp, swUDIST2);
	lwTmp = L_shr_r(lwTmp, 6);
	lwTmp = L_sub(stSpecPeak_be.lwFreq, lwTmp);

	test();
	if (lwTmp < 0)
	  {
	  bZerothShapeFallsInside = TRUE;               move16();
	  iTo = 1;		move16();
	  }
	else
	  {
	  iTo = 0;		move16();
	  lwTmp2 = L_sub(lwTmp, lwMaxFreq);
	  test();
	  while (lwTmp2 > 0)
	    {
	    iTo = add(iTo, 1);
	    lwTmp2 = L_sub(lwTmp2, lwMaxFreq);
	    test();
	    }
	  bZerothShapeFallsInside = FALSE;		move16();
	  iTo = add(iTo, 1);
	  }

		
	// Check complexity
	iN = sub(iFrom, iTo);
	iN = add(iN, 1);
	test();
	if (iN < 0)
	  {
	  iN = 0;		move16();
	  }

	*piCreatePieceWiseFunc_LoopCount =
				add(*piCreatePieceWiseFunc_LoopCount, iN);
	swTmp = sub(*piCreatePieceWiseFunc_LoopCount,
				iCreatePieceWiseFunc_LoopLimit);

	test();
	if (swTmp > 0)
	  {
	  return 1;
	  }

	iStartIndx = 0;					move16();
	iIndx = 0;					move16();

	L_Extract(stSpecPeak_be.lwFreq, &swSpecPeakFreq_hi, &swSpecPeakFreq_lo);

	iTo2 = iTo;					move16();
	test();
	if (sub(iTo, 1) <= 0)
	  {
	  iTo = add(iTo, 1);
	  }

	for (i=iFrom; i >= iTo; i--)
	  {
	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[i].swFarLow);
	  pstPoint_be[iIndx].lwFreq = lwTmp;		move32();
	  pstPoint_be[iIndx++].swAmp= swDaLowPos;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[i].swLow);
	  pstPoint_be[iIndx].lwFreq = lwTmp;		move32();
	  pstPoint_be[iIndx++].swAmp= swDaHighPos;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[i].swHigh);
	  pstPoint_be[iIndx].lwFreq = lwTmp;		move32();
	  pstPoint_be[iIndx++].swAmp= swDaHighNeg;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[i].swFarHigh);
	  pstPoint_be[iIndx].lwFreq = lwTmp;		move32();
	  pstPoint_be[iIndx++].swAmp= swDaLowNeg;	move16();
	  } // endfor i

	for (j=i; j >= iTo2; j--)
	  {
	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[j].swFarLow);
	  pstPoint_be[iIndx].lwFreq = L_shl(lwTmp, 1);	move32();
	  pstPoint_be[iIndx++].swAmp= swDaLowPos;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[j].swLow);
	  pstPoint_be[iIndx].lwFreq = L_shl(lwTmp, 1);	move32();
	  pstPoint_be[iIndx++].swAmp= swDaHighPos;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[j].swHigh);
	  pstPoint_be[iIndx].lwFreq = L_shl(lwTmp, 1);	move32();
	  pstPoint_be[iIndx++].swAmp= swDaHighNeg;	move16();

	  lwTmp = Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[j].swFarHigh);
	  pstPoint_be[iIndx].lwFreq = L_shl(lwTmp, 1);	move32();
	  pstPoint_be[iIndx++].swAmp= swDaLowNeg;	move16();
	  }

		
	test();
	if (iIndx > 0)
	  { 
	  // Non-empty list - it can be empty because
	  // a "DC-peak" and we didn't include the 0-th shape yet 

	  // Check if the fMinFreq falls into the first shape,
	  // and if yes then modify the points list
	  for (iStartIndx=0; iStartIndx <= 3; iStartIndx++)
	    {
	    lwTmp = L_sub(pstPoint_be[iStartIndx].lwFreq, lwMinFreq);
	    test();
	    if (lwTmp >= 0)
	      {
	      break;
	      }
	    }
	  
	  test();
	  if (iStartIndx > 0)
	    {
	    iStartIndx= sub(iStartIndx, 1);
	    pstPoint_be[iStartIndx].lwFreq = lwMinFreq;		move32();
	    for (i=0; i < iStartIndx; i++)
	      {
	      pstPoint_be[iStartIndx].swAmp =
		add(pstPoint_be[iStartIndx].swAmp, pstPoint_be[i].swAmp);
	      							move16();
	      }
	    }

	  // Check if the fMaxFreq falls into the last shape,
	  // and if yes truncate the points list
	  iLastIndx = sub(iIndx, 1);
	  i = sub(iIndx, 4);
          test();
          if (sub(i,iStartIndx) < 0)
            {
              i = iStartIndx;					move16();
            }
		  
         sub(iIndx,1);
	  for (iLastIndx = iIndx-1 ; iLastIndx >= i ; iLastIndx--)
	    {
	    lwTmp = L_sub(pstPoint_be[iLastIndx].lwFreq, lwMaxFreq);
	    test();
	    if (lwTmp < 0)
	      {
	      break;
	      }
	    }
  
	  iIndx = add(iLastIndx, 1);

	  } // endif non-empty list
	
	test();	
	if (bZerothShapeFallsInside)
	  {
	  lwF0 = L_shr(Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[0].swFarLow), 12);
	  lwF1 = L_shr(Mpy_32_16(swSpecPeakFreq_hi, swSpecPeakFreq_lo, pstFrac_be[0].swLow), 12);

	  lwTmp = L_sub(lwF0, lwMinFreq);
	  test();
	  if (lwTmp < 0)
	    {
	    lwTmp = L_sub(lwF1, lwMinFreq);
	    test();
	    if (lwTmp <= 0)
	      {
	      pstPoint_be[iIndx].lwFreq = lwMinFreq;	move32();
	      pstPoint_be[iIndx++].swAmp = swAHigh;	move16();
	      }
	    else
	      {
	      pstPoint_be[iIndx].lwFreq = lwMinFreq;	move32();
	      pstPoint_be[iIndx++].swAmp = swALow;	move16();

	      lwTmp = L_sub(lwF1, lwMaxFreq);
	      test();
	      if (lwTmp <= 0)
		{
		pstPoint_be[iIndx].lwFreq   = lwF1;	move32();
		pstPoint_be[iIndx++].swAmp  = swDaHighPos;	move16();
		}
	      }
	    }
	  else
	    {
	    pstPoint_be[iIndx].lwFreq   = lwF0;		move32();
	    pstPoint_be[iIndx++].swAmp  = swALow;	move16();
	    lwTmp = L_sub(lwF1, lwMaxFreq);
	    test();
	    if (lwTmp <= 0)
	      {
	      pstPoint_be[iIndx].lwFreq   = lwF1;		move32();
	      pstPoint_be[iIndx++].swAmp  = swDaHighPos;	move16();
	      }
	    }

	  } // endif 0-th shape falls in the frequency interval

	*piNoOfPoints = sub(iIndx, iStartIndx);
	*piOutOffset = iStartIndx;			move16();

	return 0;

}



/*
***********************************************************************
*   FUNCTION NAME: LinkArrayOfPoints_be
*
*   PURPOSE: To convert an array of points to a linked list
*
*   INPUTS:
*
*     iArrLen - Array length
*
*   INPUT/OUTPUT:
*
*     pstPointArr - Array / Linked list
*
*   OUTPUTS: None
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void LinkArrayOfPoints_be(XPOINT_be *pstPointArr,
                              X_INT16 iArrLen)
{

  X_INT16 i;
  sub(iArrLen,1);
  for (i = 0; i < iArrLen-1; i++)
  {
    pstPointArr[i].next = &(pstPointArr[i+1]);  move16();
  }
  pstPointArr[i].next = NULL; move16();

}



/*
***********************************************************************
*   FUNCTION NAME: AddSortedArrayOfPoints_be
*
*   PURPOSE: To insert sorted array of points to sorted linked list of
*            points preserving the sorting order. The points are sorted
*            in ascending order of "lwFreq" data-member
*
*   INPUTS:
*
*     iArrLen - Array length
*
*   INPUT/OUTPUT:
*
*     ppstPointHead - The linked list head
*
*     ppstPointTail - The linked list tail
*
*     pstPointArr - The array - its elements are embedded into the
*                               linked list
*
*   OUTPUTS: None
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void AddSortedArrayOfPoints_be(XPOINT_be **ppstPointHead,
                                   XPOINT_be **ppstPointTail,
                                   XPOINT_be *pstPointArr,
                                   X_INT16 iArrLen)
{

  XPOINT_be *pstPointHead;
  XPOINT_be *pstPointTail;
  XPOINT_be *pScan;
  XPOINT_be *pPrev;

  X_INT16 i;
  X_INT16 i1;
  X_INT16 i2;

  Longword lwFreq;

  // Pointer operations - No counts
  pstPointHead = *ppstPointHead;                        move16();
  pstPointTail = *ppstPointTail;                        move16();

  /*
   * What if the input linked list contains only one element
   */
  test(); sub(0,0);
  if (pstPointHead == pstPointTail)
  {
    LinkArrayOfPoints_be(pstPointArr,iArrLen);
    lwFreq = pstPointHead->lwFreq;			move32();

    for (i = 0; i < iArrLen; i++)
    {
      test();
      if (L_sub(lwFreq,pstPointArr[i].lwFreq) < 0)
      {
        break;
      }

    }

    test();
    if (i == 0)
    {
      /* Insert the single element at the beginning */
      pstPointHead->next = pstPointArr; 	move16();
      *ppstPointTail = &(pstPointArr[iArrLen-1]); move16();
    }
    else 
    {
        test();
        if (sub(i,iArrLen) == 0)
        {
            /* Insert the single element at the end */
            pstPointArr[iArrLen-1].next = pstPointHead; move16();
            pstPointHead->next = NULL;   	    	move16();
            *ppstPointTail = pstPointHead; 	        move16();
            *ppstPointHead = pstPointArr;            move16();
        }
        else
        {
            /* Insert the single element in between */
            pstPointArr[i-1].next = pstPointHead; move16();
            pstPointHead->next = &(pstPointArr[i]); move16();
            *ppstPointHead = &(pstPointArr[0]);           move16();
            *ppstPointTail = &(pstPointArr[iArrLen-1]);   move16();
        }
    }


    return;

  } // end if the input linked list contains only one element


  /*
   * From the current line it's guaranteed that the input linked list
   * contains at least two elements
   */

  /*
   * Check edge conditions in order to reduce the number of
   * compare-operations in the body of the main loop
   */
  test();
  if (L_sub(pstPointArr[0].lwFreq,pstPointHead->lwFreq) < 0)
  {
    /* The first point becomes the list head */
    pstPointArr[0].next = pstPointHead; move16();
    *ppstPointHead = &(pstPointArr[0]);  move16();
    i1 = 1;						move16();
    pScan = pstPointHead; move16();
    pPrev = &(pstPointArr[0]); move16();
  }
  else
  {
    i1 = 0;						move16();
    pScan = pstPointHead->next; move16();
    pPrev = pstPointHead; move16();
  }


  /*
   * From the current line it is guaranteed that any point to be checked
   * cannot become the list head
   */

  test();
  if (L_sub(pstPointArr[iArrLen-1].lwFreq,pstPointTail->lwFreq) > 0)
  {
    pstPointTail->next = &(pstPointArr[iArrLen-1]); move16();
    *ppstPointTail =  &(pstPointArr[iArrLen-1]); move16();
    pstPointArr[iArrLen-1].next = NULL; move16();
    i2 = sub(iArrLen,2);
  }
  else
  {
    i2 = sub(iArrLen,1);
  }


  /*
   * From the current line it's guaranteed that any point to be checked
   * cannot become the list tail
   */

  /*
   *  The main loop
   */
  for (i = i1; i <= i2; i++)
  {

    /* this is not a real for() loop; use while() loop instead */
    move16(); sub(lwFreq, pScan->lwFreq); test();
    for (lwFreq = pstPointArr[i].lwFreq; lwFreq > pScan->lwFreq;)
     {
         pPrev = pScan;  move16(); 
         pScan = pScan->next; move16();
         sub(lwFreq, pScan->lwFreq); test();
     }

    pPrev->next = &(pstPointArr[i]); move16();

    pstPointArr[i].next = pScan;  move16();
    pPrev = &(pstPointArr[i]);  move16();
  }

}



/*
***********************************************************************
*   FUNCTION NAME:  ConvertLinkedListOfDiffPointsToUtilFunc_be
*
*   PURPOSE: To convert differential utility function given by a linked
*            list to utility function
*
*   INPUTS:
*
*     pstPointHead - The linked list head
*
*   OUTPUTS:
*
*     pstPointArr - Array of the utility function nodes
*
*     *piArrLen - Array length
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void ConvertLinkedListOfDiffPointsToUtilFunc_be(XPOINT_be *pstPointHead,
                                                    POINT_be *pstPointArr,
                                                    X_INT16 *piArrLen)
  {

  X_INT16 i;
  XPOINT_be *pCurr;

  pstPointArr[0].swAmp = pstPointHead->swAmp;		move16();
  pstPointArr[0].lwFreq = pstPointHead->lwFreq;		move32();

  i = 0;		move16();

  /* this is not a real for() loop; use while() loop instead */
  move16(); test();
  for (pCurr = pstPointHead->next; pCurr!=NULL; pCurr = pCurr->next)
  {
    test();
    if (L_sub(pCurr->lwFreq,pstPointArr[i].lwFreq) > 0)
    {
      i = add(i,1);					move16();
      pstPointArr[i].swAmp = add(pstPointArr[i-1].swAmp,pCurr->swAmp);
							move16();
      pstPointArr[i].lwFreq = pCurr->lwFreq;		move32();
    }
    else
    {
      pstPointArr[i].swAmp = add(pstPointArr[i].swAmp,pCurr->swAmp);
      							move16();
    }
    move16(); test();
  }

  *piArrLen = add(i,1);

  }



/*
***********************************************************************
*   FUNCTION NAME: CalcUtilityFunction_be
*
*   PURPOSE: To calculate pitch utility function within given
*            frequency band
*
*   INPUTS:
*
*     *pstRom - Pitch ROM tables and parameters
*
*     pvScratchMemory - Scratch memory
*
*     lwMinFreq - Lower limit of the frequency band
*
*     lwMaxFreq - Upper limit of the frequency band
*
*     pstPeaks_be  - Array of spectral peaks
*
*     iNoOfPeaks - Number of spectral peaks
*
*   INPUT/OUTPUT:
*
*     *pstMeter - Pitch meter data structure
*
*   OUTPUTS:
*
*     pstPoints_be - Array of nodes (edge points) defining piecewise
*                    constant utility function
*
*     *piNoOfPoints - Number of nodes
*
*   RETURN VALUE:
*
*     None
*
***********************************************************************
*/

static void CalcUtilityFunction_be(
				const RVC_PITCH_ROM_be *pstRom,
				void  *pvScratchMemory,
                                const Longword lwMinFreq,
                                const Longword lwMaxFreq,
                                const POINT_be *pstPeaks_be, 
				X_INT16   iNoOfPeaks,
				POINT_be *pstPoints_be,
				X_INT16   *piNoOfPoints,
				RVC_PITCH_METER_be *pstMeter
				)
  {

  X_INT16 i, rc, iMaxNoOfBreakPoints;
  const FRACTIONS_be *pstFrac_be = pstRom->astFrac;

  XPOINT_be *pstTmpPointsBuf_be;
  XPOINT_be *pstPointHead_be, *pstPointTail_be;
  XPOINT_be *pstTmpPoints_be;
  ARRAY_OF_XPOINTS_be  *pstArrayOfXpoints_be;
  Longword lwTmp;
  X_INT16 iTmpOffset;


  iMaxNoOfBreakPoints = add(shl(iNoOfPeaks, 1),
  shl(pstMeter->iCreatePieceWiseFunc_LoopLimit, 2));

  //
  // Allocate memory from reusable scratch memory space
  //
  pstTmpPointsBuf_be = (XPOINT_be *)pvScratchMemory;
  pstArrayOfXpoints_be = (ARRAY_OF_XPOINTS_be *)(pstTmpPointsBuf_be +
      iMaxNoOfBreakPoints);

  pstTmpPoints_be = pstTmpPointsBuf_be;
  for(i=0 ; i< iNoOfPeaks; i++)
    {
    rc = CreatePieceWiseConstantFunction_be(
	pstPeaks_be[i],
	lwMinFreq,
	lwMaxFreq,
	pstFrac_be,
	pstTmpPoints_be,
	&(pstArrayOfXpoints_be[i].length),
	&(pstMeter->iCreatePieceWiseFunc_LoopCount),
	pstMeter->iCreatePieceWiseFunc_LoopLimit,
	&iTmpOffset);

    test();
    if (rc) 
      {
      break;
      }

    // Save the address of the newly created array of points
    move16();
    pstArrayOfXpoints_be[i].addr = pstTmpPoints_be + iTmpOffset; //Pointer op.

    // Calculate output address for the following call to the
    // CreatePieceWiseConstantFunction()
    move16();
    pstTmpPoints_be = pstArrayOfXpoints_be[i].addr + pstArrayOfXpoints_be[i].length; // Pointer op.

    }

  iNoOfPeaks = i; move16(); // We could leave the loop upon rc

  //
  // Order the XPoints arrays in ascending order of their lengths
  //
//  qsort_be(pstArrayOfXpoints_be, shl(iNoOfPeaks, 7),
//      sizeof(ARRAY_OF_XPOINTS_be)*256, (void *)Compare_ARRAY_OF_XPOINTS_be);
  qsort_be(pstArrayOfXpoints_be, shl(iNoOfPeaks, 7),
      sizeof(ARRAY_OF_XPOINTS_be)*128, (void *)Compare_ARRAY_OF_XPOINTS_be);

  // 
  // Skip zero-length arrays if any
  //
  test();
  for (i=0; (i< iNoOfPeaks) && (pstArrayOfXpoints_be[i].length==0); i++) 
    {test();}

  test();
  if (sub(i, iNoOfPeaks) == 0)
    {
    *piNoOfPoints = 0;					move16();
    return;
    }

  //
  //   Merge individual arrays together
  //
  LinkArrayOfPoints_be(pstArrayOfXpoints_be[i].addr,
      pstArrayOfXpoints_be[i].length);
  pstPointHead_be = pstArrayOfXpoints_be[i].addr;             move16();
  move16();
  pstPointTail_be = pstPointHead_be + pstArrayOfXpoints_be[i].length -1;
  for (i++; i< iNoOfPeaks; i++)
    {
    AddSortedArrayOfPoints_be(&pstPointHead_be, &pstPointTail_be,
	pstArrayOfXpoints_be[i].addr, pstArrayOfXpoints_be[i].length);
    }

  //
  //  Convert linked list of differential points
  //  to an array representing piecewise constant utility function
  //
  ConvertLinkedListOfDiffPointsToUtilFunc_be(pstPointHead_be,
      pstPoints_be, piNoOfPoints);
  // Terminate the util-function
  lwTmp = L_sub(pstPoints_be[*piNoOfPoints-1].lwFreq, lwMaxFreq);
  test();
  if (lwTmp < 0)
    {
    pstPoints_be[*piNoOfPoints].swAmp = 0;		move16();
    pstPoints_be[*piNoOfPoints].lwFreq = lwMaxFreq;	move32();
    *piNoOfPoints = add(*piNoOfPoints, 1);
    }

  }



/*
***********************************************************************
*   FUNCTION NAME: RefineSpectralPeaks_be
*
*   PURPOSE: To refine power spectral peaks' location and amplitude
*            by parabolic interpolation; apply square root to
*            represent the amplitude spectrum peaks.
*
*   INPUTS:
*
*     plwSqrAmp - Power spectrum
*
*     pstInpPeaks -  Power spectral peaks
*
*     iNoOfPeaks - Number of peaks
*
*
*   OUTPUTS: pstOutPeaks - Output amplitude spectral peaks
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void RefineSpectralPeaks_be(
				Longword *plwSqrAmp,
				IPOINT_be *pstInpPeaks,
				X_INT16    iNoOfPeaks,
				POINT_be  *pstOutPeaks
				)
{
	X_INT16 i, j;
	Longword lwTmp;
	Shortword swTmp;
	X_INT16 i2aScale;
	Longword swFactor;
	Longword lw2a, lw2b, lwc;
	Shortword sw2a, sw2b;
	Longword lwX, lwY;
	Shortword swX;
	Shortword swi;

	// Frequency normalization factor in Q9 
        swFactor = 8000;			move16();

	// If y2 is a local maxima, the values [y1 y2 y3] are used to find
	// a precise location and value of the maxima using quadratic interpolation
	// 
	for(j=0; j < iNoOfPeaks; j++)
	  {
 		i = pstInpPeaks[j].ind;			move16();

                fflush(stdout);
   
		lw2a = L_sub(plwSqrAmp[i-1], plwSqrAmp[i]);
		lw2a = L_sub(lw2a, plwSqrAmp[i]);
		lw2a = L_add(lw2a, plwSqrAmp[i+1]);

		lw2b = L_sub(plwSqrAmp[i+1], plwSqrAmp[i-1]);
		lw2b = L_shr(lw2b, 1);
		
                lwc = plwSqrAmp[i];				move32();

		test();
		if (lw2b == 0)
		  {
		  sw2b = 0;				move16();
		  swX = 0;				move16();
		  i2aScale = 0;				move16();
		  }
		else
		  {
		  i2aScale = norm_l(lw2a);
		  lwTmp = L_shl(lw2b, i2aScale);
		  sw2b = round(lwTmp);
		  lwTmp = L_shl(lw2a, i2aScale);
		  sw2a = round(lwTmp);

		  swX = shr(sw2b, 1);
		  test();
		  if (sw2a < 0)
		    {
		    sw2a = negate(sw2a);
		    swX = negate(swX);
		    }
		  test();
		  if (swX < 0)
		    {
		    swX = negate(swX);
		    swX = div_s(swX, sw2a);
		    }
		  else
		    {
		    swX = div_s(swX, sw2a);
		    swX = negate(swX);
		    }
		  }
		lwY = L_mult(sw2b, swX);
		lwY = L_shr_r(lwY, i2aScale);
		lwY = L_add(lwc, lwY);

		// Peak location in Hz
		// swi in Q6 format
		swi = shl(i, 6);
		lwX = L_mult(swFactor, swX);
		// Convert from Q25 to Q16 format
		lwX = L_shr(lwX, 9-1);
		lwX = L_mac(lwX, swFactor, swi);

		pstOutPeaks[j].lwFreq = lwX;	  	move32();

		// Convert to amplitude
		swTmp = sqrt_l_fix(lwY);
		pstOutPeaks[j].swAmp = swTmp;	move16();
	  }

}



/*
***********************************************************************
*   FUNCTION NAME: CalcSpectrum_be
*
*   PURPOSE: To calculate the power spectrum and smooth it
*
*   INPUTS:
*
*     pswDft - Complex STFT spectrum after Dirichlet interpolation
*
*     plwPowSpecAtEvenPoints - Power spectrum associated with even
*                              elements of the complex spectrum given
*                              by pstDft argument, or NULL
*
*     iDftLen - Length of STFT array
*
*     plwTemp - Temporary work memory
*
*   OUTPUTS:
*
*     plwSqrAmp - Smoothed power spectrum
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void CalcSpectrum_be(
                         CMPLXsw_be *pswDft,
                         const Longword *plwPowSpecAtEvenPoints,
                         const X_INT16 iDftLen,
                         Longword *plwTemp,
                         Longword *plwSqrAmp
                           )

  {
  X_INT16 i, j;
  Longword lwTmp;

 // Calculate power spectrum of given DFT (squared amplitude)

  test();
  if (NULL == plwPowSpecAtEvenPoints)
  {
    for(i = 0; i < iDftLen; i++)
    {
      plwTemp[i] = L_mac(L_mult(pswDft[i].swReal,pswDft[i].swReal),
                                pswDft[i].swImag,pswDft[i].swImag);  move32();
    }
  }
  else
  {
    for (i = 0, j = 0; i < iDftLen; i += 2, j++)
    {
      plwTemp[i] = plwPowSpecAtEvenPoints[j]; move32();
    }

    for (i = 1; i < iDftLen; i += 2)
    {
      plwTemp[i] = L_mac(L_mult(pswDft[i].swReal,pswDft[i].swReal),
                                pswDft[i].swImag,pswDft[i].swImag); move32();
    }
  }


 // Smooth the spectrum with a three-tap MA filter

  plwSqrAmp[0] = plwTemp[0];				move32();
  plwSqrAmp[iDftLen-1] = plwTemp[iDftLen-1];		move32();
  for (i = 1; i < iDftLen-1; i++)
  {
    lwTmp = Mpy_lw_sw(plwTemp[i],swCENTER_WEIGHT);
	  plwSqrAmp[i] = Mpy_lw_sw_Add(lwTmp,L_add(plwTemp[i-1],plwTemp[i+1]),swSIDE_WEIGHT);move32();
  }

}



/*
***********************************************************************
*   FUNCTION NAME: FindPeaks_be
*
*   PURPOSE: To find peaks in the power spectrum
*
*   INPUTS:
*
*     pswSqrAmp - Smoothed power spectrum
*
*     iDftLen - Length of STFT array
*
*     iHighPassCutOffDftPoint - Lower limit of the work frequency band
*
*   OUTPUTS:
*
*     pstPeaks - Array of power spectral peaks
*
*     *piNoOfPeaks - Number of peaks
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void FindPeaks_be(
                      Longword *plwSqrAmp,
                      const X_INT16 iDftLen,
                      IPOINT_be *pstPeaks,
                      X_INT16 *piNoOfPeaks,
                      X_INT16 iHighPassCutOffDftPoint
                        )
  {
  X_INT16 i;
  X_INT16 iMaxCount;


  iMaxCount = 0; move16();

 // Find local maxima
  for (i = iHighPassCutOffDftPoint+2; i < iDftLen-2; i++)
  {
	  
	  test(); 
	  if(L_sub(plwSqrAmp[i],plwSqrAmp[i+1]) > 0) { 
		  test();
		  if (L_sub(plwSqrAmp[i],plwSqrAmp[i-1]) > 0) {
			  test(); test();
			  if ((L_sub(plwSqrAmp[i-1],plwSqrAmp[i-2]) >= 0) ||
				  (L_sub(plwSqrAmp[i+1],plwSqrAmp[i+2]) >= 0))
			  {
				  pstPeaks[iMaxCount].ind = i; move16();
				  pstPeaks[iMaxCount++].swAmp = round(plwSqrAmp[i]); move16();
			  }
			  
			  i++;
		  }
	  }
	  
  }

  (*piNoOfPeaks) = iMaxCount; move16();

}



/*
***********************************************************************
*   FUNCTION NAME: IsLowLevelInput_be
*
*   PURPOSE: To compare frame energy value to a pre-defined threshold
*
*   INPUTS:
*
*     nsFrameEnergy - Frame energy
*
*
*   OUTPUTS: None
*
*
*   RETURN VALUE:
*
*     TRUE - Frame energy is less than the threshold
*     FALSE - Otherwise
*
***********************************************************************
*/

#define		MIN_ENERGY_MANTISSA		25600
#define		MIN_ENERGY_SHIFT		18

static BOOL IsLowLevelInput_be(const NormSw nsFrameEnergy)/* not a pointer ? */
  {
  X_INT16 iTmp;
  Shortword swTmp;
  BOOL b;

  b = FALSE;				move16();
  iTmp = sub(nsFrameEnergy.iShift, MIN_ENERGY_SHIFT); 

  test();
  if(iTmp > 0)
    {
    b=TRUE;				move16();
    return (TRUE);
    }
  else
    {
    iTmp = negate(iTmp);
    swTmp = shr_r(MIN_ENERGY_MANTISSA, iTmp);
    swTmp = sub(nsFrameEnergy.swMantissa, swTmp);
    test();
    if (swTmp < 0)
      {
      b=TRUE;				move16();
      }
    }

  return(b);

  }



/*
***********************************************************************
*   FUNCTION NAME: DirichletInterpolation_be
*
*   PURPOSE: To increase the resolution of a DFT by a factor of two using
*            the approximated Dirichlet interpolation. This is a frequency
*            domain equivalent of the time-domain zero padding.
*
*   INPUTS:
*
*     pswInputDft - Complex DFT array ordered as:
*                   real[0], imag[0], real[1], imag[1], ...
*
*     lwSpecAverage - Average value of the complex DFT spectrum
*
*     iDftSize - FFT length used to compute pswInputDft
*
*     iLastInterpBin - Index of the highest bin of the interpolated
*                      spectrum to be computed
*
*     pswDiricFilter - Imaginary part of the Dirichlet kernel
*
*
*   OUTPUTS: pstOutputDft - Interpolated spectrum
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void DirichletInterpolation_be(
                                   const Shortword *pswInputDft,
                                   const Longword lwSpecAverage,
                                   const X_INT16 iDftSize,
                                   const X_INT16 iLastInterpBin, 
                                   const Shortword *pswDiricFilter,
                                   CMPLXsw_be *pstOutputDft
                                   )
  {
    X_INT16 iLastAvailableBin, iLastNeededBin;
    X_INT16 i,k,j, j_right, j_left;
    X_INT16 iMaxIndex;
    Longword lwReal;
    Longword lwImag;
    Shortword swTmp;


    for (i=1, j=2; i < DIRICHLET_KERNEL_SPAN; i++)
      {
        pstOutputDft[-2*i].swReal =  pswInputDft[j++];		 move16();
        pstOutputDft[-2*i].swImag = negate(pswInputDft[j++]);	 move16();
      }

    iLastAvailableBin = shr(iDftSize, 1);
    iLastNeededBin = add(iLastInterpBin, (DIRICHLET_KERNEL_SPAN -1));
    iMaxIndex = MIN(iLastNeededBin, iLastAvailableBin);      
    sub(iLastNeededBin,iLastAvailableBin);test();move16();

    for (i=0, j=0; i <= iMaxIndex; i++)
      {    
        pstOutputDft[2*i].swReal = pswInputDft[j++];		move16();
        pstOutputDft[2*i].swImag = pswInputDft[j++];		move16();
      }
    
    shl(sub(iLastAvailableBin,1),1);
    for ( j=2*(iLastAvailableBin-1); i <= iLastNeededBin; i++, j-=2)
      {
        pstOutputDft[2*i].swReal =  pswInputDft[j];		move16();
        pstOutputDft[2*i].swImag = negate(pswInputDft[j+1]);	move16();
      }
 
    
    j= shl(iLastInterpBin,1); 
    for (i=1; i < j; i+=2)
      {
      lwReal = lwSpecAverage;					move32();
      lwImag = 0x00000000L;					move32();
      j_right = add(i, 1);
      j_left  = sub(i, 1);
      for (k=0; k < DIRICHLET_KERNEL_SPAN; k++, j_right+=2, j_left-=2)
	{
	swTmp = sub(pstOutputDft[j_left].swImag, pstOutputDft[j_right].swImag);
	lwReal = L_msu(lwReal, pswDiricFilter[k], swTmp);

	swTmp = sub(pstOutputDft[j_left].swReal, pstOutputDft[j_right].swReal);
	lwImag = L_mac(lwImag, pswDiricFilter[k], swTmp);
	}
      pstOutputDft[i].swReal = round(lwReal);			move16();
      pstOutputDft[i].swImag = round(lwImag);			move16();
      }
            
  }



/*
***********************************************************************
*   FUNCTION NAME: PrepareSpectralPeaks_be
*
*   PURPOSE: To determine spectral peaks to be used for pitch candidates
*            generation.
*
*   INPUTS:
*
*     *pstRom - Pitch ROM data structure
*
*     pswInterpDft - Interpolated complex DFT spectrum
*
*     plwPowSpecAtEvenPoints - Power spectrum at even points or NULL
*
*   INPUT/OUTPUT:
*
*     *pstMeter - Pitch meter data structure. Following entries are
*                 modified: pstSpectralPeaks and iNoOfSpectralPeaks
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static X_INT16 PrepareSpectralPeaks_be(
                                 const RVC_PITCH_ROM_be *pstRom,
                                 RVC_PITCH_METER_be *pstMeter,
                                 CMPLXsw_be *pswInterpDft,
                                 const Longword *plwPowSpecAtEvenPoints
                                      )

{

  X_INT16 iNoOfPeaks;
  Longword *plwTemp;
  Longword *plwSqrAmp;
  IPOINT_be *pstPeakLocs;

 // Calculate power spectrum and find peak locations

 // Pointer ops.
  plwSqrAmp = (Longword *)pstMeter->pvScratchMemory;
  plwTemp = plwSqrAmp + pstMeter->iInterpDftLen;              move16();

  CalcSpectrum_be(pswInterpDft,
                  plwPowSpecAtEvenPoints,
                  pstMeter->iInterpDftLen,
                  plwTemp,
                  plwSqrAmp);

  pstPeakLocs = (IPOINT_be *)(plwSqrAmp + pstMeter->iInterpDftLen);

  FindPeaks_be(plwSqrAmp,
               pstMeter->iInterpDftLen,
               pstPeakLocs,
               &iNoOfPeaks,
               pstMeter->iHighPassCutOffDftPoint);

  test();
  if (sub(iNoOfPeaks,MAX_LOCAL_MAXIMA_ON_SPECTRUM) > 0)
  {
    pstMeter->iNoOfSpectralPeaks = 0; move16();
    return 1;
  }


  // Scale down high peaks at high frequnecy if their amplitude is too high
  Prelim_ScaleDownAmpsOfHighFreqPeaks_be(pstPeakLocs,
                                         iNoOfPeaks);

  test();
  if (sub(iNoOfPeaks,MAX_PEAKS_FOR_SORT) > 0)
  {
    iNoOfPeaks = MAX_PEAKS_FOR_SORT; move16();
  }

  // Sort peaks by decreasing amplitude
//  qsort_be(pstPeakLocs,shl(iNoOfPeaks, 7),256*sizeof(IPOINT_be),
//          (void *)CompareIpointAmp_be);
  qsort_be(pstPeakLocs,shl(iNoOfPeaks, 7),128*sizeof(IPOINT_be),
          (void *)CompareIpointAmp_be);

  test();
  if (sub(iNoOfPeaks,MAX_PEAKS_FINAL) > 0)
  {
    iNoOfPeaks = MAX_PEAKS_FINAL; move16();
  }


  // Parabolic interpolation
  RefineSpectralPeaks_be(plwSqrAmp,
                         pstPeakLocs,
                         iNoOfPeaks,
                         pstMeter->pstSpectralPeaks);

  pstMeter->iNoOfSpectralPeaks = iNoOfPeaks; move16();

  Final_ScaleDownAmpsOfHighFreqPeaks_be(pstMeter->pstSpectralPeaks,iNoOfPeaks);

  test();
  if (sub(iNoOfPeaks,MIN_PEAKS) > 0)
  {

   // Try to reduce the number of peaks in order to reduce
   // the number of iterations over utility function building loops

    Shortword swThr;
    Longword lwSum, lwThr;
    X_INT16 i;

    lwSum = 0;						move32();
    for (i = 0; i < iNoOfPeaks; i++)
    {
      lwSum = L_mac(lwSum,pstMeter->pstSpectralPeaks[i].swAmp,0x0001);
    }
    lwThr = Mpy_lw_sw(lwSum, swSUM_FRACTION);

    lwSum = 0;						move32();
    for (i = 0; i< iNoOfPeaks; i++)
    {
      lwSum = L_mac(lwSum, pstMeter->pstSpectralPeaks[i].swAmp,0x0001);
      test();
      if (L_sub(lwSum, lwThr) >= 0)
      {
        i = add(i,1);
        break;
      }
    }
    test();
    if (sub(i,iNoOfPeaks) < 0)
    {
      // the i peaks build up almost all the sum of peak amplitudes
      pstMeter->iNoOfSpectralPeaks = i;		 move16();
    }
    else
    {
      // Take out the peaks which are small relative to peak[MIN_PEAKS-1]
      swThr = mult_r(pstMeter->pstSpectralPeaks[MIN_PEAKS-1].swAmp,
                        swAMP_FRACTION);
      for (i = iNoOfPeaks-1; i >= MIN_PEAKS; i--)
      {
        test();
        if (sub(pstMeter->pstSpectralPeaks[i].swAmp, swThr) >= 0)
        {
          break;
        }
      }
      pstMeter->iNoOfSpectralPeaks = add(i,1);            move16();
    }
  }

  return 0;

}



/*
***********************************************************************
*   FUNCTION NAME: FindPitchCandidates_be
*
*   PURPOSE: To determine a few most likely pitch candidates within
*            predefined frequency band. Compute the candidates'
*            spectral and correlation scores.
*
*   INPUTS:
*
*     *pstRom - Pitch ROM data structure
*
*     *pstMeter - Pitch meter data structure
*
*     lwLowPitchFreq - Frequency band lower limit
*
*     lwHighPitchFreq - Frequency band upper limit
*
*     pswProcSpeechForCorr - Low-pass filtered and downsampled segment
*                            of speech signal as produced by the
*                            pre_process() function
*
*     iNoOfTopCandidates - Maximal number of pitch candidates to be found
*
*
*   OUTPUTS: pstTopCand_be - Array of pitch candidates
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void FindPitchCandidates_be(
		    const RVC_PITCH_ROM_be   *pstRom,
		    RVC_PITCH_METER_be *pstMeter_be,
                    Longword		lwLowPitchFreq,
                    Longword		lwHighPitchFreq,
                    Shortword           *pswProcSpeechForCorr,
                    X_INT16             iNoOfTopCandidates,
                    PITCH_be            *pstTopCand_be
					)
  {
  X_INT16  iNumOfPeaks, i, iNumOfPeaksFinally;
  Shortword swCorr;

  // Pointer ops.
  POINT_be *pstSpectralPeaks_be = (POINT_be *)pstMeter_be->pvScratchMemory;
  void  *pvScratchMemoryLeft = (void *) ((char *)pstMeter_be->pvScratchMemory + MAX_PEAKS_FINAL*sizeof(POINT_be));
  move16();

  test();
  if (pstMeter_be->iNoOfSpectralPeaks == 0)
    {
    pstMeter_be->iNoOfPitchCand = 0;			move16();
    return;
    }

  iNumOfPeaks = MIN(pstMeter_be->iNoOfSpectralPeaks,MAX_PEAKS_PRELIM);
  sub(pstMeter_be->iNoOfSpectralPeaks,MAX_PEAKS_PRELIM); test(); move16();

  for (i=0; i < iNumOfPeaks; i++)
    {
    pstSpectralPeaks_be[i].lwFreq = pstMeter_be->pstSpectralPeaks[i].lwFreq;
    								move32();
    pstSpectralPeaks_be[i].swAmp = pstMeter_be->pstSpectralPeaks[i].swAmp;
    								move16();
    }


  NormalizeAmplitudes_be(pstSpectralPeaks_be,iNumOfPeaks);


  // Calculate the utility function of the given set of spectral peaks
  CalcUtilityFunction_be(
			pstRom,
			pvScratchMemoryLeft,
			lwLowPitchFreq,
			lwHighPitchFreq,
			pstSpectralPeaks_be,
			iNumOfPeaks,
			pstMeter_be->pstPoints,
			&pstMeter_be->iNoOfPoints,
			pstMeter_be
			);

  test();
  if (pstMeter_be->iNoOfPoints == 0)
    {
    pstMeter_be->iNoOfPitchCand = 0;			move16();
    return;
    }

  pstMeter_be->iNoOfPitchCand = MAX_PRELIM_CANDS; 		move16();

  FindDominantLocalMaximaInUtilityFunction_be(
      pstMeter_be->pstPoints,
      pstMeter_be->iNoOfPoints,
      pstMeter_be->stStableTrackPitch.lwFreq,
      pstMeter_be->pstPitchCand,
      &(pstMeter_be->iNoOfPitchCand)
      );

  test();
  if (pstMeter_be->iNoOfPitchCand == 0)
    {
    return;
    }


  // Refine pitch candidates swAmp

  iNumOfPeaksFinally = pstMeter_be->iNoOfSpectralPeaks;		move16();

  for (i=0; i < iNumOfPeaksFinally; i++)
    {
    pstSpectralPeaks_be[i].lwFreq = pstMeter_be->pstSpectralPeaks[i].lwFreq;
    								move32();
    pstSpectralPeaks_be[i].swAmp = pstMeter_be->pstSpectralPeaks[i].swAmp;
    								move16();
    }

  NormalizeAmplitudes_be(pstSpectralPeaks_be,iNumOfPeaksFinally);

  for (i=0; i < pstMeter_be->iNoOfPitchCand; i++)
    {
    pstMeter_be->pstPitchCand[i].swAmp =
        UtilityFunctionAtGivenPitchFreq_be(
	  pstSpectralPeaks_be,
	  iNumOfPeaksFinally,
	  pstMeter_be->pstPitchCand[i].lwFreq);		move16();
    }


  // Sort pitch candidates in ascending order of frequency
  test();
  if (pstMeter_be->iNoOfPitchCand > 0)
    {
//    qsort_be(pstMeter_be->pstPitchCand,
//	shl(pstMeter_be->iNoOfPitchCand, 7),
//	256*sizeof(PITCH_be), (void *)ComparePitchFreqAscending_be);
    qsort_be(pstMeter_be->pstPitchCand,
	shl(pstMeter_be->iNoOfPitchCand, 7),
	128*sizeof(PITCH_be), (void *)ComparePitchFreqAscending_be);
    }


  SelectTopPitchCandidates_be(
      pstMeter_be->pstPitchCand,
      pstMeter_be->iNoOfPitchCand,
      pstMeter_be->stStableTrackPitch.lwFreq,
      iNoOfTopCandidates,
      pstTopCand_be
      );


  for (i=0; i < iNoOfTopCandidates; i++)
    {
    test();
    if (pstTopCand_be[i].lwFreq != UNVOICED)
      {
      compute_pcorr_be(
	  pswProcSpeechForCorr,
	  pstTopCand_be[i].lwFreq,
	  pstMeter_be->iFrameNo,
	  &swCorr);				
      pstTopCand_be[i].swCorr = swCorr;		move16();
      }
    }

  return;
  }



/*
***********************************************************************
*   FUNCTION NAME: ClearPitch_be
*
*   PURPOSE: To set the PITCH_be structure to UNVOICED state
*            frames
*
*
*   INPUT/OUTPUT: *pstPitch - PITCH_be data structure - set to UNVOICED state
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

static void ClearPitch_be(PITCH_be *pstPitch)
  {
  pstPitch->swAmp = 0;					move16();
  pstPitch->swCorr = 0;					move16();
  pstPitch->lwFreq = UNVOICED;				move32();
  }



/*
***********************************************************************
*   FUNCTION NAME: IsContinuousPitch_be
*
*   PURPOSE: To determine if the pitch values associated with two
*            consecutive frames are close to each other
*
*   INPUTS:
*
*     *pstPrevPitch - PITCH_be structure of the previous frame
*
*     *pstPitch - PITCH_be structure of the current frame
*
*
*   OUTPUTS: None
*
*
*   RETURN VALUE: 
*
*     TRUE - Yes
*     FALSE - No
*
***********************************************************************
*/

static BOOL IsContinuousPitch_be(const PITCH_be *pstPrevPitch,
				 PITCH_be *pstPitch)
  {
  Longword lwTemp1;
  Longword lwTemp2;
  BOOL b;

  test();
  if (pstPitch->lwFreq == UNVOICED)
    {
    return FALSE;
    }
  test();
  if (pstPrevPitch->lwFreq == UNVOICED)
    {
    return FALSE;
    }

  b = FALSE;		move16();
  lwTemp1 = Mpy_lw_sw(pstPrevPitch->lwFreq, swSTABLE_FREQ_LOWER_MARGIN);
  lwTemp1 = L_sub(pstPitch->lwFreq, lwTemp1);
  test();
  if (lwTemp1 > 0)
    {
    lwTemp2 = Mpy_lw_sw(pstPrevPitch->lwFreq, swSTABLE_FREQ_UPPER_MARGIN);
    /* UPPER_MARGIN was in Q14, adjust shift */
    lwTemp2 = L_shl(lwTemp2, 1);
    lwTemp2 = L_sub(pstPitch->lwFreq, lwTemp2);
    test();
    if (lwTemp2 < 0)
      {
      b = TRUE;		move16();
      }
    }

  return b;

  }



/*
***********************************************************************
*   FUNCTION NAME: Finalize_be
*
*   PURPOSE: To finalize pitch estimation for current frame: set output
*            pitch value & update pitch estimator for subsequent calls
*
*
*   INPUTS: *pstPitch_be  - Final pitch
*
*
*   INPUT/OUTPUT
*
*     *pstMeter_be - Pitch meter data structure
*
*
*   OUTPUTS: *plwOutPitchPeriod - Pitch cycle length at 8 kHz
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

#define swREF_SAMPLE_RATE_Q0 0x1F40

static void Finalize_be(
			RVC_PITCH_METER_be *pstMeter_be,
			PITCH_be           *pstPitch_be,
			Longword           *plwOutPitchPeriod
			)
  {
  CMPLXsw_be *pstTemp;
  Longword lwTmp1;
  Shortword swTmp;
  X_INT16 iShift;


//Swap pointers, such that the single-window interpolated input-DFT is saved for
//next frame. It is required in order to caluclate the DFT of the double-window
//in the next frame

  // Pointer ops.
  pstTemp                         = pstMeter_be->pstDoubleInterpDft; move16();
  pstMeter_be->pstDoubleInterpDft = pstMeter_be->pstSingleInterpDft; move16();
  pstMeter_be->pstSingleInterpDft = pstTemp;                         move16();

  iShift = pstMeter_be->iDoubleDftShift;			move16();
  pstMeter_be->iDoubleDftShift = pstMeter_be->iSingleDftShift;	move16();
  pstMeter_be->iSingleDftShift = iShift;			move16();


  // Update return arguments
  test();
  if (pstPitch_be->lwFreq == UNVOICED)
    {
    pstMeter_be->iStablePitchCount = 0;				move16();
    *plwOutPitchPeriod = UNVOICED ;        			move16();
    }
  else
    {
    test();
    if (IsContinuousPitch_be(&(pstMeter_be->stPrevPitch),pstPitch_be) == TRUE)
      {
      pstMeter_be->iStablePitchCount = add(pstMeter_be->iStablePitchCount, 1);            
      }
    else 
      {
      pstMeter_be->iStablePitchCount = 0;			move16();
      }

 
    //Convert pitch frequency in Hz to period (in non-integer 8 khz samples)
    iShift = norm_l(pstPitch_be->lwFreq);
    lwTmp1 = L_shl(pstPitch_be->lwFreq, iShift);
    swTmp = round(lwTmp1);
    swTmp = div_s(swREF_SAMPLE_RATE_Q0, swTmp);
    lwTmp1 = L_deposit_h(swTmp);
    iShift = sub(15, iShift);
    *plwOutPitchPeriod = L_shr(lwTmp1, iShift);

    }

  // Save pitch information for next frame (only some of the fields are needed)
  pstMeter_be->stPrevPitch = *pstPitch_be; move16();

  test();
  if (sub(pstMeter_be->iStablePitchCount,MIN_STABLE_FRAMES) >= 0)
    {
    // We're on a stable track
    pstMeter_be->iDistFromStableTrack = 0;			move16();
    pstMeter_be->stStableTrackPitch = *pstPitch_be;             move16();
    }
  else
    {
    test();
    if (sub(pstMeter_be->iDistFromStableTrack,MAX_TRACK_GAP_FRAMES) <= 0)
      {
      // We've left the stable track not long ago
      if (IsContinuousPitch_be(&pstMeter_be->stStableTrackPitch, pstPitch_be))
	{
	// We've returned to the stable track soon after we left it
	pstMeter_be->iDistFromStableTrack = 0;			move16();
	pstMeter_be->stStableTrackPitch = *pstPitch_be;         move16();
	}
      else
	{
	// We did not return to the stable track yet
	  pstMeter_be->iDistFromStableTrack = add(pstMeter_be->iDistFromStableTrack, 1);   move16();
	}
      }
    else
      {
      // We're left the stable track finally
      pstMeter_be->stStableTrackPitch.lwFreq = UNVOICED;	 move32();
      pstMeter_be->iDistFromStableTrack = add(pstMeter_be->iDistFromStableTrack, 1);   move16();
      }
    }

  // Note: stStableTrackPitch.fFreq != 0 indicates that we are on 
  // a stable pitch track - whether original one or renewed one. 

  return;

  }



/*
***********************************************************************
*   FUNCTION NAME: CLOSELY_LOCATED_be
*
*   PURPOSE: To determine if two pitch candidates are closely located
*
*   INPUTS:
*
*     pit1 - First pitch candidate
*
*     pit2 - Second pitch candidate
*
*
*   OUTPUTS: None
*
*
*   RETURN VALUE:
*
*     TRUE - Yes
*     FALSE - No
*
***********************************************************************
*/

#define swCLOSE_FACTOR_Q14 0x4CCD
/* 
   I think that the correct prototype should be:
   [...](const PITCH_be *pit1, const PITCH_be *pit2)
   For the Memory assessment, I will do as if there were pointers.
   */
static BOOL CLOSELY_LOCATED_be(PITCH_be pit1, PITCH_be pit2)
  {
  Longword lwTemp1, lwTemp2;
  BOOL b;

  b = FALSE;							move16();
  lwTemp1 = Mpy_lw_sw(pit1.lwFreq, swCLOSE_FACTOR_Q14);
  /* FACTOR was in Q14, shift to compensate */
  lwTemp1 = L_shl(lwTemp1, 1);
  lwTemp1 = L_sub(lwTemp1, pit2.lwFreq);

  test();
  if (lwTemp1 > 0)
    {
    lwTemp2 = Mpy_lw_sw(pit2.lwFreq, swCLOSE_FACTOR_Q14);
    /* FACTOR was in Q14, shift to compensate */
    lwTemp2 = L_shl(lwTemp2, 1);
    lwTemp2 = L_sub(lwTemp2, pit1.lwFreq);
    test();
    if (lwTemp2 > 0)
      {
      b = TRUE;							move16();
      }
    }

  return(b);

  }



/*
***********************************************************************
*   FUNCTION NAME: GOOD_ENOUGH_be
*
*   PURPOSE: To determine if a given pitch candidate is good enough
*
*
*   INPUTS: pit -  Pitch candidate
*
*
*   OUTPUTS: None
*
*
*   RETURN VALUE:
*
*     TRUE - Yes
*     FALSE - No
*
***********************************************************************
*/

#define swFD_SCORE_THLD1_Q15 0x63D7
#define swFD_SCORE_THLD2_Q15 0x570A
#define swCORR_THLD_Q15  0x651F
#define swSUM_THLD_Q14   0x6667

/* 
   I think that the correct prototype should be:
   [...](const PITCH_be *pit)
   For the Memory assessment, I will do as if there was one pointer.
   */
static BOOL GOOD_ENOUGH_be(PITCH_be pit)
  {
  BOOL b;
  Shortword swTemp1;

  b = FALSE;							move16();
  test();
  if(sub(pit.swAmp, swFD_SCORE_THLD1_Q15) >=0)
    {
    test();
    if (sub(pit.swCorr, swCORR_THLD_Q15) >=0)
      {
      b = TRUE;							move16();
      return (b);
      }
    }

  test();
  if ( sub(pit.swAmp, swFD_SCORE_THLD2_Q15) >= 0)
    {
    swTemp1 = add(shr(pit.swAmp,1), shr(pit.swCorr,1));
    test();
    if (sub(swTemp1, swSUM_THLD_Q14) >=0)
      {
      b = TRUE;						move16();
      }
    }

  return(b);

  }



/*
***********************************************************************
*   FUNCTION NAME: BETTER_be
*
*   PURPOSE: To determine if the first pitch candidate is better than
*            the second
*
*   INPUTS:
*
*     pit1 - First pitch candidate
*
*     pit2 - Second pitch candidate
*
*
*   OUTPUTS: None
*
*
*   RETURN VALUE:
*
*     TRUE - Yes
*     FALSE - No
*
***********************************************************************
*/

/* 
   I think that the correct prototype should be:
   [...](const PITCH_be *pit1, const PITCH_be *pit2)
   For the Memory assessment, I will do as if there were pointers.
   */
static BOOL BETTER_be(PITCH_be pit1, PITCH_be pit2)
  {
  BOOL b;

  b = FALSE;							move16();

  test();
  if (sub(pit1.swCorr, pit2.swCorr) >0)
    {
    test();
    if (sub(pit1.swAmp, pit2.swAmp) > 0)
      {
      b = TRUE;							move16();
      }
    }

  return(b);

  }



/*
***********************************************************************
*   FUNCTION NAME: SelectFinalPitch_be
*
*   PURPOSE: To select the final pitch estimate out of several candidates
*
*   INPUTS:
*
*     pPitchCand_be - Array of pitch candidates
*
*     iNoOfPitchCands - Number of candidates
*
*     *pPrevPitch_be - Pitch estimate obtained for the previous frame
*
*     iStableFrameCount - Length of continuous pitch segment
*
*     *pLastStableTrackPitch_be - Pitch estimate associated with the
*                                 last frame of a stable track
*
*     bStrictMode - TRUE if the function is called with a partial list;
*                   FALSE if it is called with the complete list of
*                   candidates
*
*
*   OUTPUTS: *pBestPitchCand_be - Final pitch candidate
*
*
*   RETURN VALUE: None
*
***********************************************************************
*/

#define lwCRIT0_OFFSET_Q15 0x0000170A
#define swCANDCORR_THLD1_Q15  0x799A
#define swCANDCORR_THLD2_Q15 0x599A
#define swCANDCORR_THLD3_Q15 0x6CCD
#define swCANDAMP_THLD3_Q15  0x68F6

static void  SelectFinalPitch_be(
			      PITCH_be  *pPitchCand_be,
			      X_INT16    iNoOfPitchCands,
			      PITCH_be  *pPrevPitch_be,     
			      X_INT16    iStableFrameCount,
			      PITCH_be  *pLastStableTrackPitch_be,
			      PITCH_be  *pBestPitchCand_be,
			      BOOL       bStrictMode
			      )
  {
  X_INT16 i, iFirstGood, iGood;
  Shortword swTmp;

  // Sort the candidates in descending frequency order,
  // and figure out the actual number of candidates
//  qsort_be(pPitchCand_be, shl(iNoOfPitchCands, 7), 256*sizeof(PITCH_be),
//      (void *)ComparePitchFreqDescending_be);
  qsort_be(pPitchCand_be, shl(iNoOfPitchCands, 7), 128*sizeof(PITCH_be),
      (void *)ComparePitchFreqDescending_be);

  test();
  for (i=0; (i < iNoOfPitchCands)  && (pPitchCand_be[i].lwFreq != UNVOICED); i++) {test();}
  iNoOfPitchCands = i;						move16();
  test();
  if (0 == iNoOfPitchCands)
    {
    ClearPitch_be(pBestPitchCand_be); // Unvoiced        
    return;
    }

  // Try to find high amp & high corr pitch
  iGood = -1;							move16();
  for (i=0; i < iNoOfPitchCands; i++)
    {				
    test();
    if (GOOD_ENOUGH_be(pPitchCand_be[i]))
      {
      iGood = i;						move16();
      break;
      }
    }

  test();
  if (iGood >= 0)
    {
    Longword lwCrit0, lwCrit, lwTmp;

    iFirstGood = iGood;			move16();
    add(iGood,1);
    for (i=iGood+1; i < iNoOfPitchCands; i++)
      {
      test();
      if (!CLOSELY_LOCATED_be(pPitchCand_be[iFirstGood],pPitchCand_be[i]))
	{
	break;
	}
      test();
      if (BETTER_be(pPitchCand_be[i],pPitchCand_be[iGood]))
	{
	iGood = i;			move16();
	}
      }

    lwCrit0 = L_add(L_deposit_l(pPitchCand_be[iGood].swCorr),
                                lwCRIT0_OFFSET_Q15);
    lwCrit0 = L_add(lwCrit0,L_deposit_l(pPitchCand_be[iGood].swAmp));
    /* sub(i,iNoOfPitchCands) < 0 can be replaced by i < iNoOfPitchCands */
    for ( ; i<iNoOfPitchCands; i++)
      {
      test();
      if (GOOD_ENOUGH_be(pPitchCand_be[i]))
	{ 
	lwCrit = L_add(L_deposit_l(pPitchCand_be[i].swAmp),
                       L_deposit_l(pPitchCand_be[i].swCorr));
	lwTmp = L_sub(lwCrit, lwCrit0);
	test();
	if (lwTmp >= 0)
	  {
	  iGood = i;				move16();
	  break;
	  }
	}
      }

    iFirstGood = iGood;				move16();
    add(i,1);
    for (i++; i < iNoOfPitchCands; i++)
      {
      test();
      if (!CLOSELY_LOCATED_be(pPitchCand_be[iFirstGood],pPitchCand_be[i]))
	{
	break;
	}
      test();
      if (BETTER_be(pPitchCand_be[i],pPitchCand_be[iGood])) 
	{
	iGood = i;				move16();
	}
      }

    }


  test();
  if (bStrictMode)
    {
    X_INT16 iClearFlag = 1;
    move16(); test();
    if (iGood >= 0)
      {
      swTmp = sub(pPitchCand_be[iGood].swAmp, swCANDCORR_THLD1_Q15);
      test();
      if (swTmp >= 0)
	{
	swTmp = sub(pPitchCand_be[iGood].swCorr, swCANDCORR_THLD1_Q15);
	test();
	if (swTmp >= 0)
	  {
	  iClearFlag = 0;				move16();
	  pBestPitchCand_be->lwFreq = pPitchCand_be[iGood].lwFreq; move32();
	  }
	}
      }
    test();
    if (iClearFlag)
      {
      ClearPitch_be(pBestPitchCand_be); // Not found
      }

    return;
    }

  test();
  if (iGood >=0)
    {
    pBestPitchCand_be->lwFreq = pPitchCand_be[iGood].lwFreq;	move32();
    return;
    }


  test();
  if (pLastStableTrackPitch_be->lwFreq != UNVOICED)
    {
    // Try to continue stable pitch track
    for (i=0; i < iNoOfPitchCands; i++)
      {
      test();
      if (IsContinuousPitch_be(pLastStableTrackPitch_be, &pPitchCand_be[i]))
	{
	swTmp = sub(pPitchCand_be[i].swAmp, swCANDCORR_THLD2_Q15);
	test();
	if (swTmp > 0)
	  {
	  break;
	  }
	swTmp = sub(pPitchCand_be[i].swCorr, swCANDCORR_THLD2_Q15);
	test();
	if (swTmp > 0)
	  {
	  break;
	  }
	}
      }

    test();
    if (sub(i, iNoOfPitchCands) < 0)
      {
      iGood = i;					move16();
      iFirstGood = iGood;				move16();
      add(iGood,1);
      for (i=iGood+1; i < iNoOfPitchCands; i++)
	{
	test();
	if (!CLOSELY_LOCATED_be(pPitchCand_be[iFirstGood],pPitchCand_be[i]))
	  {
	  break;
	  }
	test();
	if (BETTER_be(pPitchCand_be[i],pPitchCand_be[iGood]))
	  {
	  iGood = i;					move16();
	  }
	}

      pBestPitchCand_be->lwFreq = pPitchCand_be[iGood].lwFreq;	move32();
      }
    else
      {
      ClearPitch_be(pBestPitchCand_be); // Unvoiced
      }

    return;


    } // Endif we're close from the last stable track


  // Try to find continuous pitch
  test(); test();
  if (pPrevPitch_be->lwFreq!=0 && sub(iStableFrameCount,1) > 0)
    {
    iGood = -1;					move16();
    for (i=0; i < iNoOfPitchCands; i++)
      {
      test();
      if (IsContinuousPitch_be(pPrevPitch_be, &pPitchCand_be[i]))
	{
	swTmp = sub(pPitchCand_be[i].swAmp, swCANDCORR_THLD2_Q15);
	test();
	if (swTmp > 0)
	  {
	  iGood = i; 				move16();
	  break;
	  }
	}
      swTmp = sub(pPitchCand_be[i].swCorr, swCANDCORR_THLD2_Q15);
      test();
      if (swTmp > 0)
	{
	iGood = i; 				move16();
	break;
	}
      }
    test();
    if (iGood >= 0)
      {
      add(i,1);
      for (i++; i < iNoOfPitchCands; i++)
	{
	test();
	if (IsContinuousPitch_be(pPrevPitch_be, &pPitchCand_be[i]))
	  {
	  test();
	  if (BETTER_be(pPitchCand_be[i], pPitchCand_be[iGood]))
	    {
	    iGood = i;				move16();
	    }
	  }
	}
      pBestPitchCand_be->lwFreq = pPitchCand_be[iGood].lwFreq;	move32();
      return;
      }

    }


  // Try to find pitch with high Amp or high Corr
  for (i=0; i < iNoOfPitchCands; i++)
    {
    test();
    if (sub(pPitchCand_be[i].swAmp, swCANDAMP_THLD3_Q15) >= 0)
      {
      break;
      }
    test();
    if (sub(pPitchCand_be[i].swCorr, swCANDCORR_THLD3_Q15) >=0)
      {
      break;
      }
    }
  test();
  if (sub(i, iNoOfPitchCands) < 0)
    {
    iGood = i;					move16();
    iFirstGood = iGood;				move16();
    for (i++; i < iNoOfPitchCands; i++)
      {
      test();
      if (!CLOSELY_LOCATED_be(pPitchCand_be[iFirstGood],pPitchCand_be[i]))
	{
	break;
	}
      test();
      if (BETTER_be(pPitchCand_be[i],pPitchCand_be[iGood])) 
	{
	iGood = i;				move16();
	}
      }
    pBestPitchCand_be->lwFreq = pPitchCand_be[iGood].lwFreq;	move32();
    return;
    }

  ClearPitch_be(pBestPitchCand_be); // Unvoiced    
  return;

  }



/*********************************************************
*
*                   API FUNCTIONS
*
***********************************************************/

/*
***********************************************************************
*   FUNCTION NAME: RVC_ResetPitchMeter_be
*
*   PURPOSE: To reset information obtained while processing previous
*            frames
*
*   INPUT/OUTPUT:
*
*     *RvcPitchMeter - Pitch meter data structure
*                      A number of values in the structure are reset
*
*   RETURN VALUE: 
*
*    RVC_OK - Success
*    RVC_ERR_ILLEGAL_ARGUMENT - RvcPitchMeter does not point to
*                               an object of expected type
*
***********************************************************************
*/

X_INT16 RVC_ResetPitchMeter_be(void *RvcPitchMeter)
  {
  X_INT16 indx;

  RVC_PITCH_METER_be *pstMeter;

  pstMeter = (RVC_PITCH_METER_be *)RvcPitchMeter;
  /* 
     I guess that the following is a debug escape... this should 
     not be taken into account in the wMops estimation. Maybe, 
     this should appear in a #ifdef _DEBUG [section] #endif
  */
  if (pstMeter->signature != RVC_PITCH_METER_SIG)
    return RVC_ERR_ILLEGAL_ARGUMENT;

  pstMeter->iStablePitchCount = 0;			move16();
  pstMeter->stPrevPitch.swAmp  = 0;			move16();
  pstMeter->stPrevPitch.lwFreq = UNVOICED;		move32();

  pstMeter->iDistFromStableTrack = 1000;		move16();
  pstMeter->stStableTrackPitch.swAmp = 0;		move16();
  pstMeter->stStableTrackPitch.lwFreq = UNVOICED;	move32();


  // Reset interpolated DFT of the previous SINGLE window (which is by
  // convention stored in Double Window.
  for (indx=0; indx < pstMeter->iInterpDftLen; indx++)
    {
    pstMeter->pstDoubleInterpDft[indx].swReal = 0;	move16();
    pstMeter->pstDoubleInterpDft[indx].swImag = 0;	move16();
    }
  pstMeter->iDoubleDftShift = 0;			move16();

  pstMeter->iFrameNo = 0;				move16();

  return RVC_OK;
  }



/*
***********************************************************************
*   FUNCTION NAME: RVC_MeasurePitch_be
*
*   PURPOSE: To estimate pitch of the current frame
*
*   INPUTS:
*
*     *RvcPitchRom_be - Pitch ROM data structure
*
*     *RvcPitchMeter_be - Pitch meter data structure
*
*     lwSpecAverage - Average spectral value of the frame
*
*     pswProcSpeechForCorr - low-pass filtered, downsampled version of
*                            speech signal created by pre_process()
*
*     pswStft - Complex STFT spectrum (Short-Time Fourier transform)
*
*     swStftScale - Shift value for the above spectrum
*
*     plwPowerSpectrum - Power spectrum
*
*     swPowerSpectrumScale - Not used
*
*     nsFrameEnergy - Frame energy
*
*     iIsSpeech - VAD flag
*
*     iLowBandNoise - Low band noise detection flag
*
*   OUTPUTS:
*
*     *plwOutPitchPeriod - pitch cycle length expressed in samples
*                          corresponding to 8 kHz sampling rate,
*                          or 0 indicating unvoiced frame
*
*   RETURN VALUE:
*
*     RVC_OK - Success
*     RVC_ERROR_ILLEGAL_ARGUMENT - The pitch ROM and/or pitch meter
*                                  structure is not of expected type
*
***********************************************************************
*/

X_INT16 RVC_MeasurePitch_be(const void      *RvcPitchRom_be,
		     void            *RvcPitchMeter_be,
		     const Longword  lwSpecAverage,
		     Shortword       *pswProcSpeechForCorr,
		     const Shortword *pswStft,
		     const X_INT16 swStftScale,
                     const Longword *plwPowerSpectrum,
                     const X_INT16 swPowerSpectrumScale,
		     const NormSw nsLogEnergy,
                     const X_INT16   iIsSpeech,
                     const X_INT16   iLowBandNoise,
		     Longword        *plwOutPitchPeriod)
  {
  X_INT16 swSingleDftScale;
  X_INT16 swDoubleDftScale;
  Longword lwTmp;
  RVC_PITCH_ROM_be   *pstRom_be;
  RVC_PITCH_METER_be *pstMeter_be;

  BOOL bSingleWinSpecPeaks_Prepared;

  // Structures to hold the pitch candidates from each analysis window
  PITCH_be astPitchCand_be[N_OF_BEST_CANDS_SHORT+N_OF_BEST_CANDS_SINGLE+N_OF_BEST_CANDS_DOUBLE];
  PITCH_be *pstShortWinCand_be;
  PITCH_be *pstSingleWinCand_be;
  PITCH_be *pstDoubleWinCand_be;

  PITCH_be stFinalPitch_be;

  X_INT16 rc, i;

  Longword lwStartFreq, lwEndFreq;
  Longword lwShortStartFreq, lwShortEndFreq;
  Longword lwSingleStartFreq, lwSingleEndFreq;
  Longword lwDoubleStartFreq, lwDoubleEndFreq;

  // Pointer ops.
  pstRom_be = (RVC_PITCH_ROM_be *)RvcPitchRom_be;
  pstMeter_be = (RVC_PITCH_METER_be *)RvcPitchMeter_be;

  bSingleWinSpecPeaks_Prepared = FALSE;			move16();

  /* 
     I guess that the following is a debug escape... this should 
     not be taken into account in the wMops estimation. Maybe, 
     this should appear in a #ifdef _DEBUG [section] #endif
  */
  if (pstRom_be->signature != RVC_PITCH_ROM_SIG ||
      pstMeter_be->signature != RVC_PITCH_METER_SIG)
    {
    return RVC_ERR_ILLEGAL_ARGUMENT;
    }

  pstMeter_be->iFrameNo = add(pstMeter_be->iFrameNo, 1);

  // Assign local pointers (to make code more easy to read)
  pstShortWinCand_be = astPitchCand_be;
  pstSingleWinCand_be = pstShortWinCand_be + N_OF_BEST_CANDS_SHORT;
  pstDoubleWinCand_be = pstSingleWinCand_be + N_OF_BEST_CANDS_SINGLE;

  for (i=0; i < N_OF_BEST_CANDS_SHORT; i++)
    {
    ClearPitch_be(&pstShortWinCand_be[i]);	
    }
  for (i=0; i < N_OF_BEST_CANDS_SINGLE; i++)
    {
    ClearPitch_be(&pstSingleWinCand_be[i]);
    }
  for (i=0; i < N_OF_BEST_CANDS_DOUBLE; i++)
    {
    ClearPitch_be(&pstDoubleWinCand_be[i]);
    }


  //
  // Interpolate input DFT of single window by a factor of two,
  // using approximated Dirichlet interpolation.
  //
  DirichletInterpolation_be(
      pswStft,
      lwSpecAverage,
      pstRom_be->iDftSize,
      pstRom_be->i4KhzDftBin, // Last point to be interpolated
      pstRom_be->aswDirichletImag,
      pstMeter_be->pstSingleInterpDft // Output complex interpolated DFT
      );

  pstMeter_be->iSingleDftShift = swStftScale;		move16();


  // If the input has a very low energy, then treat frame as unvoiced
  test(); test();    /* workaround: test to FALSE instead */
  if(iIsSpeech == FALSE || IsLowLevelInput_be(nsLogEnergy) == TRUE)
    {
    ClearPitch_be(&stFinalPitch_be);				
    Finalize_be(pstMeter_be,
	&stFinalPitch_be,
	plwOutPitchPeriod);
    return RVC_OK;
    }


  test();
  if (pstMeter_be->stStableTrackPitch.lwFreq != UNVOICED)
    {

#define swSTARTFREQ_COEFF 0x553F	/* Q15 */
#define swENDFREQ_COEFF   0x4666	/* Q13 */

    // We're on a stable pitch track - adjust frequency search range
    lwTmp = Mpy_lw_sw(pstMeter_be->stStableTrackPitch.lwFreq,
	swSTARTFREQ_COEFF);

    lwStartFreq = MAX(lwTmp, lwDOUBLE_WIN_START_FREQ);
    L_sub(lwTmp,lwDOUBLE_WIN_START_FREQ); test(); move32();

    lwTmp = Mpy_lw_sw(pstMeter_be->stStableTrackPitch.lwFreq,
	swENDFREQ_COEFF);

    /* compensate for Q13->Q15 */
    lwTmp = L_shl(lwTmp, 2);
    lwEndFreq = MIN(lwTmp, lwSHORT_WIN_END_FREQ);
    L_sub(lwTmp,lwSHORT_WIN_END_FREQ); test(); move32();

    test();
    if (L_sub(lwStartFreq, lwDOUBLE_WIN_END_FREQ) < 0)
      {
      // There is an overlap with DOUBLE window
      lwDoubleStartFreq = lwStartFreq;			move32();
      lwDoubleEndFreq = MIN(lwEndFreq, lwDOUBLE_WIN_END_FREQ);
      L_sub(lwEndFreq,lwDOUBLE_WIN_END_FREQ); test(); move32();
      }
    else 
      {
      lwDoubleEndFreq = 0xFFFF0000;			move32();
      lwDoubleStartFreq = lwDoubleEndFreq;		move32();
      }

    test();
    if (L_sub(lwStartFreq, lwSINGLE_WIN_END_FREQ) < 0)
      {
      test();
      if (L_sub(lwEndFreq, lwSINGLE_WIN_START_FREQ) > 0)
	{                
	// There is an overlap with SINGLE window
	lwSingleStartFreq = MAX(lwStartFreq, lwSINGLE_WIN_START_FREQ);
        L_sub(lwStartFreq,lwSINGLE_WIN_START_FREQ); test(); move32();
	lwSingleEndFreq = MIN(lwEndFreq, lwSINGLE_WIN_END_FREQ);
        L_sub(lwEndFreq,lwSINGLE_WIN_END_FREQ); test(); move32();
	}
      }
    else
      {
      lwSingleEndFreq = 0xFFFF0000;			move32();
      lwSingleStartFreq = lwSingleEndFreq;		move32();
      }

    test();
    if (L_sub(lwEndFreq, lwSHORT_WIN_START_FREQ) >0)
      {
      // There is an overlap with SHORT window
      lwShortStartFreq = MAX(lwStartFreq, lwSHORT_WIN_START_FREQ);
      L_sub(lwStartFreq,lwSHORT_WIN_START_FREQ); test(); move32();
      lwShortEndFreq = lwEndFreq;			move32();
      }
    else 
      {
      lwShortEndFreq = 0xFFFF0000;			move32();
      lwShortStartFreq = lwShortEndFreq;		move32();
      }

    }
  else
    {
    lwStartFreq = lwDOUBLE_WIN_START_FREQ;		move32();
    lwEndFreq = lwSHORT_WIN_END_FREQ;			move32();
    lwShortStartFreq = lwSHORT_WIN_START_FREQ;		move32();
    lwShortEndFreq = lwSHORT_WIN_END_FREQ;		move32();
    lwSingleStartFreq = lwSINGLE_WIN_START_FREQ;	move32();
    lwSingleEndFreq = lwSINGLE_WIN_END_FREQ;		move32();
    lwDoubleStartFreq = lwDOUBLE_WIN_START_FREQ;	move32();
    lwDoubleEndFreq = lwDOUBLE_WIN_END_FREQ;        	move32();
    }

  test();
  if (iLowBandNoise)
    {
    // Don't use spectral peaks resided in low frequency band
    pstMeter_be->iHighPassCutOffDftPoint = pstRom_be->iHighPassCutOffDftPoint;
    move16();
    }
  else
    {
    // Use spectral peaks wherever they are
    pstMeter_be->iHighPassCutOffDftPoint = 0;		 move16();
    }


  /********************************************************************
   *
   *
   *                           SHORT WINDOW
   *
   *
   ********************************************************************/

  test();
  if (L_sub(lwShortStartFreq, lwShortEndFreq) <0)
    {
    pstMeter_be->iCreatePieceWiseFunc_LoopCount = 0;		move16();
    pstMeter_be->iCreatePieceWiseFunc_LoopLimit =
      CREATE_PIECEWISE_FUNC_LOOP_LIM_SH;		move16();

    swSingleDftScale = pstMeter_be->iSingleDftShift;		move16();

    rc = PrepareSpectralPeaks_be(
	pstRom_be,
	pstMeter_be,
	pstMeter_be->pstSingleInterpDft,
	plwPowerSpectrum
	);

    bSingleWinSpecPeaks_Prepared = TRUE;		move16();

    FindPitchCandidates_be(
	pstRom_be,
	pstMeter_be,
	lwShortStartFreq,
	lwShortEndFreq,
	pswProcSpeechForCorr,
	N_OF_BEST_CANDS_SHORT,
	pstShortWinCand_be
	);

    SelectFinalPitch_be(
	astPitchCand_be,
	N_OF_BEST_CANDS_SHORT,
	&(pstMeter_be->stPrevPitch),
	pstMeter_be->iStablePitchCount,
	&(pstMeter_be->stStableTrackPitch),
	&stFinalPitch_be,
	TRUE);

    test();
    if (stFinalPitch_be.lwFreq != UNVOICED)
      {    				
      Finalize_be(pstMeter_be,
	  &stFinalPitch_be,
	  plwOutPitchPeriod);
      return RVC_OK;
      }

    } // end of short window anlaysis


  /********************************************************************
   *
   *
   *                           SINGLE WINDOW
   *
   *
   *********************************************************************/

  test();
  if (L_sub(lwSingleStartFreq, lwSingleEndFreq) < 0)
    {

    pstMeter_be->iCreatePieceWiseFunc_LoopCount = 0;		move16();
    pstMeter_be->iCreatePieceWiseFunc_LoopLimit =
      CREATE_PIECEWISE_FUNC_LOOP_LIM_SNG;		move16();

    test();
    if (!bSingleWinSpecPeaks_Prepared)
    {			
      rc = PrepareSpectralPeaks_be(
	  pstRom_be,
	  pstMeter_be,
	  pstMeter_be->pstSingleInterpDft,
	  plwPowerSpectrum
	  );
    }

    FindPitchCandidates_be(
	pstRom_be,
	pstMeter_be,
	lwSingleStartFreq,
	lwSingleEndFreq,
	pswProcSpeechForCorr,
	N_OF_BEST_CANDS_SINGLE,
	pstSingleWinCand_be
	);	

    SelectFinalPitch_be(
	astPitchCand_be,
	N_OF_BEST_CANDS_SHORT+N_OF_BEST_CANDS_SINGLE,
	&(pstMeter_be->stPrevPitch),
	pstMeter_be->iStablePitchCount,
	&(pstMeter_be->stStableTrackPitch),
	&stFinalPitch_be,
	TRUE);

    test();
    if (stFinalPitch_be.lwFreq != UNVOICED)
      {    				
      Finalize_be(pstMeter_be,
	  &stFinalPitch_be,
	  plwOutPitchPeriod);
      return RVC_OK;
      }

    } // end of single window anlaysis


  /********************************************************************
   *
   *
   *                           DOUBLE WINDOW
   *
   *
   ********************************************************************/	

  test();
  if (L_sub(lwDoubleStartFreq, lwDoubleEndFreq) <0)
    {

    pstMeter_be->iCreatePieceWiseFunc_LoopCount = 0;		move16();
    pstMeter_be->iCreatePieceWiseFunc_LoopLimit =
        CREATE_PIECEWISE_FUNC_LOOP_LIM_DBL;		move16();

    CalculateDoubleWindowDft_be(
	pstMeter_be->pstDoubleInterpDft,
	pstMeter_be->iDoubleDftShift,
	pstMeter_be->pstSingleInterpDft,
	pstMeter_be->iSingleDftShift,
	pstMeter_be->pstDoubleInterpDft,
	&pstMeter_be->iDoubleDftShift,
	pstRom_be->iInterpDftLen,
	pstRom_be->pstWindowShiftTable
	);

    swDoubleDftScale = pstMeter_be->iDoubleDftShift;		move16();

    rc = PrepareSpectralPeaks_be(
	pstRom_be,
	pstMeter_be,
	pstMeter_be->pstDoubleInterpDft,
	NULL);		

    FindPitchCandidates_be(
	pstRom_be,
	pstMeter_be,
	lwDoubleStartFreq,
	lwDoubleEndFreq,
	pswProcSpeechForCorr,
	N_OF_BEST_CANDS_DOUBLE,
	pstDoubleWinCand_be
	);

    } // end of double window anlaysis


  /********************************************************************
   *
   *
   *                  PITCH DECISION
   *
   *
   ********************************************************************/

  SelectFinalPitch_be(
      astPitchCand_be,
      N_OF_BEST_CANDS,
      &(pstMeter_be->stPrevPitch),
      pstMeter_be->iStablePitchCount,
      &(pstMeter_be->stStableTrackPitch),
      &stFinalPitch_be,
      FALSE);

  Finalize_be(pstMeter_be,
      &stFinalPitch_be,
      plwOutPitchPeriod);

  return RVC_OK;

  }

