/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rvc_pitch_api_B.h"
#include "rvc_pitch_B.h"
#include "ParmType.h"
#include "rvc_pitch_init_B.h"

/* ETSI Libraries */
#include "typedef.h"
#include "basic_op.h"
#include "count.h"


// ===============================================================
//
//
//                Private  Functions
//
//
// ===============================================================


/*
***********************************************************************
*
*   FUNCTION NAME: Allocate_InterpolatedDft_be
*
*   PURPOSE: To allocate space for holding interpolated DFT values
*
*   INPUTS:
*
*      *pstRom - RVC_PITCH_ROM_be structure
*
*   INPUT/OUTPUT:
*
*      **ppstInterpDft - pointer to pointer to the
*                        Interpolated DFT structure
*
*   OUTPUTS:
*
*      None
*
*   RETURN VALUE:
*
*      RVC_OK - Success
*      RVC_ERR_NOT_ENOUGH_MEMORY - Memory allocation failure
*
**********************************************************************
*/

static int Allocate_InterpolatedDft_be(
    RVC_PITCH_ROM_be   *pstRom,
    CMPLXsw_be         **ppstInterpDft
    )
  {
  CMPLXsw_be *pstInterpDft;
  int iReturnCode;

  *ppstInterpDft = NULL;

  MALLOCATE(pstInterpDft,CMPLXsw_be,pstRom->iInterpDftLen+2*2*(DIRICHLET_KERNEL_SPAN-1),iReturnCode);
  if (RVC_OK != iReturnCode) return iReturnCode;
  pstInterpDft += 2*(DIRICHLET_KERNEL_SPAN-1);
  *ppstInterpDft = pstInterpDft;

  return RVC_OK;
  }



/*
***********************************************************************
*
*   FUNCTION NAME: Deallocate_InterpolatedDft_be
*
*   PURPOSE: To de-allocate the space for holding interpolated DFT values
*
*   INPUTS:
*
*      *pstInterpDft - Pointer to the allocated space
*
*   OUTPUTS:
*
*      None - the space is de-allocated
*
*   RETURN VALUE:
*
*      None
*
**********************************************************************
*/

static void Deallocate_InterpolatedDft_be(CMPLXsw_be *pstInterpDft)
  {
  pstInterpDft -= 2*(DIRICHLET_KERNEL_SPAN-1);
  MDEALLOCATE(pstInterpDft);	
  }



/* ======================================================================

   API Functions

   ======================================================================= */

/*
***********************************************************************
*
*   FUNCTION NAME: RVC_ConstructPitchRom_be
*
*   PURPOSE: To create and initialize the RVC_PITCH_ROM_be structure
*            containing information needed for pitch extraction
*
*   INPUTS:
*
*      iSampleRate - Sampling rate of the input signal
*
*      swPreEmphFactor - Pre-emphasis factor (0 for XAFE)
*
*   INPUT/OUTPUT:
*
*      pRvcPitchRom - Address where to put the pointer to
*                     RVC_PITCH_ROM_be structure
*
*   OUTPUTS:
*
*      None - the structure is created and initialized with
*             appropriate values
*
*   RETURN VALUE:
*
*      RVC_OK - Success
*      RVC_ERR_NOT_ENOUGH_MEMORY - Memory allocation failure
*
**********************************************************************
*/

X_INT16 RVC_ConstructPitchRom_be(
        const X_INT16  iSampleRate,
        const Shortword swPreEmphFactor,
        void  **pRvcPitchRom
        )
  {
  RVC_PITCH_ROM_be *pstRom;
  X_INT16 iReturnCode;

  *pRvcPitchRom = NULL;

  MALLOCATE(pstRom,RVC_PITCH_ROM_be,1,iReturnCode);
  if (RVC_OK != iReturnCode) return iReturnCode;

  pstRom->signature = RVC_PITCH_ROM_SIG;

  //
  //   Set sampling rate related parameters   
  //
  pstRom->iSampleRate = iSampleRate;
  switch (iSampleRate) {
    case 8000:
      pstRom->iWinSize   = 200;
      pstRom->iFrameSize = 80;
      pstRom->iDftSize   = 256;			
      break;
    case 16000:
      pstRom->iWinSize   = 400;
      pstRom->iFrameSize = 160;
      pstRom->iDftSize   = 512;
      break;
    default:
      return RVC_ERR_ILLEGAL_ARGUMENT;
  } // endswitch

  pstRom->swPreEmph = swPreEmphFactor;

  // Calculate the DFT bin which coresponds to 4kHz (this
  // is the last DFT bin that will be analyzed in order to
  // find the pitch).
  // @8000  Hz sampling rate (256 DFT size) this is: 128 
  // @16000 Hz sampling rate (512 DFT size) this is: 128 
  pstRom->i4KhzDftBin = 
    (int)((REF_BANDWIDTH*pstRom->iDftSize)/(float)iSampleRate);

  // Length of interpolated DFT
  //
  // @8000  Hz sampling rate (512  DFT size) this is: 257 (0-256) 
  // @11000 Hz sampling rate (512  DFT size) this is: 187 (0-186)
  // @16000 Hz sampling rate (1024 DFT size) this is: 257 (0-256)
  pstRom->iInterpDftLen = 2*pstRom->i4KhzDftBin+1;

  // No pre-emphasis correction required
  pstRom->pswInvPreEmpFilter = NULL;

  // Initialize fractions of utility function
  pstRom->astFrac = (FRACTIONS_be *)ROM_astFrac;

  // Initialize complex exponent "shift" factor needed to generate a DFT of 
  // the double window
  pstRom->pstWindowShiftTable = ROM_pstWindowShiftTable;

  // Init Dirichlet Interpolation
  pstRom->aswDirichletImag = ROM_aswDirichletImag;

  // Factor 2: 2*iDftSize is the size of the interpolated DFT
  pstRom->iHighPassCutOffDftPoint = (int)(HIGHPASS_CUTOFF_FREQ*
      2.f*(float)pstRom->iDftSize/(float)pstRom->iSampleRate);

  *pRvcPitchRom = (void*)pstRom;
  return RVC_OK;
  }



/*
***********************************************************************
*
*   FUNCTION NAME: RVC_ConstructPitchMeter_be
*
*   PURPOSE: To create and initialize the RVC_PITCH_METER_be structure
*            containing memory and history information for pitch extraction
*
*   INPUTS:
*
*      *RvcPitchRom - RVC_PITCH_ROM_be structure
*
*   INPUT/OUTPUT:
*
*      pRvcPitchMeter - Address where to put the pointer to
*                       the RVC_PITCH_METER_be structure
*
*   OUTPUTS:
*
*      None - the structure is created and initialized with 
*             appropriate values
*
*   RETURN VALUE:
*
*      RVC_OK - Success
*      RVC_ERR_ILLEGAL_ARGUMENT - RvcPitchRom does not point to
*                                 an object of expected type
*
**********************************************************************
*/

X_INT16 RVC_ConstructPitchMeter_be(
          const void   *RvcPitchRom, 
          void         **pRvcPitchMeter			                        
          )
  {
  RVC_PITCH_ROM_be *pstRom;
  RVC_PITCH_METER_be *pstMeter;
  X_INT16 iReturnCode, size;
  X_INT16 iMaxNoOfBreakPoints;

  *pRvcPitchMeter = NULL;

  pstRom = (RVC_PITCH_ROM_be *)RvcPitchRom;
  if (pstRom->signature != RVC_PITCH_ROM_SIG)
    return RVC_ERR_ILLEGAL_ARGUMENT;

  MALLOCATE(pstMeter, RVC_PITCH_METER_be , 1, iReturnCode);
  if (RVC_OK != iReturnCode) return iReturnCode;  

  pstMeter->signature = RVC_PITCH_METER_SIG;

  // Maximal number of break points in derivative of the Utility Function = 
  // 2 points for each 0-th shape and 4 points x max number of regular shapes
  iMaxNoOfBreakPoints = 4*CREATE_PIECEWISE_FUNC_LOOP_LIM_DBL +
    2*MAX_PEAKS_PRELIM;

  // ----------------------------------------------------------------------
  // Allocate scratch memory. This memory is actually allocated in
  // DoPitchExtract() as an automatic variable.
  // ----------------------------------------------------------------------

  // CalcUtilityFunction:
  size = MAX_PEAKS_FINAL*sizeof(POINT_be) + 
    iMaxNoOfBreakPoints*sizeof(XPOINT_be) + 
    MAX_PEAKS_PRELIM*sizeof(ARRAY_OF_XPOINTS_be);

  if((size+1)/sizeof(X_INT16) > SIZE_SCRATCH_DOPITCH)
    return RVC_ERR_NOT_ENOUGH_MEMORY ;

  // Memory for single and double interpolated DFT
  iReturnCode = Allocate_InterpolatedDft_be(pstRom, &(pstMeter->pstSingleInterpDft));
  if (RVC_OK != iReturnCode) return iReturnCode;

  iReturnCode = Allocate_InterpolatedDft_be(pstRom,&(pstMeter->pstDoubleInterpDft));
  if (RVC_OK != iReturnCode) return iReturnCode;

  pstMeter->iSingleDftShift = 0;
  pstMeter->iDoubleDftShift = 0;

  pstMeter->iInterpDftLen = pstRom->iInterpDftLen;

  // Allocate another scratch memory for holding spectral peaks.
  // This memory is actually allocated in DoAdvProcess() as an
  // automatic variable.

  size = MAX_PEAKS_FINAL * sizeof(POINT_be);
  size += (iMaxNoOfBreakPoints+1) * sizeof(POINT_be);

  if((size+1)/sizeof(X_INT16) > SIZE_SCRATCH_ADVPROCESS)
    return RVC_ERR_NOT_ENOUGH_MEMORY ;

  // Memory for preliminary pitch candidates
  MALLOCATE(pstMeter->pstPitchCand, PITCH_be, MAX_PRELIM_CANDS, iReturnCode)
    if (RVC_OK != iReturnCode) return iReturnCode;

  // Initialize data member variables
  RVC_ResetPitchMeter_be((void*)pstMeter);

  // Update address of input argument
  *pRvcPitchMeter = (void*)pstMeter;

  return RVC_OK;
  }


/*
***********************************************************************
*
*   FUNCTION NAME: RVC_DestructPitchMeter_be
*
*   PURPOSE: To de-allocate memory assigned to RVC_PITCH_METER_be structure
*
*   INPUTS:
*
*      *RvcPitchMeter - RVC_PITCH_METER_be structure
*
*   OUTPUTS:
*
*      None - the memory in the structure is de-allocated
*
*   RETURN VALUE:
*
*      RVC_OK - Success
*      RVC_ERR_ILLEGAL_ARGUMENT - RvcPitchMeter does not point to
*                                 an object of expected type
*
**********************************************************************
*/

X_INT16 RVC_DestructPitchMeter_be(void *RvcPitchMeter)
  {
  RVC_PITCH_METER_be *pstMeter;

  pstMeter = (RVC_PITCH_METER_be *)RvcPitchMeter;
  if (pstMeter->signature != RVC_PITCH_METER_SIG)
    return RVC_ERR_ILLEGAL_ARGUMENT;

  Deallocate_InterpolatedDft_be(pstMeter->pstSingleInterpDft);
  Deallocate_InterpolatedDft_be(pstMeter->pstDoubleInterpDft);

  MDEALLOCATE(pstMeter->pstPitchCand);

  MDEALLOCATE(pstMeter);
  return RVC_OK;
  }


/*
***********************************************************************
*
*   FUNCTION NAME: RVC_DestructPitchRom_be
*
*   PURPOSE: To de-allocate memory assigned to RVC_PITCH_ROM_be structure
*
*   INPUTS:
*
*      *RvcPitchRom - RVC_PITCH_ROM_be structure
*
*   OUTPUTS:
*
*      None - the memory assigned to the structure is de-allocated
*
*   RETURN VALUE:
*
*      RVC_OK - Success
*      RVC_ERR_ILLEGAL_ARGUMENT - RvcPitchRom does not point to
*                                 an object of expected type
*
**********************************************************************
*/

X_INT16 RVC_DestructPitchRom_be(void *RvcPitchRom)
  {
  RVC_PITCH_ROM_be *pstRom;

  pstRom = (RVC_PITCH_ROM_be *)RvcPitchRom;
  if (pstRom->signature != RVC_PITCH_ROM_SIG)
    return RVC_ERR_ILLEGAL_ARGUMENT;

  MDEALLOCATE(pstRom);

  return RVC_OK;
  }


