/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include <stdio.h>
#include <stdlib.h>

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#include "ParmInterface_B.h"
#include "PostProcExports_B.h"


#define NB_TOPOSTPROC   (NUM_CEP_COEFF-1)
#define INDEX_ENER      (NUM_CEP_COEFF)

struct PostProcStructX {
    X_INT32 weightLMS[12] ;
} ;


/* target values for blind equalization */
static const X_INT16 targetLMS16[NB_TOPOSTPROC]  = {
    ROUND16 ( PP_FACT_COEF  * -6.618909) ,
        ROUND16 ( PP_FACT_COEF  *  0.198269 ),
        ROUND16 ( PP_FACT_COEF  * -0.740308 ),
        ROUND16 ( PP_FACT_COEF  *  0.055132 ),
        ROUND16 ( PP_FACT_COEF  * -0.227086 ),
        ROUND16 ( PP_FACT_COEF  *  0.144280 ),
        ROUND16 ( PP_FACT_COEF  * -0.112451 ),
        ROUND16 ( PP_FACT_COEF  * -0.146940 ),
        ROUND16 ( PP_FACT_COEF  * -0.327466 ),
        ROUND16 ( PP_FACT_COEF  *  0.134571 ),
        ROUND16 ( PP_FACT_COEF  *  0.027884 ),
        ROUND16 ( PP_FACT_COEF  * -0.114905 )
} ;

#define MIN_ENER_POSTPROC   (211)
#define DELTA_ENER_POSTPROC (64)


#define LAMBDA 0.0087890625

#define LAMBDA2xx16 ROUND16(65536.0 * LAMBDA)

#define LAMBDA_FACT ROUND16(32768.0 * LAMBDA * 4096 / PP_FACT_ENER)





/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoPostProcAlloc_B
 *
 * PURPOSE:  Allocation of a postprocessing struct
 *
 * INPUT: 
 *
 * OUTPUT:
 *
 * RETURN VALUE: pointer to the new allocated struct
 *   
 *              
 *
 *---------------------------------------------------------------------------*/

extern PostProcStructX * DoPostProcAlloc_B(void) {
    PostProcStructX * newPost = calloc(1,sizeof(PostProcStructX)) ;
    return newPost;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoPostProcInit_B
 *
 * PURPOSE:  Initialisation of a postprocessing struct
 *
 * INPUT:
 *    - This : global FEParamsX_F context (for this->PPX)
 *
 * OUTPUT:
 *    - initialized this->PPX
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/

extern void DoPostProcInit_B (FEParamsX_F *This) {
    PostProcStructX *PPX = This->PPX;
    int i ;
    for (i=0 ; i<NB_TOPOSTPROC ; i++) { 
      PPX->weightLMS[i] = 0 ;  move32();
    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoPostProc_B
 *
 * PURPOSE: Inplace Cepstral post-processing for a frame 
 *              
 *
 * INPUT:
 *      - Coef16  :   Cepstral ceofficient frame
 *      - This  : global FEParamsX_F context (for this->PPX)                
 *
 * OUTPUT:
 *      - Coef16  :   postprocessed Cepstral ceofficient frame
 *
 * RETURN VALUE: TRUE
 *   
 *
 *---------------------------------------------------------------------------*/

extern BOOLEAN DoPostProc_B (
        X_INT16 *Coef16         /* (i/o) coefficients to postprocess */
        , FEParamsX_F *This
    ) 

{
    PostProcStructX *PPX = This->PPX;
    int i ;
    X_INT16 iwe16 ;
    X_INT32 *weightLMS32 = PPX->weightLMS ;
#define THRESHx16    (MIN_ENER_POSTPROC * 16)

    {
        X_INT16 ener16 = shl(Coef16[INDEX_ENER],4) ; move16();
        test() ;
        if (sub(ener16,THRESHx16) <= 0)  {
            iwe16 = 0 ; move16();
        } else {
            test();
            if (sub(ener16,add(THRESHx16,DELTA_ENER_POSTPROC*16)) > 0 ) {
                iwe16 = LAMBDA2xx16  ;  move16();
            } else {
                X_INT32 val = L_msu(L_mult(ener16,LAMBDA_FACT),THRESHx16,LAMBDA_FACT);
                iwe16 = round(val) ;
            }
        }
    }
    for (i=0 ; i<NB_TOPOSTPROC ; i++) {
        X_INT16 dif16,w16 ;
        w16 = shr_r(extract_h(weightLMS32[i]),1);
        dif16 = sub(Coef16[i],add(w16,targetLMS16[i]));
        Coef16[i] = sub(Coef16[i],w16); move16();
        weightLMS32[i] = L_mac(weightLMS32[i],dif16,iwe16); move32();
    }
    return TRUE ;
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoPostProcDelete_B
 *
 * PURPOSE:  frees a  postprocessing struct
 *
 * INPUT:
 *    - PPX : postprocessing struct
 *
 * OUTPUT:
 *
 * RETURN VALUE: none
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
extern void DoPostProcDelete_B(PostProcStructX *PPX) {
    if (PPX != NULL) {
        free(PPX);
    }
}
