/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*-----------------
 * File Inclusions
 *-----------------*/
#include <stdlib.h>
#include <stdio.h>

#include "ParmInterface_B.h"
#include "VADExports_F.h"

#include "typedef.h"
#include "basic_op.h"

#include "count.h"

/*----------------
* VAD Definitions
*-----------------*/
typedef struct VADStructX_F VADStructX_F ;
struct VADStructX_F
{
  X_INT16 Focus;               //Position of circular buffer
  X_INT16 HangOver;            //Hangover length
  X_INT16 FlushFocus;          //Position in circular buffer when emptying at end (uses Focus as start/endpoint)
  X_INT16 H_CountDown;         //Main hangover countdown
  X_INT16 V_CountDown;         //Short hangover countdown
  X_INT16 **OutBuffer;       //
};

#define PRINTVAD    0
#define PRINTNULLS  0
#define BUFFER_SIZE 7    // Number of frames in analysis buffer
#define HANGOVERV   23     // Number of frames by which VAD will stay on after speech
#define H_TRIGGER   4      // Number of speech frames required to trigger hangover NB Must be <= BUFFER_SIZE!!
#define VANGOVER    5      // Number of frames for short hangover, triggered by 3 frames of speech
#define V_TRIGGER   3      // Number of speech frames required to trigger VAD
#define SANGOVER   50      // Number of frames for safety hangover


/*----------------------------------------------------------------------------
 * FUNCTION NAME: focalpoint
 *
 * PURPOSE: CONTROL VAD BUFFER
 *
 * INPUT:
 *  Focus   current focal point
 *      offset    change in focal point
 *
 * OUTPUT:
 *  None
 *
 *
 * RETURN VALUE:
 *  temp    new focal point within Buffer boundary
 *
 *
 *---------------------------------------------------------------------------*/
static X_INT16 focalpoint (X_INT16 Focus, X_INT16 offset)
{
  X_INT16 temp;
  
  
  temp = add(Focus, offset);
  test();
  if (sub(temp,BUFFER_SIZE-1) > 0) 
  {
      temp = sub(temp, BUFFER_SIZE);
  }
  test();
  if (temp < 0) 
  {
      temp = add(temp, BUFFER_SIZE);
  }


  return (temp);
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoVADAlloc
 *
 * PURPOSE: Allocation of memory to VAD structure
 *
 *
 * INPUT:
 *  None
 *
 *
 * OUTPUT:
 *  None
 *
 *
 * RETURN VALUE:
 *  VADX    Pointer to VADStructX
 *
 *
 *---------------------------------------------------------------------------*/
extern struct VADStructX_F *DoVADAlloc_F (void)
{
  VADStructX_F *VADX = calloc (1, sizeof (VADStructX_F));
  return VADX;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoVADInit
 *
 * PURPOSE: Initialises VAD structure
 *
 *
 * INPUT:
 *  This    Pointer to front end parameter structure
 *
 *
 * OUTPUT:
 *  This    Pointer to front end parameter structure
 *
 *
 * RETURN VALUE:
 *  None
 *
 *
 *---------------------------------------------------------------------------*/
extern void DoVADInit_F (FEParamsX_F *This)
{
  X_INT16 i ;
  VADStructX_F *VADX = This->VADX;

  VADX->Focus = 0;
  VADX->HangOver = HANGOVERV;
  VADX->H_CountDown = 0;
  VADX->V_CountDown = 0;
  VADX->FlushFocus = -1;

  VADX->OutBuffer = calloc (BUFFER_SIZE, sizeof (VADX->OutBuffer[0])) ;
  if (VADX->OutBuffer == NULL)
    {
      fprintf (stderr, "Allocation error Outbuffer\n");
      exit(0);
    }
  for (i=0 ; i<BUFFER_SIZE ; i++)
    {
      VADX->OutBuffer[i] = calloc (NUM_CEP_COEFF+2, sizeof (VADX->OutBuffer[0][0])) ;
      if (VADX->OutBuffer[i] == NULL)
        {
          fprintf (stderr, "Allocation error Outbuffer\n");
          exit(0);
        }
    }
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoVADDelete
 *
 * PURPOSE: Free VAD structure
 *
 *
 * INPUT:
 *  VADX    Pointer to VAD structure
 *
 *
 * OUTPUT:
 *  VADX    Pointer to empty VAD structure
 *
 *
 * RETURN VALUE:
 *  None
 *
 *
 *---------------------------------------------------------------------------*/
extern void DoVADDelete_F (VADStructX_F *VADX)
{
  X_INT16 i;

  for (i=0 ; i<BUFFER_SIZE ; i++)
    {
      if (VADX->OutBuffer[i] != NULL) free (VADX->OutBuffer[i]) ;
    }
  
  if (VADX->OutBuffer != NULL) free (VADX->OutBuffer) ;

  if (VADX != NULL)
    free (VADX);
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoVADProc
 *
 * PURPOSE: Identifies speech and non-speech frames
 *
 *
 * INPUT:
 *  FeatureBuffer   Buffer of features
 *  This      Pointer to front end parameter structure
 *
 *
 * OUTPUT:
 *  FeatureBuffer   Buffer of features
 *  This      Pointer to front end parameter structure
 *
 *
 * RETURN VALUE:
 *  TRUE    Once buffer is full
 *  FALSE   For first 10 frames
 *
 *
 *---------------------------------------------------------------------------*/
extern BOOLEAN DoVADProc_F (X_INT16 *FeatureBuffer, FEParamsX_F *This)
{
  VADStructX_F *VADX = This->VADX;

  X_INT16 i;
  X_INT16 Sum;
  X_INT16 Trigger;
  X_INT16 Refocus;

  X_INT16 Focus = VADX->Focus;
  X_INT16 HangOver = VADX->HangOver;
  X_INT16 H_CountDown = VADX->H_CountDown;
  X_INT16 V_CountDown = VADX->V_CountDown;

  X_INT16 **OutBuffer = VADX->OutBuffer;
  X_INT16  currentVAD ;

  move16(); move16(); move16(); move16(); 

  Focus = add(Focus, 1);

  test();
  if (sub(Focus, BUFFER_SIZE) == 0) 
  {
      Focus = 0; move16();
  }

  for (i=0 ; i<NUM_CEP_COEFF+1 ; i++)
  {
      OutBuffer[Focus][i] = FeatureBuffer[i]; move16();
  }

  // VAD Now 3 frames ahead!
  // There is a 3 frame latency between the measures used by the VAD
  // and the feature frame available
  test(); test(); test(); test();
  if (This->SpeechFoundSpec || This->SpeechFoundMel || This->SpeechFoundVar || This->SpeechFoundVADNS)
  {
      OutBuffer[Focus][NUM_CEP_COEFF+1] = 1; move16();
  }
  else
  {
      OutBuffer[Focus][NUM_CEP_COEFF+1] = 0; move16();
  }

  //+3 due to changes in lookahead
  test();
  if (L_sub(This->FrameCounter, (BUFFER_SIZE+3)) > 0)
  {
      Sum=0;Trigger=0; move16(); move16(); 
      for (i=0 ; i<BUFFER_SIZE ; i++) 
        {
          Refocus = focalpoint (Focus, i+1);
          test();
          if (OutBuffer[Refocus][NUM_CEP_COEFF+1])
            {
                Sum = add(Sum, 1);
            }
          else
            {
                test();
                if (sub(Sum, Trigger) > 0)
                {
                    Trigger = Sum; move16();
                }
                Sum = 0; move16();
            }
        }
      if (sub(Sum, Trigger) > 0)
        {
            Trigger = Sum; move16();
        }
      if (sub(Trigger, H_TRIGGER) >= 0)
        {
            H_CountDown = HangOver; move16();
            test();
            if (L_sub(This->FrameCounter, 35) <= 0) 
                {
                    HangOver = SANGOVER; move16();
                }

        }
      test(); test();
      if (H_CountDown && (sub(Trigger, V_TRIGGER) < 0))
      {
          H_CountDown = sub(H_CountDown, 1);
      }
      test();
      if (sub(Trigger, V_TRIGGER) >= 0)
      {
          V_CountDown = VANGOVER; move16();
      }
      test(); test();
      if (V_CountDown && (sub(Trigger, V_TRIGGER) < 0))
      {
          V_CountDown = sub(V_CountDown, 1);
      }           
      Refocus = focalpoint (Focus, 1);

      for (i=0 ; i<NUM_CEP_COEFF+1 ; i++) 
      {
          FeatureBuffer[i] = OutBuffer[Refocus][i]; move16();
      }
      currentVAD = OutBuffer[Refocus][NUM_CEP_COEFF+1]; move16();


      test(); test(); test(); 
      if (V_CountDown || H_CountDown || (sub(Trigger, V_TRIGGER) >= 0))
      {
          currentVAD = 1; move16();
      }
      else
      {
          currentVAD = 0 ; move16();
          if (PRINTNULLS) fprintf (stderr, "%d,", This->FrameCounter-(BUFFER_SIZE+3));
      }
      This->VAD = currentVAD; move16();
      VADX->Focus = Focus; move16();
      VADX->HangOver = HangOver; move16();
      VADX->H_CountDown = H_CountDown; move16();
      VADX->V_CountDown = V_CountDown; move16();

      return TRUE;
    }
  else 
    {
      VADX->Focus = Focus; move16();
      VADX->HangOver = HangOver; move16();
      VADX->H_CountDown = H_CountDown; move16();
      VADX->V_CountDown = V_CountDown; move16();

      return FALSE;
    }  
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoVADFlush
 *
 * PURPOSE: Flush remaining data from VAD Buffer
 *
 *
 * INPUT:
 *	FeatureBuffer		Buffer of features
 *	This			Pointer to front end parameter structure
 *
 *
 * OUTPUT:
 *	FeatureBuffer		Buffer of features
 *	This			Pointer to front end parameter structure
 *
 *
 * RETURN VALUE:
 *	TRUE		When VAD buffer is being flushed
 *	FALSE		When flushing is complete
 *
 *
 *---------------------------------------------------------------------------*/
extern BOOLEAN DoVADFlush_F (X_INT16 *FeatureBuffer, FEParamsX_F *This)
{
  VADStructX_F *VADX = This->VADX;

  X_INT16 i;
  X_INT16 Sum;
  X_INT16 Trigger;
  X_INT16 Refocus;

  X_INT16 Focus = VADX->Focus;
  X_INT16 HangOver = VADX->HangOver;
  X_INT16 H_CountDown = VADX->H_CountDown;
  X_INT16 V_CountDown = VADX->V_CountDown;

  X_INT16 **OutBuffer = VADX->OutBuffer;
  X_INT16 currentVAD; 
  move16();move16();move16();move16();

  // Last frame output from loop
  test(); sub(VADX->FlushFocus,-1); move16();
  if (VADX->FlushFocus == -1) VADX->FlushFocus = Focus;

  // Move forward one step
  Focus = add(Focus, 1);


  test();
  if (sub(Focus, BUFFER_SIZE) == 0) 
  {
      Focus = 0; move16();
  }

  test();
  if (sub(Focus, VADX->FlushFocus) != 0)
    {                           
        // Flush remaining data
        // Need to keep the counter incrementing!
        This->FrameCounter = L_add(This->FrameCounter, 1); move16();
      
        // +3 due to changes in lookahead
        test();
        if (L_sub(This->FrameCounter, (BUFFER_SIZE+3) > 0))
        {
            Sum = 0; move16();
            Trigger = 0; move16();
            for (i=0 ; i<BUFFER_SIZE ; i++)
            {
                Refocus = focalpoint (Focus, i+1);
                test();
                if (OutBuffer[Refocus][NUM_CEP_COEFF+1])
                {
                    Sum = add(Sum, 1);
                }
                else
                {
                    test();
                    if (sub(Sum, Trigger) > 0)
                    {
                        Trigger = Sum; move16();
                    }
                    Sum = 0; move16();
                }
            }
  
            test();
            if (sub(Sum, Trigger) > 0)
            {
                Trigger = Sum; move16();
            }
            test();
            if (sub(Trigger, H_TRIGGER) >= 0)
            {
                H_CountDown=HangOver; move16();
                test();
                if (L_sub(This->FrameCounter, 35) <= 0) 
                {
                    HangOver = SANGOVER; move16();
                }
            }
            test(); test();
            if (H_CountDown && (sub(Trigger, V_TRIGGER) < 0))
            {
                H_CountDown = sub(H_CountDown,1);
            }
            test();
            if (sub(Trigger, V_TRIGGER) >= 0)
            {
                V_CountDown = VANGOVER; move16();
            }
            test(); test() ;
            if(V_CountDown && (sub(Trigger, V_TRIGGER) < 0))
            {
                V_CountDown = sub(V_CountDown, 1);
            }           
          Refocus = focalpoint (Focus, 1);

            for (i=0 ; i<NUM_CEP_COEFF+1 ; i++) 
            {
                FeatureBuffer[i] = OutBuffer[Refocus][i]; move16();
            }
            currentVAD = OutBuffer[Refocus][NUM_CEP_COEFF+1]; move16();
            test(); test(); test();
            if (V_CountDown || H_CountDown || (sub(Trigger, V_TRIGGER) >= 0))
            {
                currentVAD = 1; move16();
            }
            else
            {
                currentVAD = 0 ; move16();
                if (PRINTNULLS) fprintf (stderr, "%d,", This->FrameCounter-(BUFFER_SIZE+3));
            }
            /*------------------
             * Output result
             *------------------*/
            This->VAD = currentVAD; move16();
        }
        VADX->Focus = Focus; move16();
        VADX->HangOver = HangOver; move16();
        VADX->H_CountDown = H_CountDown; move16();
        VADX->V_CountDown = V_CountDown; move16();
        
        return TRUE;
    } 
  else 
    {
        return FALSE;
    }
}

