/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

/*---------------------------------------------------------------------------
* PURPOSE: 1) Apply 2-stage Wiener filter on the input frame.
*          2) Apply DC offset removal on the output of 2-stage 
*             Wiener filter.
*          3) Calculate parameters for the frame dropping VAD (see 
*             SpeechQSpec(), SpeechQMel(), SpeechQVar()).
*          4)  Incorporates pitch and class extraction
*---------------------------------------------------------------------------*/

/*-----------------
* File Inclusions
*-----------------*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "ParmInterface_B.h"
#include "NoiseSup_F.h"
#include "NoiseSupExports_B.h"
#include "ff4nrFix32_B.h"
#include "MathFunc.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#if defined(F_16KHZ)
#   include "16kHzProcExports_B.h"
#endif



/*
 * For XAFE
 */
#if defined(XAFE)
#	include "ParmType.h"
#	include "CompCepsExports_B.h"
#	include "rvc_pitch_B.h"
#	include "rvc_pitch_api_B.h"
#	include "classifyFrame_B.h"
#	include "dsrAfeVad_B.h"
#	include "preProc_B.h"
#endif


#define WF_MEL_ORDER        25

/*------------------------
* Definitions and Macros
*------------------------*/
#if !defined(max)
#define max(a,b) ((a>b)?(a):(b))
#endif

#define SHFT_NOISE 14
#define RND_SHFT_NOISE ( 1 << (SHFT_NOISE -1))

#define MULT_LOW_LOW(A,B)  L_mult((A), extract_l((B)))
#define MULT_LOW_HIGH(A,B) L_shl( \
                           L_mult( (A), extract_h(L_shl(B,1)) ) \
                                          , 15) 
#define L_mult_SL(A,B)  L_add(MULT_LOW_LOW(A,B)  ,  MULT_LOW_HIGH(A,B))


typedef struct vad_data_ns_F VAD_DATA_NS_F;
typedef struct vad_data_fd VAD_DATA_FD;
typedef struct vad_data_ca VAD_DATA_CA;
typedef struct ns_var_F      NS_VAR_F;
typedef struct ns_tmp_F      NS_TMP_F;
typedef struct dc_filter   DC_FILTER;
typedef struct dc_filter_F   DC_FILTER_F;
typedef struct gain_fact_F   GAIN_FACT_F;
typedef struct buffers     BUFFERS;
typedef struct spectrum_F    SPECTRUM_F;


struct dc_filter_F
{
    X_INT32 lastSampleIn32;                  // last input sample of DC offset compensation
    X_INT32 lastDCOut32;                     // last output sample of DC offset compensation
    X_INT16 oldShift ;                       // shift factor
} ;

struct vad_data_ns_F
{
    X_INT16   nbFrame [2];                   // frame number
    X_INT16   flagVAD;                       // VAD flag (1 == SPEECH, 0 == NON SPEECH)
    X_INT16   hangOver;                      // hangover
    X_INT16   nbSpeechFrames;                // nb speech frames (used to set hangover)
    X_INT32   meanEn32;                      // mean energy
};

struct vad_data_ca
{
    X_INT16   flagVAD;                       // VAD flag (1 == SPEECH, 0 == NON SPEECH)
    X_INT16   hangOver;                      // hangover
    X_INT16   nbSpeechFrames;                // nb speech frames (used to set hangover)
    X_INT32   meanCA;                        // mean energy

};

struct vad_data_fd
{
  X_INT16 MelMean;                 // SpeechQMel
  X_INT32 VarMean;                 // SpeechQVar  
  X_INT32 AccTest;                 // SpeechQSpec
  X_INT32 AccTest2;                //   dito
  X_INT32 SpecMean;                //   dito     
  X_INT16 MelValues[2];            // SpeechQMel
  X_INT32 SpecValues;              // SpeechQSpec
  X_INT16 SpeechInVADQ;            // Flag
  X_INT16 SpeechInVADQ2;           // Flag
};

struct gain_fact_F
{
    X_INT32   logDenEn1_32[3] ;
    X_INT32   lowSNRtrack32 ;
    X_INT16   alfaGF16 ;
};

struct buffers
{
    X_INT32 nbFramesInFirstStage;            // nb frames in first stage
    X_INT32 nbFramesInSecondStage;           // nb frames in second stage
    X_INT32 nbFramesOutSecondStage;          // nb frames out of second stage
    
    X_INT16 FirstStageIn16Buffer  [NS_BUFFER_SIZE]; // first stage buffer
    X_INT32 SecondStageInBuffer32 [NS_BUFFER_SIZE]; // second stage buffer
    X_INT16 SecondDecalSig[(NS_BUFFER_SIZE+2*NS_FRAME_SHIFT-1)/NS_FRAME_SHIFT] ;
};

struct spectrum_F
{
    X_INT32   noiseSE1_32 [NS_SPEC_ORDER];   
    X_INT16   noiseSE1_dec [NS_SPEC_ORDER];   
    X_INT32   noiseSE2_32 [NS_SPEC_ORDER];   
    X_INT16   noiseSE2_dec [NS_SPEC_ORDER];   
    
    X_INT16   indexBuffer1;                  // where to enter new PSD, alternatively 0 and 1
    X_INT16   indexBuffer2;                  // where to enter new PSD, alternatively 0 and 1
    X_INT32   PSDMeanAntBuffer1 [NS_SPEC_ORDER]; // 1st stage PSD Mean buffer
    X_INT32   PSDMeanAntBuffer2 [NS_SPEC_ORDER]; // 1st stage PSD Mean buffer
    X_INT32   denSigSE1_32[NS_SPEC_ORDER];
    X_INT32   denSigSE2_32[NS_SPEC_ORDER];
    X_INT16   nSigSE1Cur_dec[NS_SPEC_ORDER] ;
    X_INT16   nSigSE1Ant_dec[NS_SPEC_ORDER] ;
    X_INT16   nSigSE2Cur_dec[NS_SPEC_ORDER] ;
    X_INT16   nSigSE2Ant_dec[NS_SPEC_ORDER] ;
    X_INT16   denSigSE1_dec[NS_SPEC_ORDER] ;
    X_INT16   denSigSE2_dec[NS_SPEC_ORDER] ;

};

struct ns_var_F
{
    X_INT16     SampFreq;                    // sampling frequency
    BUFFERS     buffers;                     // signal buffers
    DC_FILTER_F  prevSamples32;                 // previous samples for DC offset removal
    GAIN_FACT_F   gainFact;                    // gain factorization variables
    VAD_DATA_NS_F vadNS;                       // VAD for noise suppression data
    VAD_DATA_FD vadFD;                       // VAD for frame dropping data
    VAD_DATA_CA vadCA;                       // VAD for frame dropping data
    SPECTRUM_F    spectrum;                    // spectrum data
#if defined(F_16KHZ)
    BOOLEAN     Do16kHzProc;                 // do 16 kHz processing ?
#endif
};

typedef struct NoiseSupStructX_F
{
    NS_VAR_F    nsVar;                         // non sharable data
} NoiseSupStructX_F ;

/*------------
* Prototypes
*------------*/
static void DCOffsetFil_F (X_INT32 *Data32, DC_FILTER_F *prevSamples, X_INT16 DataLength,X_INT16 newShift) ;
static void DoSigWindowing16_F1(X_INT16 * DataIn16,X_INT32 * DataOut32,X_INT16 decal,X_INT16 FFTLength) ;
static void DoSigWindowing16_F2(X_INT16 * DataIn16,X_INT32 * DataOut32,X_INT16 FFTLength) ;

static void FFTtoPSD_F (X_INT32 *FFTIn32, X_INT32  *PSDOut32,X_INT16 *nSigSE_dec,X_INT16 dec0);
static void PSDMean_F(X_INT32 * PSDIn32,X_INT16 *PSDIn_dec,X_INT32 * PSDOut32,X_INT16 *PSDOut_dec,X_INT32 *PSDAnt32,X_INT16 *PSDAnt_dec) ;
static X_INT16 SpeechQVar (FEParamsX_F *This, X_INT16 *W);
static X_INT16 SpeechQMel (FEParamsX_F *This);
static X_INT16 SpeechQSpec (FEParamsX_F *This);


// scaled with 2^15 !!   
// 1/n for n = 1..20  (1 value is scaled with 2^15-1 !!)
static const X_INT16 lambda_1divX[20] = {32767, 16384, 10923, 8192, 6554, \
                                   5461, 4681,   4096, 3641, 3277, \
                                   2979, 2731,   2521, 2341, 2185, \
                                   2048, 1928, 1820, 1725, 1638};

static const X_INT16 Hann_sh32_hi[100] = {
                 2       /* 0.50 - 0.50 * cos (PI * 2 * (0 + 0.5) / (200)*/
        ,       18       /* 0.50 - 0.50 * cos (PI * 2 * (1 + 0.5) / (200)*/
        ,       51       /* 0.50 - 0.50 * cos (PI * 2 * (2 + 0.5) / (200)*/
        ,       99       /* 0.50 - 0.50 * cos (PI * 2 * (3 + 0.5) / (200)*/
        ,      163       /* 0.50 - 0.50 * cos (PI * 2 * (4 + 0.5) / (200)*/
        ,      244       /* 0.50 - 0.50 * cos (PI * 2 * (5 + 0.5) / (200)*/
        ,      340       /* 0.50 - 0.50 * cos (PI * 2 * (6 + 0.5) / (200)*/
        ,      453       /* 0.50 - 0.50 * cos (PI * 2 * (7 + 0.5) / (200)*/
        ,      581       /* 0.50 - 0.50 * cos (PI * 2 * (8 + 0.5) / (200)*/
        ,      724       /* 0.50 - 0.50 * cos (PI * 2 * (9 + 0.5) / (200)*/
        ,      883       /* 0.50 - 0.50 * cos (PI * 2 * (10 + 0.5) / (200)*/
        ,     1058       /* 0.50 - 0.50 * cos (PI * 2 * (11 + 0.5) / (200)*/
        ,     1247       /* 0.50 - 0.50 * cos (PI * 2 * (12 + 0.5) / (200)*/
        ,     1452       /* 0.50 - 0.50 * cos (PI * 2 * (13 + 0.5) / (200)*/
        ,     1671       /* 0.50 - 0.50 * cos (PI * 2 * (14 + 0.5) / (200)*/
        ,     1904       /* 0.50 - 0.50 * cos (PI * 2 * (15 + 0.5) / (200)*/
        ,     2152       /* 0.50 - 0.50 * cos (PI * 2 * (16 + 0.5) / (200)*/
        ,     2414       /* 0.50 - 0.50 * cos (PI * 2 * (17 + 0.5) / (200)*/
        ,     2690       /* 0.50 - 0.50 * cos (PI * 2 * (18 + 0.5) / (200)*/
        ,     2979       /* 0.50 - 0.50 * cos (PI * 2 * (19 + 0.5) / (200)*/
        ,     3282       /* 0.50 - 0.50 * cos (PI * 2 * (20 + 0.5) / (200)*/
        ,     3597       /* 0.50 - 0.50 * cos (PI * 2 * (21 + 0.5) / (200)*/
        ,     3926       /* 0.50 - 0.50 * cos (PI * 2 * (22 + 0.5) / (200)*/
        ,     4266       /* 0.50 - 0.50 * cos (PI * 2 * (23 + 0.5) / (200)*/
        ,     4618       /* 0.50 - 0.50 * cos (PI * 2 * (24 + 0.5) / (200)*/
        ,     4982       /* 0.50 - 0.50 * cos (PI * 2 * (25 + 0.5) / (200)*/
        ,     5357       /* 0.50 - 0.50 * cos (PI * 2 * (26 + 0.5) / (200)*/
        ,     5743       /* 0.50 - 0.50 * cos (PI * 2 * (27 + 0.5) / (200)*/
        ,     6140       /* 0.50 - 0.50 * cos (PI * 2 * (28 + 0.5) / (200)*/
        ,     6547       /* 0.50 - 0.50 * cos (PI * 2 * (29 + 0.5) / (200)*/
        ,     6963       /* 0.50 - 0.50 * cos (PI * 2 * (30 + 0.5) / (200)*/
        ,     7389       /* 0.50 - 0.50 * cos (PI * 2 * (31 + 0.5) / (200)*/
        ,     7823       /* 0.50 - 0.50 * cos (PI * 2 * (32 + 0.5) / (200)*/
        ,     8266       /* 0.50 - 0.50 * cos (PI * 2 * (33 + 0.5) / (200)*/
        ,     8717       /* 0.50 - 0.50 * cos (PI * 2 * (34 + 0.5) / (200)*/
        ,     9176       /* 0.50 - 0.50 * cos (PI * 2 * (35 + 0.5) / (200)*/
        ,     9642       /* 0.50 - 0.50 * cos (PI * 2 * (36 + 0.5) / (200)*/
        ,    10114       /* 0.50 - 0.50 * cos (PI * 2 * (37 + 0.5) / (200)*/
        ,    10593       /* 0.50 - 0.50 * cos (PI * 2 * (38 + 0.5) / (200)*/
        ,    11077       /* 0.50 - 0.50 * cos (PI * 2 * (39 + 0.5) / (200)*/
        ,    11566       /* 0.50 - 0.50 * cos (PI * 2 * (40 + 0.5) / (200)*/
        ,    12061       /* 0.50 - 0.50 * cos (PI * 2 * (41 + 0.5) / (200)*/
        ,    12559       /* 0.50 - 0.50 * cos (PI * 2 * (42 + 0.5) / (200)*/
        ,    13062       /* 0.50 - 0.50 * cos (PI * 2 * (43 + 0.5) / (200)*/
        ,    13567       /* 0.50 - 0.50 * cos (PI * 2 * (44 + 0.5) / (200)*/
        ,    14075       /* 0.50 - 0.50 * cos (PI * 2 * (45 + 0.5) / (200)*/
        ,    14586       /* 0.50 - 0.50 * cos (PI * 2 * (46 + 0.5) / (200)*/
        ,    15099       /* 0.50 - 0.50 * cos (PI * 2 * (47 + 0.5) / (200)*/
        ,    15612       /* 0.50 - 0.50 * cos (PI * 2 * (48 + 0.5) / (200)*/
        ,    16127       /* 0.50 - 0.50 * cos (PI * 2 * (49 + 0.5) / (200)*/
        ,    16641       /* 0.50 - 0.50 * cos (PI * 2 * (50 + 0.5) / (200)*/
        ,    17156       /* 0.50 - 0.50 * cos (PI * 2 * (51 + 0.5) / (200)*/
        ,    17669       /* 0.50 - 0.50 * cos (PI * 2 * (52 + 0.5) / (200)*/
        ,    18182       /* 0.50 - 0.50 * cos (PI * 2 * (53 + 0.5) / (200)*/
        ,    18693       /* 0.50 - 0.50 * cos (PI * 2 * (54 + 0.5) / (200)*/
        ,    19201       /* 0.50 - 0.50 * cos (PI * 2 * (55 + 0.5) / (200)*/
        ,    19706       /* 0.50 - 0.50 * cos (PI * 2 * (56 + 0.5) / (200)*/
        ,    20209       /* 0.50 - 0.50 * cos (PI * 2 * (57 + 0.5) / (200)*/
        ,    20707       /* 0.50 - 0.50 * cos (PI * 2 * (58 + 0.5) / (200)*/
        ,    21202       /* 0.50 - 0.50 * cos (PI * 2 * (59 + 0.5) / (200)*/
        ,    21691       /* 0.50 - 0.50 * cos (PI * 2 * (60 + 0.5) / (200)*/
        ,    22175       /* 0.50 - 0.50 * cos (PI * 2 * (61 + 0.5) / (200)*/
        ,    22654       /* 0.50 - 0.50 * cos (PI * 2 * (62 + 0.5) / (200)*/
        ,    23126       /* 0.50 - 0.50 * cos (PI * 2 * (63 + 0.5) / (200)*/
        ,    23592       /* 0.50 - 0.50 * cos (PI * 2 * (64 + 0.5) / (200)*/
        ,    24051       /* 0.50 - 0.50 * cos (PI * 2 * (65 + 0.5) / (200)*/
        ,    24502       /* 0.50 - 0.50 * cos (PI * 2 * (66 + 0.5) / (200)*/
        ,    24945       /* 0.50 - 0.50 * cos (PI * 2 * (67 + 0.5) / (200)*/
        ,    25379       /* 0.50 - 0.50 * cos (PI * 2 * (68 + 0.5) / (200)*/
        ,    25805       /* 0.50 - 0.50 * cos (PI * 2 * (69 + 0.5) / (200)*/
        ,    26221       /* 0.50 - 0.50 * cos (PI * 2 * (70 + 0.5) / (200)*/
        ,    26628       /* 0.50 - 0.50 * cos (PI * 2 * (71 + 0.5) / (200)*/
        ,    27025       /* 0.50 - 0.50 * cos (PI * 2 * (72 + 0.5) / (200)*/
        ,    27411       /* 0.50 - 0.50 * cos (PI * 2 * (73 + 0.5) / (200)*/
        ,    27786       /* 0.50 - 0.50 * cos (PI * 2 * (74 + 0.5) / (200)*/
        ,    28150       /* 0.50 - 0.50 * cos (PI * 2 * (75 + 0.5) / (200)*/
        ,    28502       /* 0.50 - 0.50 * cos (PI * 2 * (76 + 0.5) / (200)*/
        ,    28842       /* 0.50 - 0.50 * cos (PI * 2 * (77 + 0.5) / (200)*/
        ,    29171       /* 0.50 - 0.50 * cos (PI * 2 * (78 + 0.5) / (200)*/
        ,    29486       /* 0.50 - 0.50 * cos (PI * 2 * (79 + 0.5) / (200)*/
        ,    29789       /* 0.50 - 0.50 * cos (PI * 2 * (80 + 0.5) / (200)*/
        ,    30078       /* 0.50 - 0.50 * cos (PI * 2 * (81 + 0.5) / (200)*/
        ,    30354       /* 0.50 - 0.50 * cos (PI * 2 * (82 + 0.5) / (200)*/
        ,    30616       /* 0.50 - 0.50 * cos (PI * 2 * (83 + 0.5) / (200)*/
        ,    30864       /* 0.50 - 0.50 * cos (PI * 2 * (84 + 0.5) / (200)*/
        ,    31097       /* 0.50 - 0.50 * cos (PI * 2 * (85 + 0.5) / (200)*/
        ,    31316       /* 0.50 - 0.50 * cos (PI * 2 * (86 + 0.5) / (200)*/
        ,    31521       /* 0.50 - 0.50 * cos (PI * 2 * (87 + 0.5) / (200)*/
        ,    31710       /* 0.50 - 0.50 * cos (PI * 2 * (88 + 0.5) / (200)*/
        ,    31885       /* 0.50 - 0.50 * cos (PI * 2 * (89 + 0.5) / (200)*/
        ,    32044       /* 0.50 - 0.50 * cos (PI * 2 * (90 + 0.5) / (200)*/
        ,    32187       /* 0.50 - 0.50 * cos (PI * 2 * (91 + 0.5) / (200)*/
        ,    32315       /* 0.50 - 0.50 * cos (PI * 2 * (92 + 0.5) / (200)*/
        ,    32428       /* 0.50 - 0.50 * cos (PI * 2 * (93 + 0.5) / (200)*/
        ,    32524       /* 0.50 - 0.50 * cos (PI * 2 * (94 + 0.5) / (200)*/
        ,    32605       /* 0.50 - 0.50 * cos (PI * 2 * (95 + 0.5) / (200)*/
        ,    32669       /* 0.50 - 0.50 * cos (PI * 2 * (96 + 0.5) / (200)*/
        ,    32717       /* 0.50 - 0.50 * cos (PI * 2 * (97 + 0.5) / (200)*/
        ,    32750       /* 0.50 - 0.50 * cos (PI * 2 * (98 + 0.5) / (200)*/
        ,    32766       /* 0.50 - 0.50 * cos (PI * 2 * (99 + 0.5) / (200)*/
        } ;

static const X_INT16 Hann_sh32_lo[100] = {
              1393       /* 0.50 - 0.50 * cos (PI * 2 * (0 + 0.5) / (200)*/
        ,    12340       /* 0.50 - 0.50 * cos (PI * 2 * (1 + 0.5) / (200)*/
        ,   -32348       /* 0.50 - 0.50 * cos (PI * 2 * (2 + 0.5) / (200)*/
        ,    -3689       /* 0.50 - 0.50 * cos (PI * 2 * (3 + 0.5) / (200)*/
        ,    29648       /* 0.50 - 0.50 * cos (PI * 2 * (4 + 0.5) / (200)*/
        ,    -2045       /* 0.50 - 0.50 * cos (PI * 2 * (5 + 0.5) / (200)*/
        ,    27097       /* 0.50 - 0.50 * cos (PI * 2 * (6 + 0.5) / (200)*/
        ,   -20236       /* 0.50 - 0.50 * cos (PI * 2 * (7 + 0.5) / (200)*/
        ,   -20234       /* 0.50 - 0.50 * cos (PI * 2 * (8 + 0.5) / (200)*/
        ,    18825       /* 0.50 - 0.50 * cos (PI * 2 * (9 + 0.5) / (200)*/
        ,    22117       /* 0.50 - 0.50 * cos (PI * 2 * (10 + 0.5) / (200)*/
        ,   -20644       /* 0.50 - 0.50 * cos (PI * 2 * (11 + 0.5) / (200)*/
        ,    10338       /* 0.50 - 0.50 * cos (PI * 2 * (12 + 0.5) / (200)*/
        ,   -28265       /* 0.50 - 0.50 * cos (PI * 2 * (13 + 0.5) / (200)*/
        ,   -18599       /* 0.50 - 0.50 * cos (PI * 2 * (14 + 0.5) / (200)*/
        ,    25160       /* 0.50 - 0.50 * cos (PI * 2 * (15 + 0.5) / (200)*/
        ,    22365       /* 0.50 - 0.50 * cos (PI * 2 * (16 + 0.5) / (200)*/
        ,    22515       /* 0.50 - 0.50 * cos (PI * 2 * (17 + 0.5) / (200)*/
        ,     8663       /* 0.50 - 0.50 * cos (PI * 2 * (18 + 0.5) / (200)*/
        ,    28510       /* 0.50 - 0.50 * cos (PI * 2 * (19 + 0.5) / (200)*/
        ,    -2192       /* 0.50 - 0.50 * cos (PI * 2 * (20 + 0.5) / (200)*/
        ,    28063       /* 0.50 - 0.50 * cos (PI * 2 * (21 + 0.5) / (200)*/
        ,   -32200       /* 0.50 - 0.50 * cos (PI * 2 * (22 + 0.5) / (200)*/
        ,    -7593       /* 0.50 - 0.50 * cos (PI * 2 * (23 + 0.5) / (200)*/
        ,    14335       /* 0.50 - 0.50 * cos (PI * 2 * (24 + 0.5) / (200)*/
        ,    10796       /* 0.50 - 0.50 * cos (PI * 2 * (25 + 0.5) / (200)*/
        ,    23788       /* 0.50 - 0.50 * cos (PI * 2 * (26 + 0.5) / (200)*/
        ,    29044       /* 0.50 - 0.50 * cos (PI * 2 * (27 + 0.5) / (200)*/
        ,     1594       /* 0.50 - 0.50 * cos (PI * 2 * (28 + 0.5) / (200)*/
        ,   -18676       /* 0.50 - 0.50 * cos (PI * 2 * (29 + 0.5) / (200)*/
        ,     7468       /* 0.50 - 0.50 * cos (PI * 2 * (30 + 0.5) / (200)*/
        ,   -12442       /* 0.50 - 0.50 * cos (PI * 2 * (31 + 0.5) / (200)*/
        ,    25134       /* 0.50 - 0.50 * cos (PI * 2 * (32 + 0.5) / (200)*/
        ,    26554       /* 0.50 - 0.50 * cos (PI * 2 * (33 + 0.5) / (200)*/
        ,    28700       /* 0.50 - 0.50 * cos (PI * 2 * (34 + 0.5) / (200)*/
        ,     2401       /* 0.50 - 0.50 * cos (PI * 2 * (35 + 0.5) / (200)*/
        ,   -16466       /* 0.50 - 0.50 * cos (PI * 2 * (36 + 0.5) / (200)*/
        ,     7513       /* 0.50 - 0.50 * cos (PI * 2 * (37 + 0.5) / (200)*/
        ,   -21747       /* 0.50 - 0.50 * cos (PI * 2 * (38 + 0.5) / (200)*/
        ,    -4127       /* 0.50 - 0.50 * cos (PI * 2 * (39 + 0.5) / (200)*/
        ,    29053       /* 0.50 - 0.50 * cos (PI * 2 * (40 + 0.5) / (200)*/
        ,   -19402       /* 0.50 - 0.50 * cos (PI * 2 * (41 + 0.5) / (200)*/
        ,    15149       /* 0.50 - 0.50 * cos (PI * 2 * (42 + 0.5) / (200)*/
        ,   -30608       /* 0.50 - 0.50 * cos (PI * 2 * (43 + 0.5) / (200)*/
        ,     7446       /* 0.50 - 0.50 * cos (PI * 2 * (44 + 0.5) / (200)*/
        ,    31078       /* 0.50 - 0.50 * cos (PI * 2 * (45 + 0.5) / (200)*/
        ,     7409       /* 0.50 - 0.50 * cos (PI * 2 * (46 + 0.5) / (200)*/
        ,   -31053       /* 0.50 - 0.50 * cos (PI * 2 * (47 + 0.5) / (200)*/
        ,    13626       /* 0.50 - 0.50 * cos (PI * 2 * (48 + 0.5) / (200)*/
        ,   -22852       /* 0.50 - 0.50 * cos (PI * 2 * (49 + 0.5) / (200)*/
        ,    22852       /* 0.50 - 0.50 * cos (PI * 2 * (50 + 0.5) / (200)*/
        ,   -13626       /* 0.50 - 0.50 * cos (PI * 2 * (51 + 0.5) / (200)*/
        ,    31053       /* 0.50 - 0.50 * cos (PI * 2 * (52 + 0.5) / (200)*/
        ,    -7409       /* 0.50 - 0.50 * cos (PI * 2 * (53 + 0.5) / (200)*/
        ,   -31078       /* 0.50 - 0.50 * cos (PI * 2 * (54 + 0.5) / (200)*/
        ,    -7446       /* 0.50 - 0.50 * cos (PI * 2 * (55 + 0.5) / (200)*/
        ,    30608       /* 0.50 - 0.50 * cos (PI * 2 * (56 + 0.5) / (200)*/
        ,   -15149       /* 0.50 - 0.50 * cos (PI * 2 * (57 + 0.5) / (200)*/
        ,    19402       /* 0.50 - 0.50 * cos (PI * 2 * (58 + 0.5) / (200)*/
        ,   -29053       /* 0.50 - 0.50 * cos (PI * 2 * (59 + 0.5) / (200)*/
        ,     4127       /* 0.50 - 0.50 * cos (PI * 2 * (60 + 0.5) / (200)*/
        ,    21747       /* 0.50 - 0.50 * cos (PI * 2 * (61 + 0.5) / (200)*/
        ,    -7513       /* 0.50 - 0.50 * cos (PI * 2 * (62 + 0.5) / (200)*/
        ,    16466       /* 0.50 - 0.50 * cos (PI * 2 * (63 + 0.5) / (200)*/
        ,    -2401       /* 0.50 - 0.50 * cos (PI * 2 * (64 + 0.5) / (200)*/
        ,   -28700       /* 0.50 - 0.50 * cos (PI * 2 * (65 + 0.5) / (200)*/
        ,   -26554       /* 0.50 - 0.50 * cos (PI * 2 * (66 + 0.5) / (200)*/
        ,   -25134       /* 0.50 - 0.50 * cos (PI * 2 * (67 + 0.5) / (200)*/
        ,    12442       /* 0.50 - 0.50 * cos (PI * 2 * (68 + 0.5) / (200)*/
        ,    -7468       /* 0.50 - 0.50 * cos (PI * 2 * (69 + 0.5) / (200)*/
        ,    18676       /* 0.50 - 0.50 * cos (PI * 2 * (70 + 0.5) / (200)*/
        ,    -1594       /* 0.50 - 0.50 * cos (PI * 2 * (71 + 0.5) / (200)*/
        ,   -29044       /* 0.50 - 0.50 * cos (PI * 2 * (72 + 0.5) / (200)*/
        ,   -23788       /* 0.50 - 0.50 * cos (PI * 2 * (73 + 0.5) / (200)*/
        ,   -10796       /* 0.50 - 0.50 * cos (PI * 2 * (74 + 0.5) / (200)*/
        ,   -14335       /* 0.50 - 0.50 * cos (PI * 2 * (75 + 0.5) / (200)*/
        ,     7593       /* 0.50 - 0.50 * cos (PI * 2 * (76 + 0.5) / (200)*/
        ,    32200       /* 0.50 - 0.50 * cos (PI * 2 * (77 + 0.5) / (200)*/
        ,   -28063       /* 0.50 - 0.50 * cos (PI * 2 * (78 + 0.5) / (200)*/
        ,     2192       /* 0.50 - 0.50 * cos (PI * 2 * (79 + 0.5) / (200)*/
        ,   -28510       /* 0.50 - 0.50 * cos (PI * 2 * (80 + 0.5) / (200)*/
        ,    -8663       /* 0.50 - 0.50 * cos (PI * 2 * (81 + 0.5) / (200)*/
        ,   -22515       /* 0.50 - 0.50 * cos (PI * 2 * (82 + 0.5) / (200)*/
        ,   -22365       /* 0.50 - 0.50 * cos (PI * 2 * (83 + 0.5) / (200)*/
        ,   -25160       /* 0.50 - 0.50 * cos (PI * 2 * (84 + 0.5) / (200)*/
        ,    18599       /* 0.50 - 0.50 * cos (PI * 2 * (85 + 0.5) / (200)*/
        ,    28265       /* 0.50 - 0.50 * cos (PI * 2 * (86 + 0.5) / (200)*/
        ,   -10338       /* 0.50 - 0.50 * cos (PI * 2 * (87 + 0.5) / (200)*/
        ,    20644       /* 0.50 - 0.50 * cos (PI * 2 * (88 + 0.5) / (200)*/
        ,   -22117       /* 0.50 - 0.50 * cos (PI * 2 * (89 + 0.5) / (200)*/
        ,   -18825       /* 0.50 - 0.50 * cos (PI * 2 * (90 + 0.5) / (200)*/
        ,    20234       /* 0.50 - 0.50 * cos (PI * 2 * (91 + 0.5) / (200)*/
        ,    20236       /* 0.50 - 0.50 * cos (PI * 2 * (92 + 0.5) / (200)*/
        ,   -27097       /* 0.50 - 0.50 * cos (PI * 2 * (93 + 0.5) / (200)*/
        ,     2045       /* 0.50 - 0.50 * cos (PI * 2 * (94 + 0.5) / (200)*/
        ,   -29648       /* 0.50 - 0.50 * cos (PI * 2 * (95 + 0.5) / (200)*/
        ,     3689       /* 0.50 - 0.50 * cos (PI * 2 * (96 + 0.5) / (200)*/
        ,    32348       /* 0.50 - 0.50 * cos (PI * 2 * (97 + 0.5) / (200)*/
        ,   -12340       /* 0.50 - 0.50 * cos (PI * 2 * (98 + 0.5) / (200)*/
        ,    -1393       /* 0.50 - 0.50 * cos (PI * 2 * (99 + 0.5) / (200)*/
        } ;

static const X_INT16 Hann_sh24_hi[100] = {
                 0       /* 0.50 - 0.50 * cos (PI * 2 * (0 + 0.5) / (200)*/
        ,        0       /* 0.50 - 0.50 * cos (PI * 2 * (1 + 0.5) / (200)*/
        ,        0       /* 0.50 - 0.50 * cos (PI * 2 * (2 + 0.5) / (200)*/
        ,        0       /* 0.50 - 0.50 * cos (PI * 2 * (3 + 0.5) / (200)*/
        ,        1       /* 0.50 - 0.50 * cos (PI * 2 * (4 + 0.5) / (200)*/
        ,        1       /* 0.50 - 0.50 * cos (PI * 2 * (5 + 0.5) / (200)*/
        ,        1       /* 0.50 - 0.50 * cos (PI * 2 * (6 + 0.5) / (200)*/
        ,        2       /* 0.50 - 0.50 * cos (PI * 2 * (7 + 0.5) / (200)*/
        ,        2       /* 0.50 - 0.50 * cos (PI * 2 * (8 + 0.5) / (200)*/
        ,        3       /* 0.50 - 0.50 * cos (PI * 2 * (9 + 0.5) / (200)*/
        ,        3       /* 0.50 - 0.50 * cos (PI * 2 * (10 + 0.5) / (200)*/
        ,        4       /* 0.50 - 0.50 * cos (PI * 2 * (11 + 0.5) / (200)*/
        ,        5       /* 0.50 - 0.50 * cos (PI * 2 * (12 + 0.5) / (200)*/
        ,        6       /* 0.50 - 0.50 * cos (PI * 2 * (13 + 0.5) / (200)*/
        ,        7       /* 0.50 - 0.50 * cos (PI * 2 * (14 + 0.5) / (200)*/
        ,        7       /* 0.50 - 0.50 * cos (PI * 2 * (15 + 0.5) / (200)*/
        ,        8       /* 0.50 - 0.50 * cos (PI * 2 * (16 + 0.5) / (200)*/
        ,        9       /* 0.50 - 0.50 * cos (PI * 2 * (17 + 0.5) / (200)*/
        ,       11       /* 0.50 - 0.50 * cos (PI * 2 * (18 + 0.5) / (200)*/
        ,       12       /* 0.50 - 0.50 * cos (PI * 2 * (19 + 0.5) / (200)*/
        ,       13       /* 0.50 - 0.50 * cos (PI * 2 * (20 + 0.5) / (200)*/
        ,       14       /* 0.50 - 0.50 * cos (PI * 2 * (21 + 0.5) / (200)*/
        ,       15       /* 0.50 - 0.50 * cos (PI * 2 * (22 + 0.5) / (200)*/
        ,       17       /* 0.50 - 0.50 * cos (PI * 2 * (23 + 0.5) / (200)*/
        ,       18       /* 0.50 - 0.50 * cos (PI * 2 * (24 + 0.5) / (200)*/
        ,       19       /* 0.50 - 0.50 * cos (PI * 2 * (25 + 0.5) / (200)*/
        ,       21       /* 0.50 - 0.50 * cos (PI * 2 * (26 + 0.5) / (200)*/
        ,       22       /* 0.50 - 0.50 * cos (PI * 2 * (27 + 0.5) / (200)*/
        ,       24       /* 0.50 - 0.50 * cos (PI * 2 * (28 + 0.5) / (200)*/
        ,       26       /* 0.50 - 0.50 * cos (PI * 2 * (29 + 0.5) / (200)*/
        ,       27       /* 0.50 - 0.50 * cos (PI * 2 * (30 + 0.5) / (200)*/
        ,       29       /* 0.50 - 0.50 * cos (PI * 2 * (31 + 0.5) / (200)*/
        ,       31       /* 0.50 - 0.50 * cos (PI * 2 * (32 + 0.5) / (200)*/
        ,       32       /* 0.50 - 0.50 * cos (PI * 2 * (33 + 0.5) / (200)*/
        ,       34       /* 0.50 - 0.50 * cos (PI * 2 * (34 + 0.5) / (200)*/
        ,       36       /* 0.50 - 0.50 * cos (PI * 2 * (35 + 0.5) / (200)*/
        ,       38       /* 0.50 - 0.50 * cos (PI * 2 * (36 + 0.5) / (200)*/
        ,       40       /* 0.50 - 0.50 * cos (PI * 2 * (37 + 0.5) / (200)*/
        ,       41       /* 0.50 - 0.50 * cos (PI * 2 * (38 + 0.5) / (200)*/
        ,       43       /* 0.50 - 0.50 * cos (PI * 2 * (39 + 0.5) / (200)*/
        ,       45       /* 0.50 - 0.50 * cos (PI * 2 * (40 + 0.5) / (200)*/
        ,       47       /* 0.50 - 0.50 * cos (PI * 2 * (41 + 0.5) / (200)*/
        ,       49       /* 0.50 - 0.50 * cos (PI * 2 * (42 + 0.5) / (200)*/
        ,       51       /* 0.50 - 0.50 * cos (PI * 2 * (43 + 0.5) / (200)*/
        ,       53       /* 0.50 - 0.50 * cos (PI * 2 * (44 + 0.5) / (200)*/
        ,       55       /* 0.50 - 0.50 * cos (PI * 2 * (45 + 0.5) / (200)*/
        ,       57       /* 0.50 - 0.50 * cos (PI * 2 * (46 + 0.5) / (200)*/
        ,       59       /* 0.50 - 0.50 * cos (PI * 2 * (47 + 0.5) / (200)*/
        ,       61       /* 0.50 - 0.50 * cos (PI * 2 * (48 + 0.5) / (200)*/
        ,       63       /* 0.50 - 0.50 * cos (PI * 2 * (49 + 0.5) / (200)*/
        ,       65       /* 0.50 - 0.50 * cos (PI * 2 * (50 + 0.5) / (200)*/
        ,       67       /* 0.50 - 0.50 * cos (PI * 2 * (51 + 0.5) / (200)*/
        ,       69       /* 0.50 - 0.50 * cos (PI * 2 * (52 + 0.5) / (200)*/
        ,       71       /* 0.50 - 0.50 * cos (PI * 2 * (53 + 0.5) / (200)*/
        ,       73       /* 0.50 - 0.50 * cos (PI * 2 * (54 + 0.5) / (200)*/
        ,       75       /* 0.50 - 0.50 * cos (PI * 2 * (55 + 0.5) / (200)*/
        ,       77       /* 0.50 - 0.50 * cos (PI * 2 * (56 + 0.5) / (200)*/
        ,       79       /* 0.50 - 0.50 * cos (PI * 2 * (57 + 0.5) / (200)*/
        ,       81       /* 0.50 - 0.50 * cos (PI * 2 * (58 + 0.5) / (200)*/
        ,       83       /* 0.50 - 0.50 * cos (PI * 2 * (59 + 0.5) / (200)*/
        ,       85       /* 0.50 - 0.50 * cos (PI * 2 * (60 + 0.5) / (200)*/
        ,       87       /* 0.50 - 0.50 * cos (PI * 2 * (61 + 0.5) / (200)*/
        ,       88       /* 0.50 - 0.50 * cos (PI * 2 * (62 + 0.5) / (200)*/
        ,       90       /* 0.50 - 0.50 * cos (PI * 2 * (63 + 0.5) / (200)*/
        ,       92       /* 0.50 - 0.50 * cos (PI * 2 * (64 + 0.5) / (200)*/
        ,       94       /* 0.50 - 0.50 * cos (PI * 2 * (65 + 0.5) / (200)*/
        ,       96       /* 0.50 - 0.50 * cos (PI * 2 * (66 + 0.5) / (200)*/
        ,       97       /* 0.50 - 0.50 * cos (PI * 2 * (67 + 0.5) / (200)*/
        ,       99       /* 0.50 - 0.50 * cos (PI * 2 * (68 + 0.5) / (200)*/
        ,      101       /* 0.50 - 0.50 * cos (PI * 2 * (69 + 0.5) / (200)*/
        ,      102       /* 0.50 - 0.50 * cos (PI * 2 * (70 + 0.5) / (200)*/
        ,      104       /* 0.50 - 0.50 * cos (PI * 2 * (71 + 0.5) / (200)*/
        ,      106       /* 0.50 - 0.50 * cos (PI * 2 * (72 + 0.5) / (200)*/
        ,      107       /* 0.50 - 0.50 * cos (PI * 2 * (73 + 0.5) / (200)*/
        ,      109       /* 0.50 - 0.50 * cos (PI * 2 * (74 + 0.5) / (200)*/
        ,      110       /* 0.50 - 0.50 * cos (PI * 2 * (75 + 0.5) / (200)*/
        ,      111       /* 0.50 - 0.50 * cos (PI * 2 * (76 + 0.5) / (200)*/
        ,      113       /* 0.50 - 0.50 * cos (PI * 2 * (77 + 0.5) / (200)*/
        ,      114       /* 0.50 - 0.50 * cos (PI * 2 * (78 + 0.5) / (200)*/
        ,      115       /* 0.50 - 0.50 * cos (PI * 2 * (79 + 0.5) / (200)*/
        ,      116       /* 0.50 - 0.50 * cos (PI * 2 * (80 + 0.5) / (200)*/
        ,      117       /* 0.50 - 0.50 * cos (PI * 2 * (81 + 0.5) / (200)*/
        ,      119       /* 0.50 - 0.50 * cos (PI * 2 * (82 + 0.5) / (200)*/
        ,      120       /* 0.50 - 0.50 * cos (PI * 2 * (83 + 0.5) / (200)*/
        ,      121       /* 0.50 - 0.50 * cos (PI * 2 * (84 + 0.5) / (200)*/
        ,      121       /* 0.50 - 0.50 * cos (PI * 2 * (85 + 0.5) / (200)*/
        ,      122       /* 0.50 - 0.50 * cos (PI * 2 * (86 + 0.5) / (200)*/
        ,      123       /* 0.50 - 0.50 * cos (PI * 2 * (87 + 0.5) / (200)*/
        ,      124       /* 0.50 - 0.50 * cos (PI * 2 * (88 + 0.5) / (200)*/
        ,      125       /* 0.50 - 0.50 * cos (PI * 2 * (89 + 0.5) / (200)*/
        ,      125       /* 0.50 - 0.50 * cos (PI * 2 * (90 + 0.5) / (200)*/
        ,      126       /* 0.50 - 0.50 * cos (PI * 2 * (91 + 0.5) / (200)*/
        ,      126       /* 0.50 - 0.50 * cos (PI * 2 * (92 + 0.5) / (200)*/
        ,      127       /* 0.50 - 0.50 * cos (PI * 2 * (93 + 0.5) / (200)*/
        ,      127       /* 0.50 - 0.50 * cos (PI * 2 * (94 + 0.5) / (200)*/
        ,      127       /* 0.50 - 0.50 * cos (PI * 2 * (95 + 0.5) / (200)*/
        ,      128       /* 0.50 - 0.50 * cos (PI * 2 * (96 + 0.5) / (200)*/
        ,      128       /* 0.50 - 0.50 * cos (PI * 2 * (97 + 0.5) / (200)*/
        ,      128       /* 0.50 - 0.50 * cos (PI * 2 * (98 + 0.5) / (200)*/
        ,      128       /* 0.50 - 0.50 * cos (PI * 2 * (99 + 0.5) / (200)*/
        } ;

static const X_INT16 Hann_sh24_lo[100] = {
               517       /* 0.50 - 0.50 * cos (PI * 2 * (0 + 0.5) / (200)*/
        ,     4656       /* 0.50 - 0.50 * cos (PI * 2 * (1 + 0.5) / (200)*/
        ,    12930       /* 0.50 - 0.50 * cos (PI * 2 * (2 + 0.5) / (200)*/
        ,    25330       /* 0.50 - 0.50 * cos (PI * 2 * (3 + 0.5) / (200)*/
        ,   -23692       /* 0.50 - 0.50 * cos (PI * 2 * (4 + 0.5) / (200)*/
        ,    -3080       /* 0.50 - 0.50 * cos (PI * 2 * (5 + 0.5) / (200)*/
        ,    21610       /* 0.50 - 0.50 * cos (PI * 2 * (6 + 0.5) / (200)*/
        ,   -15183       /* 0.50 - 0.50 * cos (PI * 2 * (7 + 0.5) / (200)*/
        ,    17585       /* 0.50 - 0.50 * cos (PI * 2 * (8 + 0.5) / (200)*/
        ,   -11190       /* 0.50 - 0.50 * cos (PI * 2 * (9 + 0.5) / (200)*/
        ,    29526       /* 0.50 - 0.50 * cos (PI * 2 * (10 + 0.5) / (200)*/
        ,     8623       /* 0.50 - 0.50 * cos (PI * 2 * (11 + 0.5) / (200)*/
        ,    -8408       /* 0.50 - 0.50 * cos (PI * 2 * (12 + 0.5) / (200)*/
        ,   -21614       /* 0.50 - 0.50 * cos (PI * 2 * (13 + 0.5) / (200)*/
        ,   -31049       /* 0.50 - 0.50 * cos (PI * 2 * (14 + 0.5) / (200)*/
        ,    28770       /* 0.50 - 0.50 * cos (PI * 2 * (15 + 0.5) / (200)*/
        ,    26711       /* 0.50 - 0.50 * cos (PI * 2 * (16 + 0.5) / (200)*/
        ,    28248       /* 0.50 - 0.50 * cos (PI * 2 * (17 + 0.5) / (200)*/
        ,   -32222       /* 0.50 - 0.50 * cos (PI * 2 * (18 + 0.5) / (200)*/
        ,   -23697       /* 0.50 - 0.50 * cos (PI * 2 * (19 + 0.5) / (200)*/
        ,   -11785       /* 0.50 - 0.50 * cos (PI * 2 * (20 + 0.5) / (200)*/
        ,     3438       /* 0.50 - 0.50 * cos (PI * 2 * (21 + 0.5) / (200)*/
        ,    21890       /* 0.50 - 0.50 * cos (PI * 2 * (22 + 0.5) / (200)*/
        ,   -22046       /* 0.50 - 0.50 * cos (PI * 2 * (23 + 0.5) / (200)*/
        ,     2616       /* 0.50 - 0.50 * cos (PI * 2 * (24 + 0.5) / (200)*/
        ,    30250       /* 0.50 - 0.50 * cos (PI * 2 * (25 + 0.5) / (200)*/
        ,    -4771       /* 0.50 - 0.50 * cos (PI * 2 * (26 + 0.5) / (200)*/
        ,    28529       /* 0.50 - 0.50 * cos (PI * 2 * (27 + 0.5) / (200)*/
        ,    -1018       /* 0.50 - 0.50 * cos (PI * 2 * (28 + 0.5) / (200)*/
        ,   -27977       /* 0.50 - 0.50 * cos (PI * 2 * (29 + 0.5) / (200)*/
        ,    13085       /* 0.50 - 0.50 * cos (PI * 2 * (30 + 0.5) / (200)*/
        ,    -9009       /* 0.50 - 0.50 * cos (PI * 2 * (31 + 0.5) / (200)*/
        ,   -28830       /* 0.50 - 0.50 * cos (PI * 2 * (32 + 0.5) / (200)*/
        ,    19048       /* 0.50 - 0.50 * cos (PI * 2 * (33 + 0.5) / (200)*/
        ,     3440       /* 0.50 - 0.50 * cos (PI * 2 * (34 + 0.5) / (200)*/
        ,   -10231       /* 0.50 - 0.50 * cos (PI * 2 * (35 + 0.5) / (200)*/
        ,   -22080       /* 0.50 - 0.50 * cos (PI * 2 * (36 + 0.5) / (200)*/
        ,   -32227       /* 0.50 - 0.50 * cos (PI * 2 * (37 + 0.5) / (200)*/
        ,    24747       /* 0.50 - 0.50 * cos (PI * 2 * (38 + 0.5) / (200)*/
        ,    17648       /* 0.50 - 0.50 * cos (PI * 2 * (39 + 0.5) / (200)*/
        ,    11889       /* 0.50 - 0.50 * cos (PI * 2 * (40 + 0.5) / (200)*/
        ,     7348       /* 0.50 - 0.50 * cos (PI * 2 * (41 + 0.5) / (200)*/
        ,     3899       /* 0.50 - 0.50 * cos (PI * 2 * (42 + 0.5) / (200)*/
        ,     1416       /* 0.50 - 0.50 * cos (PI * 2 * (43 + 0.5) / (200)*/
        ,     -227       /* 0.50 - 0.50 * cos (PI * 2 * (44 + 0.5) / (200)*/
        ,    -1159       /* 0.50 - 0.50 * cos (PI * 2 * (45 + 0.5) / (200)*/
        ,    -1507       /* 0.50 - 0.50 * cos (PI * 2 * (46 + 0.5) / (200)*/
        ,    -1401       /* 0.50 - 0.50 * cos (PI * 2 * (47 + 0.5) / (200)*/
        ,     -971       /* 0.50 - 0.50 * cos (PI * 2 * (48 + 0.5) / (200)*/
        ,     -345       /* 0.50 - 0.50 * cos (PI * 2 * (49 + 0.5) / (200)*/
        ,      345       /* 0.50 - 0.50 * cos (PI * 2 * (50 + 0.5) / (200)*/
        ,      971       /* 0.50 - 0.50 * cos (PI * 2 * (51 + 0.5) / (200)*/
        ,     1401       /* 0.50 - 0.50 * cos (PI * 2 * (52 + 0.5) / (200)*/
        ,     1507       /* 0.50 - 0.50 * cos (PI * 2 * (53 + 0.5) / (200)*/
        ,     1159       /* 0.50 - 0.50 * cos (PI * 2 * (54 + 0.5) / (200)*/
        ,      227       /* 0.50 - 0.50 * cos (PI * 2 * (55 + 0.5) / (200)*/
        ,    -1416       /* 0.50 - 0.50 * cos (PI * 2 * (56 + 0.5) / (200)*/
        ,    -3899       /* 0.50 - 0.50 * cos (PI * 2 * (57 + 0.5) / (200)*/
        ,    -7348       /* 0.50 - 0.50 * cos (PI * 2 * (58 + 0.5) / (200)*/
        ,   -11889       /* 0.50 - 0.50 * cos (PI * 2 * (59 + 0.5) / (200)*/
        ,   -17648       /* 0.50 - 0.50 * cos (PI * 2 * (60 + 0.5) / (200)*/
        ,   -24747       /* 0.50 - 0.50 * cos (PI * 2 * (61 + 0.5) / (200)*/
        ,    32227       /* 0.50 - 0.50 * cos (PI * 2 * (62 + 0.5) / (200)*/
        ,    22080       /* 0.50 - 0.50 * cos (PI * 2 * (63 + 0.5) / (200)*/
        ,    10231       /* 0.50 - 0.50 * cos (PI * 2 * (64 + 0.5) / (200)*/
        ,    -3440       /* 0.50 - 0.50 * cos (PI * 2 * (65 + 0.5) / (200)*/
        ,   -19048       /* 0.50 - 0.50 * cos (PI * 2 * (66 + 0.5) / (200)*/
        ,    28830       /* 0.50 - 0.50 * cos (PI * 2 * (67 + 0.5) / (200)*/
        ,     9009       /* 0.50 - 0.50 * cos (PI * 2 * (68 + 0.5) / (200)*/
        ,   -13085       /* 0.50 - 0.50 * cos (PI * 2 * (69 + 0.5) / (200)*/
        ,    27977       /* 0.50 - 0.50 * cos (PI * 2 * (70 + 0.5) / (200)*/
        ,     1018       /* 0.50 - 0.50 * cos (PI * 2 * (71 + 0.5) / (200)*/
        ,   -28529       /* 0.50 - 0.50 * cos (PI * 2 * (72 + 0.5) / (200)*/
        ,     4771       /* 0.50 - 0.50 * cos (PI * 2 * (73 + 0.5) / (200)*/
        ,   -30250       /* 0.50 - 0.50 * cos (PI * 2 * (74 + 0.5) / (200)*/
        ,    -2616       /* 0.50 - 0.50 * cos (PI * 2 * (75 + 0.5) / (200)*/
        ,    22046       /* 0.50 - 0.50 * cos (PI * 2 * (76 + 0.5) / (200)*/
        ,   -21890       /* 0.50 - 0.50 * cos (PI * 2 * (77 + 0.5) / (200)*/
        ,    -3438       /* 0.50 - 0.50 * cos (PI * 2 * (78 + 0.5) / (200)*/
        ,    11785       /* 0.50 - 0.50 * cos (PI * 2 * (79 + 0.5) / (200)*/
        ,    23697       /* 0.50 - 0.50 * cos (PI * 2 * (80 + 0.5) / (200)*/
        ,    32222       /* 0.50 - 0.50 * cos (PI * 2 * (81 + 0.5) / (200)*/
        ,   -28248       /* 0.50 - 0.50 * cos (PI * 2 * (82 + 0.5) / (200)*/
        ,   -26711       /* 0.50 - 0.50 * cos (PI * 2 * (83 + 0.5) / (200)*/
        ,   -28770       /* 0.50 - 0.50 * cos (PI * 2 * (84 + 0.5) / (200)*/
        ,    31049       /* 0.50 - 0.50 * cos (PI * 2 * (85 + 0.5) / (200)*/
        ,    21614       /* 0.50 - 0.50 * cos (PI * 2 * (86 + 0.5) / (200)*/
        ,     8408       /* 0.50 - 0.50 * cos (PI * 2 * (87 + 0.5) / (200)*/
        ,    -8623       /* 0.50 - 0.50 * cos (PI * 2 * (88 + 0.5) / (200)*/
        ,   -29526       /* 0.50 - 0.50 * cos (PI * 2 * (89 + 0.5) / (200)*/
        ,    11190       /* 0.50 - 0.50 * cos (PI * 2 * (90 + 0.5) / (200)*/
        ,   -17585       /* 0.50 - 0.50 * cos (PI * 2 * (91 + 0.5) / (200)*/
        ,    15183       /* 0.50 - 0.50 * cos (PI * 2 * (92 + 0.5) / (200)*/
        ,   -21610       /* 0.50 - 0.50 * cos (PI * 2 * (93 + 0.5) / (200)*/
        ,     3080       /* 0.50 - 0.50 * cos (PI * 2 * (94 + 0.5) / (200)*/
        ,    23692       /* 0.50 - 0.50 * cos (PI * 2 * (95 + 0.5) / (200)*/
        ,   -25330       /* 0.50 - 0.50 * cos (PI * 2 * (96 + 0.5) / (200)*/
        ,   -12930       /* 0.50 - 0.50 * cos (PI * 2 * (97 + 0.5) / (200)*/
        ,    -4656       /* 0.50 - 0.50 * cos (PI * 2 * (98 + 0.5) / (200)*/
        ,     -517       /* 0.50 - 0.50 * cos (PI * 2 * (99 + 0.5) / (200)*/
        } ;

#define SHFT_FACT_MUL   14
#define GAIN_FACT_MUL   ( ( ( 1 << SHFT_FACT_MUL ) / 20 ) * 20)         


//# 2**14 pond[i]
static const X_INT16 pondMelFilterNoise[] = {
            1,   0 ,16384 /* Channel 0 Central = 0.00 */
        ,   1,   1 ,16384 /* Channel 1 Central = 1.00 */
        ,   1,   2 ,16384 /* Channel 2 Central = 2.00 */
        ,   1,   3 ,16384 /* Channel 3 Central = 3.00 */
        ,   1,   4 ,16384 /* Channel 4 Central = 4.00 */
        ,   2,   5 ,10924,5460 /* Channel 5 Central = 5.33 */
        ,   2,   6 ,5460,10924 /* Channel 6 Central = 6.67 */
        ,   2,   8 ,10924,5460 /* Channel 7 Central = 8.33 */
        ,   3,   9 ,4096,8192,4096 /* Channel 8 Central = 10.00 */
        ,   3,  11 ,4096,8192,4096 /* Channel 9 Central = 12.00 */
        ,   3,  13 ,4096,8192,4096 /* Channel 10 Central = 14.00 */
        ,   3,  15 ,4096,8192,4096 /* Channel 11 Central = 16.00 */
        ,   3,  17 ,4096,8192,4096 /* Channel 12 Central = 18.00 */
        ,   4,  19 ,3276,6552,4368,2184 /* Channel 13 Central = 20.33 */
        ,   5,  21 ,1820,3640,5460,3640,1820 /* Channel 14 Central = 23.00 */
        ,   5,  24 ,1820,3640,5460,3640,1820 /* Channel 15 Central = 26.00 */
        ,   5,  27 ,1820,3640,5460,3640,1820 /* Channel 16 Central = 29.00 */
        ,   6,  30 ,1560,3120,4680,3512,2340,1172 /* Channel 17 Central = 32.33 */
        ,   6,  33 ,1172,2340,3512,4680,3120,1560 /* Channel 18 Central = 35.67 */
        ,   7,  37 ,1364,2732,4096,3276,2456,1640,820 /* Channel 19 Central = 39.67 */
        ,   8,  40 ,728,1456,2184,2912,3640,2732,1820,912 /* Channel 20 Central = 43.67 */
        ,   8,  45 ,912,1820,2732,3640,2912,2184,1456,728 /* Channel 21 Central = 48.33 */
        ,   9,  49 ,656,1312,1968,2620,3276,2620,1968,1312,656 /* Channel 22 Central = 53.00 */
        ,  10,  54 ,596,1192,1788,2384,2980,2484,1984,1488,992,496 /* Channel 23 Central = 58.33 */
        ,   6,  59 ,780,1560,2340,3120,3900,4680 /* Channel 24 Central = 64.00 */
        , 0 /* no more filter, length = 0 */
        } ;

#define IDCT_ORDER  25
//# idctMel16[i,j] = 2**(15+idctMel16[i,0]) * idcMel[i]
static const X_INT16 idctMel16[234] = {
           2  , 1024, 2048, 2048, 2048, 2389, 2731, 3072, 3413, 3755, 4096, 4096, 4096, 4437, 5120, 5803, 6144, 6485, 6827, 7509, 8192, 8875, 9557,10240,11264, 5803
        ,  2 ,  989, 1976, 1969, 1957, 2264, 2549, 2811, 3026, 3200, 3291, 3059, 2799, 2720, 2681, 2397, 1723,  919, -108,-1299,-2909,-4647,-6636,-8487,-10465,-5607
        ,  2 ,  890, 1772, 1747, 1704, 1919, 2056, 2119, 2029, 1814, 1363,  695,    0, -753,-1837,-3201,-4442,-5396,-5933,-6111,-5199,-3185,  272, 4197, 8317, 5045
        ,  2 ,  740, 1464, 1417, 1338, 1436, 1396, 1234,  831,  266, -578,-1396,-2094,-2829,-3661,-4069,-3433,-2004,  242, 2791, 5353, 6346, 4640,  363,-5468,-4195
        ,  2 ,  559, 1097, 1033,  930,  923,  746,  434, -122, -785,-1582,-2067,-2237,-2239,-1844, -618, 1284, 2945, 3720, 3083,293,-3196,-5208,-3107, 2721, 3169
        ,  3 ,  744, 1443, 1312, 1102,  964,  513, -146,-1133,-2108,-2918,-2847,-2104, -936, 1021, 3385, 4441, 3163, -405,-4272,-5665,-1770, 5293, 6724,-1465,-4215
        ,  4 ,  814, 1558, 1353, 1033,  727,    0, -934,-2097,-2927,-3007,-1809,    0, 1959, 3894, 4067,  953,-3270,-5399,-2813, 4130, 6749, -744,-8099, -877, 4612
        ,  5 ,  614, 1155,  949,  631,  279, -423,-1214,-1966,-2153,-1363,  241, 1735, 2646, 2339, -171,-3246,-3333,  467, 4283, 2384,-4054,-3560, 4928, 2480,-3477
        ,  8 ,  558, 1031,  789,  427,    0, -744,-1450,-1844,-1447,    0, 1578, 2232, 1710, -364,-2921,-2367, 1352, 3688,  534,-4425, -631, 5163,-2135,-3736, 3162
} ;


// 2**10

static const X_INT16 pondMelFilter16k[] = {
  7, 1, 1024, 878, 731, 585, 439, 293, 146,
  17, 2, 146, 293, 439, 585, 731, 877, 1024, 931, 838, 745, 652, 559, 465, 372, 279, 186, 93, 
  28, 9, 93, 186, 279, 372, 465, 559, 652, 745, 838, 931, 1024, 967, 910, 853, 796, 740, 683, 626, 569, 512, 455, 398, 341, 284, 228, 171, 114, 57, 
  44, 20, 57, 114, 171, 228, 284, 341, 398, 455, 512, 569, 626, 683, 740, 796, 853, 910, 967, 1024, 986, 948, 910, 872, 834, 796, 759, 721, 683, 645, 607, 569, 531, 493, 455, 417, 379, 341, 303, 265, 228, 190, 152, 114, 76, 38, 
  27, 38, 38, 76, 114, 152, 190, 228, 265, 303, 341, 379, 417, 455, 493, 531, 569, 607, 645, 683, 721, 759, 796, 834, 872, 910, 948, 986, 1024,
  0
};

/* tabulation for tab[i] = norm_s((i << 7) |0x7f) */
const static X_INT16 decal7[256] = { 8 ,
                               7 ,
							   6, 6,
							   5, 5, 5, 5,
							   4, 4, 4, 4, 4, 4, 4, 4,
							   3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
							   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
							   2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
							   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
							   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
							   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
							   1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
							   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
} ;



/************************************ local functions **************************************************************************/
/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoMelFB_B
 *
 * PURPOSE:       Performs mel filtering on FFT magnitude spectrum using the
 *                filter bank defined by a chained list of FFT window
 *                structures
 *
 * INPUT:
 *   - SigFFT32       Pointer to signal FFT magnitude spectrum
 *
 * OUTPUT
 *   - intFilter32    Filter bank outputs s
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/
static void DoMelFB_B (
        X_INT32 *SigFFT32           /* (i) power spectrum of bins */
        , X_INT32 *intFilter32      /* (o) power spectrum of mel filters */
)
{
    const X_INT16 *ptPond16 = pondMelFilter16k ;
    int nf = 0 ;
    int length ;

    test(); move16();
    while((length = * ptPond16++) > 0) {
        int j ;
        X_INT32 sum = L_mult(0,0)  ;
        X_INT16 first = *ptPond16++ ; move16();
        for(j=0;j<length;j++) {
            X_INT32 sf = SigFFT32[first+j] ;
            X_INT16 low = shr_r(extract_l(sf),1) ;
            X_INT16 high  = extract_h(sf) ;
            high = sub(high,shr(low,15)) ;
            sum=L_mac(L_mac(sum,(*ptPond16),high),1,mult_r((*ptPond16),low));
            ptPond16++ ;
        }
        intFilter32[nf++] = sum    ;    move32();
        test(); move16();
    }
    return;
}

/*---------------------------------------------------------------------------
 * FUNCTION NAME: FilterBank16
 *
 * PURPOSE:      applies Mel-frequency scale to the wiener filter 
 *
 * INPUT:
 *   - W_16       Wiener filter 
 *
 * OUTPUT
 *   - Filter16    Mel-warped Wiener filter
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/
static void FilterBank16 (
                X_INT16 *W_16           /* (i) Wiener filter */
                ,X_INT16 *Filter16      /* (o) Mel-warped Wiener filter */
) {
    const X_INT16 *ptPond16  = pondMelFilterNoise ;
    int nf = 0 ;
    int length ;
    test(); move16();
    while((length = * ptPond16++) > 0) {
        X_INT16 first = *ptPond16++ ;
        X_INT32 sum = 0 ;
        int j ;
        move32(); move16();
        for(j=0;j<length;j++) {
            sum = L_mac(sum,W_16[first+j],(X_INT16) ( *ptPond16 ) ); 
            ptPond16++ ;
        }
        Filter16[nf++] = extract_h(sum) ; move16();
        test(); move16();
    }
    return;
}

/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoMelIDCT_F16
 *
 * PURPOSE:      computes Mel-warped inverse DCT for Wiener filter 
 *
 * INPUT:
 *   - W_16       Mel-warped Wiener filter 
 *   - idctMel16  Mel-warped inverse DCT coeffcients
 *
 * OUTPUT
 *   - W_32    Time-domain response of Wiener filter
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/
static void DoMelIDCT_F16 (
         X_INT16 * W_16         /* (i) Mel-warped Wiener filter */
        ,X_INT32 *W_32          /* (o) Time-domain response of Wiener filter */
        ,const X_INT16 *idctMel16     /* (i) Mel-warped inverse DCT coefficients */
) {
    int i ;
    for (i=0;i<NS_HALF_FILTER_LENGTH+1;i++) {
        X_INT32 val = L_mult(0,0) ;
        int j ;
        X_INT16 shft = *idctMel16++ ; move16();
        test();
        if(sub(shft, 2)==0) {
            for(j=0;j<IDCT_ORDER;j++) {
                val = L_add(val,L_shr_r(L_mult(*idctMel16,W_16[j]),2)) ; 
                idctMel16++ ;
            }
            W_32[NS_HALF_FILTER_LENGTH - i]   =  val  ; move32();
        } else {
            for(j=0;j<IDCT_ORDER;j++) {
                val = L_mac(val,*idctMel16,W_16[j]);
                idctMel16++ ;
            }
            W_32[NS_HALF_FILTER_LENGTH - i] = L_shr_r(val,shft) ;  move32();
        }
    }
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: DCOffsetFil_F
 *
 * PURPOSE:      removes DC offset , inplace
 *
 * INPUT:
 *   - Data32       Input signal 
 *   - prevSamples  DC offset context struct
 *   - DataLength   Signal length
 *   - curShift     Shift factor for input signal Data32[i] = Data[i] * 2**curShift
 *
 * OUTPUT
 *   - Data32     Output DC offset removed signal 
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/

static void DCOffsetFil_F (
            X_INT32 *Data32             /* (i/o)        input signal */
            ,DC_FILTER_F *prevSamples   /* (i/o)    DC offset context   */
            , X_INT16 DataLength        /* (i)      size of Data32      */
            ,X_INT16 curShift           /* (i)      shift factor for input signal   */ 
)
{
    X_INT16 i;
    X_INT32 aux;
    X_INT32 Prev_x32 = prevSamples->lastSampleIn32;
    X_INT32 prevDCOut32 = prevSamples->lastDCOut32;
    X_INT16 diff = sub(curShift,prevSamples->oldShift);
    move32(); move32(); 
    prevSamples->oldShift = curShift ; move16();
    test();
    if(diff > 0) {
        Prev_x32 = L_shl(Prev_x32,diff);
        prevDCOut32 = L_shl(prevDCOut32,diff);
    } else {
        diff = sub(0,diff);
        Prev_x32 = L_shr_r(Prev_x32,diff);
        prevDCOut32 = L_shr_r(prevDCOut32,diff);
    }
    for (i=0 ; i<DataLength ; i+=2)
    {
        aux = L_shr_r(Data32[i],2);		/* even point */
        prevDCOut32 = L_add(prevDCOut32,L_shr_r(L_sub(Prev_x32,prevDCOut32),10)) ;
        Data32[i] = L_sub(aux,prevDCOut32);  move32();
        
        Prev_x32 = L_shr_r(Data32[i+1],2); /* odd point */
        prevDCOut32 = L_add(prevDCOut32,L_shr_r(L_sub(aux,prevDCOut32),10)) ;
        Data32[i+1] = L_sub(Prev_x32,prevDCOut32);  move32();
    }
    prevSamples->lastDCOut32 = prevDCOut32 ; move32();
    prevSamples->lastSampleIn32 = Prev_x32; move32();
}

/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoSigWindowing16_F1
 *
 * PURPOSE:     windowing for first pass (plus zero padding for fft)
 *
 * INPUT:
 *   - DataIn16     Input signal 
 *   - prevSamples  DC offset context struct
 *   - FFTLength    FFT length (for zero padding)
 *   - decal        Shift factor for input signal DataIn16[i] = Data[i] * 2**decal
 *
 * OUTPUT
 *   - DataOut32     Output windowed signal 
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/

static void DoSigWindowing16_F1(
            X_INT16 * DataIn16          /* (i) Input signal             */
            ,X_INT32 * DataOut32        /* (o) Output windowed signal   */
            ,X_INT16 decal              /* (i) shift factor for input signal    */
            ,X_INT16 FFTLength) {       /* (i) FFT length (for zero padding)    */
    X_INT16 i;
    const X_INT16 frameLength = 200 ;
    test();
    if(sub(decal,8) < 0 ) {
        for (i=0 ; i<frameLength/2 ; i++) {
            X_INT16 lo = Hann_sh24_lo[i] ;
            X_INT16 hi = Hann_sh24_hi[i] ;
            X_INT16 di16 = shl(DataIn16[i],decal) ;
            DataOut32[i] = L_mac(L_deposit_l(mult(di16,lo)),di16,hi ) ;  move32();
        }
        for (i=frameLength/2-1 ; i>=0 ; i--) {
            X_INT16 lo = Hann_sh24_lo[i] ;
            X_INT16 hi = Hann_sh24_hi[i] ;
            X_INT16 dfi16 = shl(DataIn16[frameLength-1-i],decal) ;
            DataOut32[frameLength-1-i] = L_mac(L_deposit_l(mult(dfi16,lo)),dfi16,hi )  ; move32();
        }
    } else {
        X_INT16 decal_m8 = sub(decal,8) ;
        for (i=0 ; i<frameLength/2 ; i++) {
            X_INT16 lo = Hann_sh32_lo[i] ;
            X_INT16 hi = Hann_sh32_hi[i] ;
            X_INT16 di16 = shl(DataIn16[i], decal_m8) ;
            DataOut32[i] = L_mac(L_deposit_l(mult(di16,lo)),di16,hi ) ; move32(); 
        }
        for (i=frameLength/2-1 ; i>=0 ; i--) {
            X_INT16 lo = Hann_sh32_lo[i] ;
            X_INT16 hi = Hann_sh32_hi[i] ;
            X_INT16 dfi16 = shl(DataIn16[frameLength-1-i], decal_m8) ;
            DataOut32[frameLength-1-i] = L_mac(L_deposit_l(mult(dfi16,lo)),dfi16,hi )  ; move32(); 
        }
    }
    for (i=frameLength ; i<FFTLength ; i++) {
        DataOut32 [i] = 0; move32();
    }
    return;
    
}

/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoSigWindowing16_F2
 *
 * PURPOSE:     windowing for second pass (plus zero padding for fft)
 *
 * INPUT:
 *   - DataIn16     Input signal 
 *   - prevSamples  DC offset context struct
 *   - FFTLength    FFT length (for zero padding)
 *
 * OUTPUT
 *   - DataOut32     Output windowed signal 
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/

static void DoSigWindowing16_F2(
    X_INT16 * DataIn16          /* (i) Input signal             */
    ,X_INT32 * DataOut32        /* (o) Output windowed signal   */
    ,X_INT16 FFTLength          /* (i) FFT length (for zero padding)    */
) {
    X_INT16 i;
    const X_INT16 frameLength = 200 ;
    {
        for (i=0 ; i<frameLength/2 ; i++) {
            X_INT16 lo = Hann_sh24_lo[i] ;
            X_INT16 hi2 = (X_INT16) (Hann_sh24_hi[i] ) ;
            X_INT16 di16 = DataIn16[i] ;
            DataOut32[i] = L_mac(L_deposit_l(mult(di16,lo)),di16,hi2 ) ; move32(); 
        }
        for (i=frameLength/2-1 ; i>=0 ; i--) {
            X_INT16 lo = Hann_sh24_lo[i] ;
            X_INT16 hi2 = (X_INT16) (Hann_sh24_hi[i] ) ;
            X_INT16 dfi16 = DataIn16[frameLength-1-i] ;
            DataOut32[frameLength-1-i] = L_mac(L_deposit_l(mult(dfi16,lo)),dfi16,hi2 )  ; move32();
        }
    } 
    for (i=frameLength ; i<FFTLength ; i++) {
        DataOut32 [i] = 0; move32() ;
    }
    return;
    
}

#ifdef F_16KHZ

/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoSigWindowing16_F3
 *
 * PURPOSE:     windowing for second pass at 16Khz (plus zero padding for fft)
 *
 * INPUT:
 *   - DataIn16     Input signal 
 *   - prevSamples  DC offset context struct
 *   - FFTLength    FFT length (for zero padding)
 *
 * OUTPUT
 *   - DataOut32     Output windowed signal 
 *
 * RETURN VALUE
 *   none
 *
 *---------------------------------------------------------------------------*/
static void DoSigWindowing16_F3(
                                X_INT32 * DataIn32              /* (i) Input signal             */
                                ,X_INT32 * DataOut32            /* (o) Output windowed signal   */
                                ,X_INT16 FFTLength              /* (i) FFT length (for zero padding)    */
                                ) {
    X_INT16 i;
    const X_INT16 frameLength = 200 ;
    {
        for (i=0 ; i<frameLength/2 ; i++) {
            X_INT16 low = extract_l(DataIn32[i]) ;
            X_INT16 high  = extract_h(DataIn32[i]) ;
            X_INT16 a16 = Hann_sh32_hi[i];
            high = sub(high,shr(low,15)) ;
            DataOut32[i] = L_mac(L_mult(a16,high),mult_r(a16,shr_r(low,1)),1) ; move32();
            low = extract_l(DataIn32[frameLength - 1 - i]) ;
            high  = extract_h(DataIn32[frameLength - 1 - i]) ;
            high = sub(high,shr(low,15)) ;
            DataOut32[frameLength - 1 - i] = L_mac(L_mult(a16,high),mult_r(a16,shr_r(low,1)),1) ; move32();
        }
    } 
    for (i=frameLength ; i<FFTLength ; i++) {
        DataOut32 [i] = 0; move32();
    }
    return;
    
}
#endif




/*---------------------------------------------------------------------------
 * FUNCTION NAME: Square24_B
 *
 * PURPOSE:     computes x*x / 2**(16+2*decal)
 *
 * INPUT:
 *   - x        : X_INT32 value
 *   - decal    : shift factor
 *
 * OUTPUT
 *
 * RETURN VALUE x*x / 2**(16+2*decal)
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Square24_B(X_INT32 x,X_INT16 decal) {
    X_INT32 val32 = L_shl(x,decal) ;
    X_INT16 h = extract_h(val32);
    return L_shr(L_msu(L_mult(h,h),h,shr(extract_l(val32),15)),1); 
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: Square24d2_B
 *
 * PURPOSE:     computes x*x / 2**(16+2*decal+1)
 *
 * INPUT:
 *   - x        : X_INT32 value
 *   - decal    : shift factor
 *
 * OUTPUT
 *
 * RETURN VALUE x*x / 2**(16+2*decal+1)
 *   
 *
 *---------------------------------------------------------------------------*/
static X_INT32 Square24d2_B(X_INT32 x,X_INT16 decal) {
        X_INT32 val32 = L_shl(x,decal);
        X_INT16 h = extract_h(val32);
        return L_shr(L_msu(L_mult(h,h),h,shr(extract_l(val32),15)),2);
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: FFTtoPSD_F
 *
 * PURPOSE:     computes power spectrum of each bins
 *
 * INPUT:
 *   - FFTIn32  : FFT bins
 *   - decIn    : global shift factor for FFT bins FFTIn32[i] = FFT[i] * 2**decIn
 *
 * OUTPUT
 *   - PSDOut32 : power spectrum  bins
 *   - nSigSE_dec   : shift factor for each FFT bin PSDOut32(i] = PSDOut[i] * 2**nSigSE_dec[i]
 *
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void FFTtoPSD_F (
            X_INT32 *FFTIn32        /* (i) FFT bins */
            ,X_INT32  *PSDOut32     /* (o) Power spectrum bins */
            ,X_INT16 *nSigSE_dec    /* (o) shift factor for Power spectrum */
            ,X_INT16 decIn          /* (i) shift factor for FFT bins */
    ) {
    X_INT16 i ;
	X_INT16 max1 ;
    X_INT32 * psd32 =  PSDOut32 ;
    {
        X_INT16 decal ;
        X_INT16 max = extract_h(L_abs(FFTIn32[0])) | extract_h(L_abs(FFTIn32[2])) | extract_h(L_abs(FFTIn32[3])) ;
        logic16() ; logic16() ; 
		decal = decal7[shr(max,7)] ; move16();
        psd32[0] =  L_add( Square24d2_B( FFTIn32[0],decal),
                        L_add(Square24d2_B(FFTIn32[2],decal),Square24d2_B(FFTIn32[3],decal))
                    )  ; move32();
        nSigSE_dec[0] = sub(decIn,sub(8,decal)) ; move16();
		max1 = extract_h(L_abs(FFTIn32[1])) ; /* save before overlap */
	}
    for(i=1;i<NS_SPEC_ORDER-1;i++) {
        X_INT16   decal ;
       X_INT16 maxx = extract_h(L_abs(FFTIn32[4*i])) | extract_h(L_abs(FFTIn32[4*i+1]))
                        | extract_h(L_abs(FFTIn32[4*i+2])) | extract_h(L_abs(FFTIn32[4*i+3])) ;
        logic16() ; logic16() ; logic16() ; 
		decal = decal7[shr(maxx,7)] ; move16();
        nSigSE_dec[i] = sub(decIn,sub(8,decal)) ; move16();
        psd32[i] =  L_add(L_add(Square24d2_B( FFTIn32[4*i], decal),Square24d2_B(FFTIn32[4*i+1],decal)),
                        L_add(Square24d2_B(FFTIn32[4*i+2],decal),Square24d2_B(FFTIn32[4*i+3],decal))
                  )  ; move32();
    }
    {   /* last bin */
		X_INT16 decal;
 		decal = decal7[shr(max1,7)] ; move16();
        nSigSE_dec[i] = sub(decIn,sub(8,decal)) ; move16();
        psd32[i] =  Square24_B( FFTIn32[1],decal) ; move32() ;
    }
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: PSDMean_F
 *
 * PURPOSE:     computes the power spectrum mean over the 2 last frames
 *
 * INPUT:
 *   - PSDIn32  : current frame power spectrum
 *   - PSDIn_dec : shift factor for current frame power spectrum PSDIn32[i] = PSDIn[i] * 2**PSDIn_dec[i]
 *   - PSDAnt32 : previous frame power spectrum
 *   - PSDAnt_dec : shift factor for previous frame power spectrum PSDAnt32[i] = PSDIn[i] * 2**PSDAnt_dec[i]
 *
 * OUTPUT
 *   - PSDOut32 :  frame power spectrum mean
 *   - PSDIn_dec : shift factor for  frame power spectrum mean PSDOut32[i] = PSDOut[i] * 2**PSDOut_dec[i]
 *     PSDAnt32[i] = PSDIn32[i] ; PSDAnt_dec[i] = PSDIn_dec[i]
 *
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void PSDMean_F(
            X_INT32 * PSDIn32       /* (i)  current frame power spectrum  */
            ,X_INT16 *PSDIn_dec     /* (i)  shift factor for current frame power spectrum  */
            ,X_INT32 * PSDOut32     /* (o)   power spectrum mean */
            ,X_INT16 *PSDOut_dec    /* (o)   shift factor for power spectrum mean */
            ,X_INT32 *PSDAnt32      /* (i)  previous frame power spectrum  */
            ,X_INT16 *PSDAnt_dec    /* (i)  shift factor for previous frame power spectrum  */
) {
    X_INT16 i;
    for (i=0 ; i<NS_SPEC_ORDER ; i++){
        X_INT16 difDec = sub(PSDAnt_dec[i],PSDIn_dec[i]);
        test();
        if(difDec == 0) {
            X_INT32 tmp = L_add(L_shr_r(PSDAnt32[i],1),L_shr_r(PSDIn32[i],1));    
            PSDAnt32[i] = PSDIn32[i];    move32();
			PSDOut32[i] = tmp ; move32();
            PSDOut_dec[i] = PSDIn_dec[i] ; move16();
        }else {
            test();
            if ( difDec > 0 ) {
                X_INT32 tmp = L_add(L_shr_r(PSDAnt32[i],add(add(difDec,difDec),1)),L_shr_r(PSDIn32[i],1));     
                PSDAnt32[i] = PSDIn32[i]; move32() ;
				PSDOut32[i] =  tmp ;  move32();
                PSDOut_dec[i] = PSDAnt_dec[i] = PSDIn_dec[i] ;      move16(); move16();
            } else {
                X_INT32 tmp = L_add(L_shr_r(PSDAnt32[i],1),L_shr_r(PSDIn32[i],sub(1,add(difDec,difDec))));     
                PSDAnt32[i] = PSDIn32[i]; move32();
				PSDOut32[i] = tmp ;  move32();
                PSDOut_dec[i] = PSDAnt_dec[i] ; move16();
                PSDAnt_dec[i] = PSDIn_dec[i] ; move16();
            }
        }
    }
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: ApplyWF
 *
 * PURPOSE:     applies Wiener filter
 *
 * INPUT:
 *   - curFrame16   : current frame shift signal
 *   - W_32 : Wiener filter
 *   - frameShift   : frame shift size
 *   - halfMel : half length of filter
 *
 * OUTPUT
 *   - signalOut32  :  filtered signal
 *
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void     ApplyWF (
                X_INT16 *curFrame16     /* (i) signal to filter */
                ,X_INT32 *W_32          /* (i) filter           */
                , X_INT32 *signalOut32  /* (o) filtered signal  */
                ,int frameShift         /* (i) nb samples to filter */
                ,int halfMel            /* (i) half size of the filter  */
) {
    int i, j;
    X_INT16 W_lo[NS_HALF_FILTER_LENGTH+1] ;
    X_INT16 W_hi[NS_HALF_FILTER_LENGTH+1] ;
    for (j=0 ; j<halfMel+1 ; j++){
        W_hi[halfMel-j]  = extract_h(W_32[halfMel-j]) ; move16(); 
        W_lo[halfMel-j] = shr(extract_l(W_32[halfMel-j]),5) & 0x7ff ; logic16() ; move16();
    }
    for (i=0 ; i<frameShift ; i++)
    {
        X_INT32 val = L_mult(W_hi[halfMel],curFrame16[i]) ;
        X_INT32 val1 = L_mult(W_lo[halfMel],curFrame16[i]) ;
        for (j=halfMel ; j>0 ; j--){
            val = L_mac(val,W_hi[halfMel-j],curFrame16[i-j]) ;
            val1 = L_mac(val1,W_lo[halfMel-j],curFrame16[i-j]) ;
        }
        for (j=1 ; j<halfMel+1 ; j++){
            val = L_mac(val,W_hi[halfMel-j],curFrame16[i+j]) ;
            val1 = L_mac(val1,W_lo[halfMel-j],curFrame16[i+j]) ;
        }
        val = L_add(val,L_shr(val1,11)) ;
        signalOut32[i] = val  ; move32() ;
    }
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: VAD_F
 *
 * PURPOSE:   computes VAD
 *
 * INPUT:
 *   - vadNS    : VAD context struct 
 *   - newShiftFrame16 : frame shift signal
 *   - decalSig : shift factor for newShiftFrame16 : newShiftFrame16[i] = newShiftFrame[i] * 2**decalSig
 *
 * OUTPUT
 *   - vadNs updated
 *
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void VAD_F (
            VAD_DATA_NS_F *vadNS                /* (i/o) associated VAD structure */
            , const X_INT16 *newShiftFrame16    /* (i) new frame shift  */
            ,int decalSig                       /* (i) shift factor for input frame */
){
    X_INT16 i;
    X_INT16 flagVAD;
    X_INT16 hangOver;
    X_INT16 nbSpeechFrames;
    X_INT16 nbFrame;
    X_INT32  meanEn32 ;
    X_INT16 frameEn16  ;
#define NS_LAMBDA_LTE_LOWER_E16 ROUND16((1 - NS_LAMBDA_LTE_LOWER_E) * 32768.0)
#define NS_LAMBDA_LTE_HIGHER_E16 ROUND16((1 - NS_LAMBDA_LTE_HIGHER_E) * 32768.0)
    static const X_INT16  M1_LamdaLTE[NS_NB_FRAME_THRESHOLD_LTE - 2] = {
        ROUND16((1/2.0) * 32768.0) ,
        ROUND16((1/3.0) * 32768.0) ,
        ROUND16((1/4.0) * 32768.0) ,
        ROUND16((1/5.0) * 32768.0) ,
        ROUND16((1/6.0) * 32768.0) ,
        ROUND16((1/7.0) * 32768.0) ,
        ROUND16((1/8.0) * 32768.0) ,
        ROUND16((1/9.0) * 32768.0) 
    } ;

    nbFrame        = vadNS->nbFrame [0];
    meanEn32       = vadNS->meanEn32;         move32();
    flagVAD        = vadNS->flagVAD;          move16();
    hangOver       = vadNS->hangOver;         move16();
    nbSpeechFrames = vadNS->nbSpeechFrames;   move16();
    

    test();
    if(sub(decalSig, 3) >= 0) {
        X_INT32 frameEn32 = 128 ;   move32();
        for (i=0 ; i<NS_FRAME_SHIFT ; i++) {
            frameEn32 = L_mac(frameEn32,newShiftFrame16[i],newShiftFrame16[i]);
        }
        frameEn16 = sub(Log_2(frameEn32),7 * 1024) ;
    } else {
        X_INT32 frameEn32 = 2 ; move32();
        for (i=0 ; i<NS_FRAME_SHIFT ; i++) {
            X_INT16 ech = shr_r(newShiftFrame16[i],3);
            frameEn32 = L_mac(frameEn32,ech,ech);
            
        }
        frameEn16 = sub(Log_2(frameEn32),1024) ;
    }
    {
        X_INT16 diff =sub(frameEn16,round(meanEn32) ); 
        test(); test();
        if ((sub(diff,ROUND16( 64 * NS_SNR_THRESHOLD_UPD_LTE )) < 0 ) || (sub(nbFrame,NS_MIN_FRAME) < 0))
        {
            test(); test();
            if ( (diff  < 0 ) || (sub(nbFrame, NS_MIN_FRAME) < 0)) {
                test();
                if( sub(nbFrame,NS_NB_FRAME_THRESHOLD_LTE) < 0) {
                    test();
                    if(sub(nbFrame,1) == 0) {
                        meanEn32 = L_deposit_h(frameEn16) ;
                    } else {
                        meanEn32 = L_mac(meanEn32,M1_LamdaLTE[nbFrame - 2],diff) ;
                    }
                } else {
                    meanEn32 = L_mac(meanEn32,NS_LAMBDA_LTE_LOWER_E16,diff) ;
                }
            } else {
                meanEn32 = L_mac(meanEn32,NS_LAMBDA_LTE_HIGHER_E16,diff) ;
            }
            test();
            /* ROUND16() should be used instead of ROUND32() */
            if(sub(extract_h(meanEn32),ROUND32(64 * NS_ENERGY_FLOOR -32)) < 0) {
              /* ROUND16() should be used instead of ROUND32() */
              /* Too many WMOPs are counted for the following operation */
                meanEn32 = L_deposit_h(ROUND32(64 * NS_ENERGY_FLOOR -32));
            }
        }
    }
    test();
    if (sub(nbFrame,4) > 0){
        test();
        /* ROUND16() should be used instead of ROUND32() */
        if (sub(sub(frameEn16,extract_h(meanEn32)),ROUND32( 64 * NS_SNR_THRESHOLD_VAD ) )  > 0  )
        {
            flagVAD = 1; move16();
            nbSpeechFrames= add(nbSpeechFrames,1);
        }   else {
            test();
            if (sub(nbSpeechFrames,NS_MIN_SPEECH_FRAME_HANGOVER) > 0) 
            {
                hangOver = NS_HANGOVER; move16();
            }
            nbSpeechFrames = 0; move16();
            test(); move16();
            if (hangOver != 0){
                hangOver = sub(hangOver,1) ;
                flagVAD = 1;
            } else {
                flagVAD = 0;
            }
        }
    }
    vadNS->meanEn32 =   meanEn32 ; move32();
    vadNS->flagVAD        = flagVAD; move16();
    vadNS->hangOver       = hangOver; move16();
    vadNS->nbSpeechFrames = nbSpeechFrames; move16();
    return;
}


#define ADJUST_SHFT(val32,valDec)   \
{	X_INT16 kk ;\
	X_INT16 hig= extract_h(val32) | shl(0x20,valDec); logic16(); \
	kk = decal7[shr(shl(hig,1),7)] ; move16();		\
    val32 = L_shl(val32,kk) ; move32();	\
	valDec = add(valDec,kk) ; move16() ;	\
}

static const X_INT16 M1_LambdaNSEx2[NS_NB_FRAME_THRESHOLD_NSE] = {
        0,                          0x7fff               ,  2*ROUND16(16384.0 / 2) , 2*ROUND16(16384.0 / 3) ,2*ROUND16(16384.0 / 4) ,
        2*ROUND16(16384.0 / 5) ,  2*ROUND16(16384.0 / 6) ,  2*ROUND16(16384.0 / 7) , 2*ROUND16(16384.0 / 8) ,2*ROUND16(16384.0 / 9) , 
        2*ROUND16(16384.0 /10) ,  2*ROUND16(16384.0 /11) ,  2*ROUND16(16384.0 /12) , 2*ROUND16(16384.0 /13) ,2*ROUND16(16384.0 /14) ,
        2*ROUND16(16384.0 /15) ,  2*ROUND16(16384.0 /16) ,  2*ROUND16(16384.0 /17) , 2*ROUND16(16384.0 /18) ,2*ROUND16(16384.0 /19) , 
        2*ROUND16(16384.0 /20) ,  2*ROUND16(16384.0 /21) ,  2*ROUND16(16384.0 /22) , 2*ROUND16(16384.0 /23) ,2*ROUND16(16384.0 /24) ,
        2*ROUND16(16384.0 /25) ,  2*ROUND16(16384.0 /26) ,  2*ROUND16(16384.0 /27) , 2*ROUND16(16384.0 /28) ,2*ROUND16(16384.0 /29) , 
        2*ROUND16(16384.0 /30) ,  2*ROUND16(16384.0 /31) ,  2*ROUND16(16384.0 /32) , 2*ROUND16(16384.0 /33) ,2*ROUND16(16384.0 /34) ,
        2*ROUND16(16384.0 /35) ,  2*ROUND16(16384.0 /36) ,  2*ROUND16(16384.0 /37) , 2*ROUND16(16384.0 /38) ,2*ROUND16(16384.0 /39) , 
        2*ROUND16(16384.0 /40) ,  2*ROUND16(16384.0 /41) ,  2*ROUND16(16384.0 /42) , 2*ROUND16(16384.0 /43) ,2*ROUND16(16384.0 /44) ,
        2*ROUND16(16384.0 /45) ,  2*ROUND16(16384.0 /46) ,  2*ROUND16(16384.0 /47) , 2*ROUND16(16384.0 /48) ,2*ROUND16(16384.0 /49) , 
        2*ROUND16(16384.0 /50) ,  2*ROUND16(16384.0 /51) ,  2*ROUND16(16384.0 /52) , 2*ROUND16(16384.0 /53) ,2*ROUND16(16384.0 /54) ,
        2*ROUND16(16384.0 /55) ,  2*ROUND16(16384.0 /56) ,  2*ROUND16(16384.0 /57) , 2*ROUND16(16384.0 /58) ,2*ROUND16(16384.0 /59) , 
        2*ROUND16(16384.0 /60) ,  2*ROUND16(16384.0 /61) ,  2*ROUND16(16384.0 /62) , 2*ROUND16(16384.0 /63) ,2*ROUND16(16384.0 /64) ,
        2*ROUND16(16384.0 /65) ,  2*ROUND16(16384.0 /66) ,  2*ROUND16(16384.0 /67) , 2*ROUND16(16384.0 /68) ,2*ROUND16(16384.0 /69) , 
        2*ROUND16(16384.0 /70) ,  2*ROUND16(16384.0 /71) ,  2*ROUND16(16384.0 /72) , 2*ROUND16(16384.0 /73) ,2*ROUND16(16384.0 /74) ,
        2*ROUND16(16384.0 /75) ,  2*ROUND16(16384.0 /76) ,  2*ROUND16(16384.0 /77) , 2*ROUND16(16384.0 /78) ,2*ROUND16(16384.0 /79) , 
        2*ROUND16(16384.0 /80) ,  2*ROUND16(16384.0 /81) ,  2*ROUND16(16384.0 /82) , 2*ROUND16(16384.0 /83) ,2*ROUND16(16384.0 /84) ,
        2*ROUND16(16384.0 /85) ,  2*ROUND16(16384.0 /86) ,  2*ROUND16(16384.0 /87) , 2*ROUND16(16384.0 /88) ,2*ROUND16(16384.0 /89) , 
        2*ROUND16(16384.0 /90) ,  2*ROUND16(16384.0 /91) ,  2*ROUND16(16384.0 /92) , 2*ROUND16(16384.0 /93) ,2*ROUND16(16384.0 /94) ,
        2*ROUND16(16384.0 /95) ,  2*ROUND16(16384.0 /96) ,  2*ROUND16(16384.0 /97) , 2*ROUND16(16384.0 /98) ,2*ROUND16(16384.0 /99)  


} ;



/*---------------------------------------------------------------------------
 * FUNCTION NAME: NoiseEstimation_F1
 *
 * PURPOSE:     computes noise estimation for first pass
 *
 * INPUT:
 *   - NSX  : noise context struct 
 *   - PSDMeaned32 : power spectrum mean
 *   - PSDMeaned_dec : shift factor for PSDMeaned32 : PSDMeaned32[i] = PSDMeaned[i] * 2**PSDMeaned_dec[i]
 *   - nSigSE32 : signal power spectrum 
 *
 * OUTPUT
 *   - PSDMeaned16 :  square root of power spectrum density
 *   - NSX updated
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void NoiseEstimation_F1(
                NoiseSupStructX_F *NSX  /* (i/o) associated structure */
                ,X_INT32 *PSDMeaned32   /* (i) power spectrum density mean */
                ,X_INT16 *PSDMeaned16   /* (o) square root of power spectrum density */
                ,X_INT16 *PSDMeaned_dec /* (i) shift factor for PSDMeaned16 */
				,X_INT32 *nSigSE32
) {
    /* 
       the following line is obviously a mistake - 
       NSX->nsVar should not be copied in the stack !! 
       I propose to delete this line.
    */
    NS_VAR_F    nsVar = NSX->nsVar;
    
    X_INT16   i;
    X_INT16   flagVAD;
    X_INT16   nbFrame;
    X_INT32 *  noiseSE_32 ;
    X_INT32 * denSigSE_32 ;
    X_INT16 * nSigSE_dec ;
    X_INT16 * denSigSE_dec ;
    X_INT16 * noiseSE_dec = NSX->nsVar.spectrum.noiseSE1_dec ;
    nSigSE_dec = NSX->nsVar.spectrum.nSigSE1Cur_dec ;
 

    denSigSE_32 = NSX->nsVar.spectrum.denSigSE1_32 ;
    denSigSE_dec = NSX->nsVar.spectrum.denSigSE1_dec ;
    noiseSE_32 = NSX->nsVar.spectrum.noiseSE1_32 ;

    /* 
      Since nsVar should be deleted, the two following lines
      becomes:
      nbFrame   = NSX->nsVar.vadNS.nbFrame [0];
      flagVAD   = NSX->nsVar.vadNS.flagVAD;
    */
    nbFrame   = nsVar.vadNS.nbFrame [0];
    flagVAD   = nsVar.vadNS.flagVAD;
    
    
    /*----------------------------------------------------------------------------
    * Spectral estimations of noisy signal are now stored in magnitude in nSigSE
    *----------------------------------------------------------------------------*/

    {
        for (i=0 ; i<NS_SPEC_ORDER ; i++) {
            nSigSE32[i] = Sqrt_2(nSigSE32[i]) ; move32() ;
            test();
            if(L_sub(PSDMeaned32[i] , 0x3fff0000) >= 0) {
                PSDMeaned16[i] = Sqrt16_2(L_shr(PSDMeaned32[i],2)) ; move16() ;
                PSDMeaned_dec[i] = sub(PSDMeaned_dec[i],1) ; move16();
            } else {
                PSDMeaned16[i] = Sqrt16_2(PSDMeaned32[i]) ; move16() ;
            }
        }
    }
    /*-------------------------------------------
    * VAD based Noise Estimation (in magnitude)
    *-------------------------------------------*/
    {
        X_INT16 M1_lambdaNSE_Fx2 ;
        test(); move16();
        if (sub(nbFrame,NS_NB_FRAME_THRESHOLD_NSE) < 0 ) {
            M1_lambdaNSE_Fx2 = M1_LambdaNSEx2[nbFrame] ;
        } else {
           M1_lambdaNSE_Fx2 = 2 * ROUND16((1 << 14 ) * (1 - NS_LAMBDA_NSE)) ;
        }
        test();
        if (flagVAD == 0){
            test(); sub(nbFrame,1);
            if(nbFrame == 1) {
                for (i=0 ; i<NS_SPEC_ORDER ; i++){
                    noiseSE_32[i] = L_mult(PSDMeaned16[i],1<<(SHFT_NOISE-1)); move32();
                    noiseSE_dec[i] = PSDMeaned_dec[i] ; move16();
                }
            } else { /* nbFrame == 1 */
                for (i=0 ; i<NS_SPEC_ORDER ; i++){
                    ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]);
                    {
                        X_INT16 difDec = sub(noiseSE_dec[i],PSDMeaned_dec[i]);
                        test();
                        if(difDec == 0) {
                            X_INT16 difh = round(L_msu(noiseSE_32[i],(1<<(SHFT_NOISE-1)),PSDMeaned16[i]));    
                            noiseSE_32[i] = L_msu(noiseSE_32[i],M1_lambdaNSE_Fx2,difh); move32();
                        } else {
                            test();
                            if (difDec > 0 )  {
                                noiseSE_32[i] = L_shr_r(noiseSE_32[i],difDec) ; move32();
                                noiseSE_dec[i] = PSDMeaned_dec[i] ; move16();
                                {
                                    X_INT16 difh = round(L_msu(noiseSE_32[i],(1<<(SHFT_NOISE-1)),PSDMeaned16[i]));    
                                    noiseSE_32[i] = L_msu(noiseSE_32[i],M1_lambdaNSE_Fx2,difh); move32();
                                }
                            } else {
                                X_INT16 difl = sub(shr_r(PSDMeaned16[i],sub(1,difDec)),round(L_shl(noiseSE_32[i],16-SHFT_NOISE-1)));
                                noiseSE_32[i] = L_mac(noiseSE_32[i],shr(M1_lambdaNSE_Fx2,1),difl); move32() ;
                            }
                        }
                    }
                }
            }
        }
    }
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: FilterCalc_F
 *
 * PURPOSE:     computes the Wiener filter (in frequency domain)
 *
 * INPUT:
 *   - NSX  : noise context struct 
 *   - PSDMeaned16 :  square root of power spectrum density
 *   - PSDMeaned_dec : shift factor for PSDMeaned32 : PSDMeaned32[i] = PSDMeaned[i] * 2**PSDMeaned_dec[i]
 *   - nSigSE32 : signal power spectrum 
 *
 * OUTPUT
 *   - W_16 : Wiener filter
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/

static void FilterCalc_F (
            X_INT16 fstage              /* (i) current stage */
            , NoiseSupStructX_F *NSX    /* (i/o) associated structure */
            , X_INT16 *W_16             /* (o) filter       */
            ,X_INT16 *PSDMeaned16       /* (i) square root of power spectrum density */
            ,X_INT16 *PSDMeaned_dec     /* (i) shift factor for PSDMeaned16 */
			, X_INT32 *nSigSE32
){
    /*--------------------------------------
    * noise suppression filter calculation
    *--------------------------------------*/
    {
        X_INT32 SNR_prio32 ;
        X_INT16 mSNR_post16 ;
        X_INT16 Wi_16 ;
        X_INT32 *  noiseSE_32 ;
        X_INT32 * denSigSE_32 ;
        X_INT16 * noiseSE_dec ;
        X_INT16 * denSigSE_dec ;
        X_INT16 * nSigSE_dec ;
        int i ;

        test();
        move16(); move16(); move16(); move16(); move16(); 
        if(fstage == 0) {
            denSigSE_32 = NSX->nsVar.spectrum.denSigSE1_32 ;
            denSigSE_dec = NSX->nsVar.spectrum.denSigSE1_dec ;
            noiseSE_32 = NSX->nsVar.spectrum.noiseSE1_32 ;
            noiseSE_dec = NSX->nsVar.spectrum.noiseSE1_dec ;
            nSigSE_dec = NSX->nsVar.spectrum.nSigSE1Cur_dec ;
        } else {
            denSigSE_32 = NSX->nsVar.spectrum.denSigSE2_32 ;
            denSigSE_dec = NSX->nsVar.spectrum.denSigSE2_dec ;
            noiseSE_32 = NSX->nsVar.spectrum.noiseSE2_32 ;
            noiseSE_dec = NSX->nsVar.spectrum.noiseSE2_dec ;
            nSigSE_dec = NSX->nsVar.spectrum.nSigSE2Cur_dec ;
        }


#define BETA16 ROUND16(4096.0 * NS_BETA)
#define M1_BETA16 ((X_INT16)(4096 - BETA16))
#define x8M1_BETA16 8*M1_BETA16
        for (i=0 ; i<NS_SPEC_ORDER ; i++){
            {
                X_INT16 difDecPSD_Noise = sub(noiseSE_dec[i],PSDMeaned_dec[i]);
                X_INT16 minDec ;
                test();
                if(difDecPSD_Noise>=0) {
                    mSNR_post16 = round(L_msu(L_shr(noiseSE_32[i],difDecPSD_Noise),PSDMeaned16[i],1<<(SHFT_NOISE-1)));
                    minDec = PSDMeaned_dec[i] ; move16();
                } else {
                    mSNR_post16 = round(L_msu(noiseSE_32[i],PSDMeaned16[i],shl((1<<(SHFT_NOISE-1)),difDecPSD_Noise))) ;
                    minDec = noiseSE_dec[i] ; move16();
                }
                test();
                if(mSNR_post16 <= 0) {
                    X_INT16 difDec = sub(minDec,denSigSE_dec[i]);
                    test() ;
                    if(difDec >= 0) {
                        X_INT16 hi = add(shr(mSNR_post16,difDec),extract_h(denSigSE_32[i]));
                        SNR_prio32 = L_msu(denSigSE_32[i],x8M1_BETA16,hi);
                        minDec = denSigSE_dec[i] ; move16();
                    } else {
                        X_INT32 sh_den_SigSE32 = L_shl(denSigSE_32[i],difDec) ;
                        X_INT16 hi =add(mSNR_post16,extract_h(sh_den_SigSE32));
                        SNR_prio32 = L_msu(sh_den_SigSE32,x8M1_BETA16,hi);
                    }
                    {
                        X_INT32 num = SNR_prio32 ;
                        X_INT32 den = L_add(SNR_prio32,L_shr(noiseSE_32[i],sub(noiseSE_dec[i],minDec))) ;
                        X_INT16 den16 = round(den) ;
                        test();
                        if(sub(den16,0x7ff) > 0)  {
                            Wi_16 = div_s(round(num),den16);
                        } else {
                            test();
                            if(den16 == 0) {
                                Wi_16 = 0 ; move16();
                            } else {
                                Wi_16 = div_s(round(L_shl(num,4)),round(L_shl(den,4)));
                            }
                        }
                    } 
                } else {
                    X_INT16 difDec = sub(noiseSE_dec[i],denSigSE_dec[i]);
                    X_INT16 hi = round(denSigSE_32[i]) ;
                    SNR_prio32 = L_msu(denSigSE_32[i],x8M1_BETA16,hi);
                    test();
                    if(difDec >= 0) {
                        const X_INT32 num = SNR_prio32 ;
                        X_INT32 den = L_add(SNR_prio32,L_shr_r(noiseSE_32[i],difDec)) ;
                        X_INT16 den16 = round(den) ;
                        test();
                        if(sub(den16,0x7ff) > 0)  {
                            Wi_16 = div_s(round(num),den16);
                        } else {
                            test() ;
                            if(den16 == 0) {
                                Wi_16 = 0 ; move16();
                            } else {
                                 Wi_16 = div_s(round(L_shl(num,4)),round(L_shl(den,4)));
                            }
                        }
                    } else {
                        X_INT16 num = shr_r(extract_h(SNR_prio32),-difDec) ;
                        X_INT16 den  = add(num,round(noiseSE_32[i])) ;
                        test();
                        if(den >= (0x4000 >>(-difDec-1)))  {
                            Wi_16 = div_s(num,den);
                        } else {
                            test() ;
                            if(den == 0) {
                                Wi_16 = 0 ; move16();
                            } else {
                                num = round(SNR_prio32) ;
                                den = add(num,round(L_shl(noiseSE_32[i],-difDec)));
                                Wi_16 = div_s(num,den);
                            }
                        }
                    }
                }
                W_16[i] = Wi_16; move16();
                SNR_prio32 = L_shr(L_mult(Wi_16,PSDMeaned16[i]),1);

            }
#define     Wi_MIN_M1 ROUND16( (1<<(SHFT_NOISE-1 )) * NS_RSB_MIN / (1 + NS_RSB_MIN))
#define     Wi_MIN_P1 ROUND16( (1<<(SHFT_NOISE+1 )) * NS_RSB_MIN / (1 + NS_RSB_MIN))
#define     Wi_MIN_P3 ROUND16( (1<<(SHFT_NOISE+3 )) * NS_RSB_MIN / (1 + NS_RSB_MIN))

            {
                X_INT32 num32   ;
                X_INT32 den32  ;
                X_INT16 den16 ;
                X_INT16 difDec = sub(noiseSE_dec[i],PSDMeaned_dec[i]);
                test();
                if(difDec >= 0) {
                    X_INT16 difDec1 = sub(difDec,1);
                    den32 = L_add(SNR_prio32,L_shr_r(noiseSE_32[i],difDec1)) ;
                    num32 = SNR_prio32 ; move32(); 
                } else {
                    num32 = L_shr_r(SNR_prio32,negate(difDec)) ;
                    den32 = L_add(L_add(num32,noiseSE_32[i]),noiseSE_32[i]) ;
                }
                den16 = round(den32) ;
                test();
                if(sub(den16,0x7ff) > 0) {
                    Wi_16 = div_s(round(num32),round(den32));
                } else {
                    test();
                    if(sub(den16,0x7f) > 0 )  {
                       Wi_16 = div_s(round(L_shl(num32,4)),round(L_shl(den32,4)));
                    }else {
                        test();
                        if(den16 == 0) {
                            Wi_16 = 0 ; move16();
                        } else {
                            test();
                            if(L_msu(num32,0x2000,extract_h(den32)) >= 0) {
                                Wi_16 = div_s(round(L_shl(num32,8)),round(L_shl(den32,8)));
                            } else {
                                Wi_16 = shr_r(div_s(round(L_shl(num32,10)),round(L_shl(den32,8))),2);
                            }
                        }
                    }
                }
                test(); sub(Wi_16, Wi_MIN_P1);
                if(Wi_16 < Wi_MIN_P1) {
                    W_16[i] = (X_INT16) ((Wi_MIN_P3  + 2)/4) ; move16(); 
                    denSigSE_32[i] = L_mult(Wi_MIN_P3,round(L_shl(nSigSE32[i],15))) ; move32();
                    denSigSE_dec[i]  =  add(3,nSigSE_dec[i]) ; move16();
                } else  {
                    W_16[i] = Wi_16 ; move16();
                    denSigSE_32[i] = L_mult(Wi_16,round(L_shl(nSigSE32[i],15))) ; move32();
                    denSigSE_dec[i]  =  add(1,nSigSE_dec[i]) ; move16();
                }
            }
        }
    }
    return;
}

static const X_INT16  M1_LamdaNSE[11-2] = {
        ROUND16((1/2.0) * 32768.0) ,
        ROUND16((1/3.0) * 32768.0) ,
        ROUND16((1/4.0) * 32768.0) ,
        ROUND16((1/5.0) * 32768.0) ,
        ROUND16((1/6.0) * 32768.0) ,
        ROUND16((1/7.0) * 32768.0) ,
        ROUND16((1/8.0) * 32768.0) ,
        ROUND16((1/9.0) * 32768.0) ,
        ROUND16((1/10.0) * 32768.0) 
    } ;


/*---------------------------------------------------------------------------
 * FUNCTION NAME: NoiseEstimation_F2
 *
 * PURPOSE:     computes noise estimation for second pass
 *
 * INPUT:
 *   - NSX  : noise context struct 
 *   - PSDMeaned32 : power spectrum mean
 *   - PSDMeaned_dec : shift factor for PSDMeaned32 : PSDMeaned32[i] = PSDMeaned[i] * 2**PSDMeaned_dec[i]
 *   - nSigSE32 : signal power spectrum 
 *
 * OUTPUT
 *   - PSDMeaned16 :  square root of power spectrum density
 *   - NSX updated
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void NoiseEstimation_F2(
            NoiseSupStructX_F *NSX  /* (i/o) associated structure */
            ,X_INT32 *PSDMeaned32   /* (i) power spectrum density mean */
            ,X_INT16 *PSDMeaned16   /* (o) square root of power spectrum density */
            ,X_INT16 *PSDMeaned_dec /* (i) shift factor for PSDMeaned16 */
			,X_INT32 *nSigSE32
) {
    /* 
       the following line is obviously a mistake - 
       NSX->nsVar should not be copied in the stack !! 
       I propose to delete this line.
    */
    NS_VAR_F    nsVar = NSX->nsVar;
    X_INT16   i;
    X_INT16   flagVAD;
    X_INT16   nbFrame;
    
    X_INT32 * noiseSE_32 ;
    X_INT32 * denSigSE_32 ;
    X_INT16 * nSigSE_dec ;
    X_INT16 * denSigSE_dec ;
    X_INT16 * noiseSE_dec = NSX->nsVar.spectrum.noiseSE2_dec ;
    nSigSE_dec = NSX->nsVar.spectrum.nSigSE2Cur_dec ;
    denSigSE_32 = NSX->nsVar.spectrum.denSigSE2_32 ;
    denSigSE_dec = NSX->nsVar.spectrum.denSigSE2_dec ;
    noiseSE_32 = NSX->nsVar.spectrum.noiseSE2_32 ;

    /* 
      Since nsVar should be deleted, the two following lines
      becomes:
      nbFrame   = NSX->nsVar.vadNS.nbFrame [0];
      flagVAD   = NSX->nsVar.vadNS.flagVAD;
    */
    nbFrame   = nsVar.vadNS.nbFrame [1];
    flagVAD   = nsVar.vadNS.flagVAD;
    
  /*--------------------------------
   * non VAD based Noise Estimation
   *--------------------------------*/
    // noise estimation in energy
    test();    
    if(sub(nbFrame, 11)< 0) {
        test();
        if(sub(nbFrame,1) == 0) {
            for (i=0 ; i<NS_SPEC_ORDER ; i++) {
                nSigSE32[i] = Sqrt_2(nSigSE32[i]) ; move32();
                test();
                if(L_sub(PSDMeaned32[i], 0x3fff0000) >= 0 ) {
                    PSDMeaned32[i] = L_shr(PSDMeaned32[i],2); move32();
                    PSDMeaned_dec[i] = sub(PSDMeaned_dec[i],1) ; move16();
                }
                PSDMeaned16[i] = Sqrt16_2(PSDMeaned32[i]) ; move16();
            }
            for (i=0 ; i<NS_SPEC_ORDER ; i++) {
                noiseSE_32[i] = L_mult(PSDMeaned16[i],(1<<(SHFT_NOISE-1))); 
                noiseSE_dec[i] = PSDMeaned_dec[i]; move16();
                ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]) ;
            }
        } else  { /* nbFrame != 1 */
                  /* prefer [nbFrame-2] instead of [sub(nbFrame,2)] */
            X_INT16 lambdaNSE16 =  M1_LamdaNSE[sub(nbFrame,2)]; move16();
            for (i=0 ; i<NS_SPEC_ORDER ; i++) {
                X_INT16 difDec = sub(noiseSE_dec[i],PSDMeaned_dec[i]);
                test();
                if(difDec == 0 ) {
                        X_INT16 n16 = round(noiseSE_32[i]);
                        X_INT32 x2noiseSE_32 = L_mult(n16,n16);
                        x2noiseSE_32 = L_mac(x2noiseSE_32,lambdaNSE16,
                                             sub(round(L_shr(PSDMeaned32[i],3)),mult_r(n16,n16)));
                        noiseSE_32[i] = L_msu(0,-32768,Sqrt16_2(L_shr(x2noiseSE_32,1))) ; move32();
                        ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]) ;
                } else {
                    test();
                    if (difDec > 0 )  {
                        X_INT16 n16 = round(L_shr(noiseSE_32[i],difDec)) ;
                        X_INT32 x2noiseSE_32 = L_mult(n16,n16);
                        x2noiseSE_32 = L_mac(x2noiseSE_32,lambdaNSE16,
                                             sub(round(L_shr(PSDMeaned32[i],3)),mult_r(n16,n16)));
                        noiseSE_32[i] = L_msu(0,0x8000,Sqrt16_2(L_shr(x2noiseSE_32,1))) ; move32();
                        noiseSE_dec[i] = PSDMeaned_dec[i] ; move16();
                        ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]) ;
                    } else  {
                        X_INT32 x2noiseSE_32 ;
                        X_INT16 n16 = round(noiseSE_32[i]) ;
                        test();
                        if(sub(n16,0x080) >= 0) {
                          sub(0,0); 
                          /* the following operation is non saturating,
                             prefer however -32768 to 32768!! */
                            x2noiseSE_32 =  L_mult((X_INT16)(32768 - lambdaNSE16),mult_r(n16,n16)) ;
                        } else  {
                            n16 = round(L_shl(noiseSE_32[i],8)) ;
                          sub(0,0); 
                          /* the following operation is non saturating,
                             prefer however -32768 to 32768!! */
                            x2noiseSE_32 = mult_r((X_INT16)(32768 - lambdaNSE16),mult_r(n16,n16)) ;
                        }
                        x2noiseSE_32 = L_add(x2noiseSE_32,L_shr(L_mult(lambdaNSE16,round(PSDMeaned32[i])),sub(3,add(difDec,difDec))));
                        noiseSE_32[i] = L_msu(0,0x8000,Sqrt16_2(L_shr(x2noiseSE_32,1))) ; move32();
                        ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]) ;
                    }
                } 
            }
            for (i=0 ; i<NS_SPEC_ORDER ; i++) {
                nSigSE32[i] = Sqrt_2(nSigSE32[i]) ; move32();
                test();
                if(L_sub(PSDMeaned32[i],0x3fff0000)>= 0) {
					PSDMeaned16[i] = Sqrt16_2(L_shr(PSDMeaned32[i],2)); move16();
                    PSDMeaned_dec[i] = sub(PSDMeaned_dec[i],1) ;    move16();
                } else {
					PSDMeaned16[i] = Sqrt16_2(PSDMeaned32[i]) ; move16();
				}
            }
        } /* else nbFrame == 1 */

    } else { /* nbFrame < 11 */
        X_INT16 ix16 ;
        for (i=0 ; i<NS_SPEC_ORDER ; i++) {
            nSigSE32[i] = Sqrt_2(nSigSE32[i]) ; move32();
            test();
            if(L_sub(PSDMeaned32[i],0x3fff0000)>= 0) {
				  PSDMeaned16[i] = Sqrt16_2(L_shr(PSDMeaned32[i],2)); move16();
                  PSDMeaned_dec[i] = sub(PSDMeaned_dec[i],1) ; move16();
            } else {
				PSDMeaned16[i] = Sqrt16_2(PSDMeaned32[i]) ; move16();
			}
        }
        for (i=0 ; i<NS_SPEC_ORDER ; i++) {
            X_INT16 val16 ;
            X_INT16 difDec2 ;
            ADJUST_SHFT(noiseSE_32[i],noiseSE_dec[i]) ;
            difDec2 = sub(sub(noiseSE_dec[i],PSDMeaned_dec[i]),2);
            test();
            if(difDec2 >= 0) {
                X_INT16 num = extract_h(L_shr(noiseSE_32[i],difDec2)) ;
                test();
                if(sub(num,PSDMeaned16[i]) < 0) {
                    X_INT16 den = PSDMeaned16[i] ;
                    test();
                    if(den <= 0) {
                       ix16 = 0 ;       move16();
                    } else {
                       test();
                       if(sub(num,den) >= 0) num = sub(den,1) ;
                       ix16 = div_s(num,den);
                    }
                    test();
                    if(sub(ix16,8192)<=0) {
                         val16 = add(add(mult_r(54784 - 32768,ix16),ix16),-99329+ 3 * 32768); 
                         val16 =  add(mult_r(val16,ix16),sub(38224 - 32768,add(ix16,add(ix16,ix16))))   ;
                         val16 = add(mult_r(val16,ix16),add(ix16,-268));
                         val16 = add(mult_r(val16,ix16),1);
                    } else {
                         val16 = add(mult_r(-6162,ix16),22784) ;
                         val16 = add(mult_r(val16,ix16),-31199) ;
                         val16 = add(mult_r(val16,ix16),15769) ;
                         val16 = add(mult_r(val16,ix16),64194-0x10000);
                    }
                } else {
                    test();
                    if(PSDMeaned16[i] == 0) {
                        ix16 = 0;               move16();
                    } else {
                        ix16 = div_s(PSDMeaned16[i],num) ;
                    }
                    test();
                    if(sub(ix16,16384) < 0) {
                         val16 = add(mult_r(-2050,ix16),-2978) ;
                         val16 = add(mult_r(val16,ix16),7471) ;
                         val16 = add(mult_r(val16,ix16),-36) ;
                         val16 = add(mult_r(val16,ix16),62173-0x10000);
                    } else {
                         val16 = add(mult_r(2348,ix16),-8514) ;
                         val16 = add(mult_r(val16,ix16),9592) ;
                         val16 = add(mult_r(val16,ix16),-163) ;
                         val16 = add(mult_r(val16,ix16),62124-0x10000);
                    }
                }
            } else { /* difDec2 >= 0 */
                X_INT16 num = round(noiseSE_32[i]) ;
                X_INT16 den = shr_r(PSDMeaned16[i],-difDec2) ;
                /* 
                   I understand that shl_r() could have been 
                   defined in ETSI basicops.
                */
                negate(difDec2);
                test();
                if(sub(num,den)<= 0) {
                    /* warning - the following variables num and den 
                       shadow already defined variables */
                    X_INT16 num = extract_h(L_shr(noiseSE_32[i],difDec2)) ;
                    X_INT16 den = PSDMeaned16[i] ;
                    test();
                    if(den <= 0) {
                        ix16 = 0 ; move16();
                    } else {
                        if(num >= den) num = sub(den,1);
                        ix16 = div_s(num,den);
                    }
                    test();
                    if(sub(ix16,8192) <= 0 ) {
                         val16 = add(add(mult_r(54784 - 32768,ix16),ix16),-99329+ 3 * 32768); 
                         val16 =  add(mult_r(val16,ix16),sub(38224 - 32768,add(ix16,add(ix16,ix16))))   ;
                         val16 = add(mult_r(val16,ix16),add(ix16,-268));
                         val16 = add(mult_r(val16,ix16),1);
                    } else {
                         val16 = add(mult_r(-6162,ix16),22784) ;
                         val16 = add(mult_r(val16,ix16),-31199) ;
                         val16 = add(mult_r(val16,ix16),15769) ;
                         val16 = add(mult_r(val16,ix16),64194-0x10000);
                    }
                } else {
                    ix16 = div_s(den,num);
                    test();
                    if(sub(ix16,16384) < 0 ) {
                         val16 = add(mult_r(-2050,ix16),-2978) ;
                         val16 = add(mult_r(val16,ix16),7471) ;
                         val16 = add(mult_r(val16,ix16),-36) ;
                         val16 = add(mult_r(val16,ix16),62173-0x10000);
                    } else {
                         val16 = add(mult_r(2348,ix16),-8514) ;
                         val16 = add(mult_r(val16,ix16),9592) ;
                         val16 = add(mult_r(val16,ix16),-163) ;
                         val16 = add(mult_r(val16,ix16),62124-0x10000);
                    }
                }
            } /* else difDec2 >= 0 */
            test();
            if(val16 >= 0) {
                noiseSE_dec[i] = sub(noiseSE_dec[i],1) ; move16() ;
                noiseSE_32[i] = L_shr(L_mac(noiseSE_32[i],val16,shr_r(extract_h(noiseSE_32[i]),1)),1); move32();
            } else {
                noiseSE_32[i] = L_add(noiseSE_32[i],L_shr(L_mult(val16,extract_h(noiseSE_32[i])),1)); move32();
            }
        }
    }
    return;
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoGainFact_F1
 *
 * PURPOSE:     computes Wiener filter gain  for first pass
 *
 * INPUT:
 *   - NSX  : noise context struct 
 *
 * OUTPUT
 *   - NSX updated (NSX->nsVar.gainFact)
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void DoGainFact_F1 (NoiseSupStructX_F *NSX)
{
    X_INT16 i;
    GAIN_FACT_F *gainFact = &(NSX->nsVar.gainFact);
    X_INT32 *denSigSE32 = NSX->nsVar.spectrum.denSigSE1_32 ;
    X_INT16 * DEN_Dec = NSX->nsVar.spectrum.denSigSE1_dec ;
    X_INT32 den2_32;
    X_INT16 m;

    den2_32 = 0 ; move32();
    m=32; move16();
    for (i=0 ; i<NS_SPEC_ORDER ; i++) {
        test();
        if(sub(DEN_Dec[i],m)  < 0) {
            m = DEN_Dec[i] ; move16();
        }
    }
    m = sub(m,4) ;
    for (i=0 ; i<NS_SPEC_ORDER ; i++) {
        den2_32 = L_add(den2_32,L_shr(denSigSE32[i],sub(DEN_Dec[i],m)));
    }
    gainFact->logDenEn1_32[0] = gainFact->logDenEn1_32[1] ; move32();
    gainFact->logDenEn1_32[1] = gainFact->logDenEn1_32[2] ; move32();
    gainFact->logDenEn1_32[2] = L_msu(L_deposit_l(Log_2(den2_32)),512,m) ; move32();
}

 static const X_INT16  mInvLambda16[NS_MIN_FRAME] = {
        0,
        2*ROUND16((1/1.0) * 16384.0) ,
        2*ROUND16((1/2.0) * 16384.0) ,
        2*ROUND16((1/3.0) * 16384.0) ,
        2*ROUND16((1/4.0) * 16384.0) ,
        2*ROUND16((1/5.0) * 16384.0) ,
        2*ROUND16((1/6.0) * 16384.0) ,
        2*ROUND16((1/7.0) * 16384.0) ,
        2*ROUND16((1/8.0) * 16384.0) ,
        2*ROUND16((1/9.0) * 16384.0) 
    } ;


/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoGainFact_F2
 *
 * PURPOSE:     computes Wiener filter gain  for second pass
 *
 * INPUT:
 *   - NSX  : noise context struct 
 *
 * OUTPUT
 *   - NSX updated (NSX->nsVar.gainFact)
 * RETURN VALUE none
 *   
 *
 *---------------------------------------------------------------------------*/
static void DoGainFact_F2 ( 
        NoiseSupStructX_F *NSX  /* (i/o) associated structure */
      , X_INT16 *W_16           /* (i/o) filter */
){
    X_INT16 i;
    GAIN_FACT_F *gainFact = &(NSX->nsVar.gainFact);
    X_INT32 *noiseEN_32 = NSX->nsVar.spectrum.noiseSE2_32 ;
    X_INT16 * noise_Dec = NSX->nsVar.spectrum.noiseSE2_dec ;
    X_INT32 nois2_32;
    X_INT32 nois2log ;
    X_INT32 logAvrSNR_32 ;
    X_INT16 m;
    X_INT16 m4 ;
    X_INT16 nbFrame ;

    nois2_32 = 0 ; move32();
    m = 32 ; move16();
    for (i=0 ; i<NS_SPEC_ORDER ; i++) {
        test();
        if(sub(noise_Dec[i], m) < 0) {
            m = noise_Dec[i] ; move16();
        }
    }
    m4 =sub(m,4);
    for (i=0 ; i<NS_SPEC_ORDER ; i++) {
        nois2_32 = L_add(nois2_32,L_shr(noiseEN_32[i],sub(noise_Dec[i],m4)));
    }
    nois2log = L_msu(L_deposit_l(Log_2(nois2_32)),512,m); 
    logAvrSNR_32 = L_add(L_add(gainFact->logDenEn1_32[0],gainFact->logDenEn1_32[1]),gainFact->logDenEn1_32[2]);
	logAvrSNR_32 = L_sub(logAvrSNR_32,nois2log);
	logAvrSNR_32 = L_sub(logAvrSNR_32,nois2log);
	logAvrSNR_32 = L_sub(logAvrSNR_32,nois2log);
#define THRESHOLD_LogAvr    (X_INT32)(-100.0 * 1024 / 20.0 * log(10.0) / log(2.0))
    test();
    if(L_sub(logAvrSNR_32,THRESHOLD_LogAvr) < 0) {
        logAvrSNR_32 = THRESHOLD_LogAvr ; move32();
    }
    nbFrame = NSX->nsVar.vadNS.nbFrame[1] ;
#define THRESHOLD_difAvr (X_INT32)(10.0 * 1024 / 20.0 * 3.0 * log(10.0) / log(2.0) + 1.0)
    test(); test();
    if ( (L_sub(L_sub(logAvrSNR_32,gainFact->lowSNRtrack32), THRESHOLD_difAvr) < 0) || (sub(nbFrame,NS_MIN_FRAME)< 0) ){
        X_INT16 lambda16 ;
        test();
        if (sub(nbFrame,NS_MIN_FRAME)<0) {
	  test();
          if(sub(nbFrame,1)==0) {
	     gainFact->lowSNRtrack32 = logAvrSNR_32 ;
          } else {
	     lambda16 = mInvLambda16[nbFrame] ; move16();
             gainFact->lowSNRtrack32 = L_add(gainFact->lowSNRtrack32,
	     mult_r(lambda16,extract_l(L_sub(logAvrSNR_32,gainFact->lowSNRtrack32)))) ; move32();
	    }		

        } else  {
            test();
            if (L_sub(logAvrSNR_32,gainFact->lowSNRtrack32)<0) {
                lambda16 = 2*ROUND16(16384 * 0.05) ; move16();
            } else {
                lambda16 = 2*ROUND16(16384 * 0.01) ; move16();
            }
            gainFact->lowSNRtrack32 = L_add(gainFact->lowSNRtrack32,
            mult_r(lambda16,extract_l(L_sub(logAvrSNR_32,gainFact->lowSNRtrack32)))) ; move32();
        }
    }
    test();  
    if(L_sub(gainFact->logDenEn1_32[2],(X_INT32)((log(100.0) /log(2.0) + SHFT_NOISE - 4) * 1024 )) > 0 )
    {
        test(); sub(sub(logAvrSNR_32,gainFact->lowSNRtrack32),0);
        if( (logAvrSNR_32 - gainFact->lowSNRtrack32 ) < 3.5 * 1024 / 20.0 * 3.0 * log(10.0) / log(2.0)) {
            gainFact->alfaGF16 = add(gainFact->alfaGF16, ROUND16(0.15 * GAIN_FACT_MUL)); move16();
            test();
            if (sub(gainFact->alfaGF16,ROUND16(0.8 * GAIN_FACT_MUL)) > 0) {
                gainFact->alfaGF16 = ROUND16(0.8 * GAIN_FACT_MUL);  move16();
            }
        }else {
            gainFact->alfaGF16 = sub(gainFact->alfaGF16, ROUND16(0.3 * GAIN_FACT_MUL));  move16();
            test();
            if (sub(gainFact->alfaGF16,ROUND16(0.1 * GAIN_FACT_MUL)) < 0 ) {
                gainFact->alfaGF16 = ROUND16(0.1 * GAIN_FACT_MUL); move16();
            }
        }

    }
    {
        X_INT16 alfaGF16 = add(gainFact->alfaGF16,gainFact->alfaGF16)  ;
        X_INT16 ofsW16 = shr(extract_l(L_msu(1 <<(SHFT_FACT_MUL+SHFT_NOISE-13),gainFact->alfaGF16,1<<(SHFT_NOISE-14))),1);
        for (i=0 ; i<WF_MEL_ORDER ; i++) {
            W_16[i] = add(ofsW16,mult(alfaGF16,W_16[i])) ; move16() ;
        }
    }
}


/*------------------------
* start Functions for VAD
*------------------------*/

/*----------------------------------------------------------------------------
 * FUNCTION NAME: SpeechQSpec
 *
 * PURPOSE:       Voice Activity Detection for Frame Dropping
 *              
 * INPUT:          FEParamsX *  Pointer to frontend parameter structure   
 *
 * OUTPUT:        VAD frame dropping structure updated
 *
 *
 * RETURN VALUE:
 *   none
 *
 *---------------------------------------------------------------------------*/

static X_INT16 SpeechQSpec (FEParamsX_F *This)
{
    NoiseSupStructX_F *NSX = This->NSX;
    VAD_DATA_FD *vadFD = &(NSX->nsVar.vadFD);
    const X_INT16 FrameCounter = This->NSX->nsVar.vadNS.nbFrame[0];
    X_INT32 InvAcceleration; // == 1/Acceleration in float version
    X_INT16 tSpecVal;
    X_INT16 AccTestIntern;
    X_INT32 tSpecVal_L, tmp, tmpSpecMean;
    
    test();
    if (sub(FrameCounter,15) < 0)
    {
        // Scale of SpecMean , SpecValues == 2^20
        test();
        if (sub(FrameCounter,1) == 0) {
            vadFD->SpecMean = vadFD->SpecValues;    move32();
        }
        
        // Scale of tSpecVal , AccTest == 2^16
        tSpecVal = extract_h(L_shl(vadFD->SpecValues, 12)); 

        // calc 1.1 * (AccTest * (FC-1) + SpecVal ) / FC
        sub(FrameCounter,1);
        tmp = L_add(L_shr(L_mult(vadFD->AccTest, (FrameCounter - 1)) ,1) , (X_INT32)tSpecVal);
        
        tmp = L_shr(L_mult_SL(lambda_1divX[FrameCounter-1], tmp) ,6); // +15 +1 -6 == shl(10)
        
        tmp = L_mult_SL(35, L_shr(tmp, 6));  // (*35/32) ~= * 1.1
        
        vadFD->AccTest = extract_h(L_shl(tmp,6));      // shr(10) --> AccTest == 2^16

        // check if ratio is greater 1 
        AccTestIntern = (X_INT16)vadFD->AccTest; move16();
        test();
        if (sub(tSpecVal,AccTestIntern) >= 0)
        {
            // ratio < 1 so div_s may be used
            InvAcceleration = div_s(AccTestIntern, tSpecVal);
            // 1/2.5 * 2^15 = 13107
            // if InvAccelration < 1/2.5
            test() ;
            if (sub(InvAcceleration,13107) < 0 )
            {
                vadFD->SpeechInVADQ = 1;  move16(); 
            }
        }
        test();
        if (vadFD->SpeechInVADQ == 0)
        {
          test();
            if(L_sub(vadFD->SpecValues,vadFD->SpecMean) > 0) {
                vadFD->SpecMean = vadFD->SpecValues; move32();
            }
        }
    }
    tSpecVal_L = L_shl(vadFD->SpecValues, 10); // compensate Scaling of constants
    // in the following comparisons
    tmpSpecMean = L_shr(vadFD->SpecMean, 1);                                                  
    // Scaling of all constants is 2^10
    test(); test();
    if (L_sub(tSpecVal_L ,L_mult_SL(1536, tmpSpecMean)) < 0 && \
        L_sub(tSpecVal_L ,L_mult_SL( 768, tmpSpecMean)) > 0 )
    {
        vadFD->SpecMean = L_add(L_mult_SL(819, tmpSpecMean), \
            L_mult_SL(205, L_shr(vadFD->SpecValues,1) ));
        tmpSpecMean = L_shr(vadFD->SpecMean, 11);
    }
    test();
    if (L_sub(tSpecVal_L, L_mult_SL(512, tmpSpecMean)) <= 0)
    {
        vadFD->SpecMean = L_add(L_mult_SL(993, tmpSpecMean), \
            L_mult_SL(31, L_shr(vadFD->SpecValues,1) ));
        tmpSpecMean = L_shr(vadFD->SpecMean, 11);
    }
    
    vadFD->SpecMean = L_shl(tmpSpecMean, 1);
    test();
    if (L_sub(tSpecVal_L,L_mult_SL(1690, tmpSpecMean)) > 0 )
    {
        return 1;
    }
    else
    {
        return 0;
    }
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME: SpeechQMel
 *
 * PURPOSE:       Voice Activity Detection for Frame Dropping
 *              
 * INPUT:         FEParamsX_F *  Pointer to frontend parameter structure   
 *
 * OUTPUT:        VAD frame dropping structure updated
 *
 *
 * RETURN VALUE:
 *   none
 *
 *---------------------------------------------------------------------------*/
static X_INT16 SpeechQMel (FEParamsX_F *This)
{
    NoiseSupStructX_F *NSX = This->NSX;
    VAD_DATA_FD *vadFD = &(NSX->nsVar.vadFD);
    const X_INT16 FrameCounter = This->NSX->nsVar.vadNS.nbFrame[0];
    X_INT16   SmoothMel;
    X_INT16   MM;
    
    SmoothMel = add(mult(6144, vadFD->MelValues[1]) ,mult(2048, vadFD->MelValues[0]));
    SmoothMel = add(mult(12288, vadFD->MelValues[1]) ,mult(4096, vadFD->MelValues[0]));
    
    // MelValues 2^13 , Constants == 2^14 --> SmoothMel == 2^12
    vadFD->MelValues[0] = vadFD->MelValues[1]; move16();
    test();
    if (sub(FrameCounter,15) < 0 )
    {
        test();
        if(sub(vadFD->MelMean,SmoothMel) < 0) {
            vadFD->MelMean = SmoothMel; move16();
        }
    }
    MM = shl(vadFD->MelMean,2); 
    test(); test();
    if (sub(SmoothMel,mult(12288, MM)) < 0  && sub(SmoothMel,mult(6144, MM)) > 0 )
    {
        vadFD->MelMean = add( mult(26216, vadFD->MelMean), mult(6552, SmoothMel));
    } 
    
    MM = shl(vadFD->MelMean,2); 
    test();
    if (sub(SmoothMel,mult(4096, MM)) <= 0)
    {
        vadFD->MelMean = add( mult(31784, vadFD->MelMean), mult(984, SmoothMel));
    } 
    MM = shl(vadFD->MelMean,2); 
    test();
    if (sub(SmoothMel,mult(26624, MM)) > 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }     
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME: SpeechQVar
 *
 * PURPOSE:       Voice Activity Detection for Frame Dropping
 *              
 * INPUT:         FEParamsX *  Pointer to frontend parameter structure
 *                *W           Pointer to noise suppression filter  
 *
 * OUTPUT:        VAD frame dropping structure updated
 *
 *
 * RETURN VALUE:
 *   none
 *
 *---------------------------------------------------------------------------*/
static X_INT16 SpeechQVar (FEParamsX_F *This, X_INT16 *W)
{
    NoiseSupStructX_F *NSX = This->NSX;
    VAD_DATA_FD *vadFD = &(NSX->nsVar.vadFD);
    X_INT16 i, tmpW; 
    const X_INT16 ssize = NS_FFT_LENGTH / 4;
    const X_INT16 FrameCounter = This->NSX->nsVar.vadNS.nbFrame[0];
    X_INT32 var;
    X_INT32 mean, tmpMean;
    X_INT32 specVar, tmpSpecVar;
    
    var = 0L; move32();
    mean = 0L; move32();
    specVar = 0L; move32();

    for (i=0 ; i<ssize ; i++)
    {
        tmpW = shr(W[i],8); 
        mean = L_add(tmpW, mean) ;   // mean is scaled by 2^13
        var  = L_mac(var, tmpW, tmpW); // var is scaled by 2^21
    }
    specVar = L_sub(L_shr(var,1) , L_shr(L_mult(mean,  mean),7));
    test();  
    if (sub(FrameCounter,15) < 0 )
    {
        L_sub(vadFD->VarMean, specVar);
        vadFD->VarMean = max (vadFD->VarMean, specVar); test(); move32();
    }
    
    // scale up for comparison
    tmpSpecVar = L_shl(specVar, 10); 
    tmpMean = L_shr(vadFD->VarMean, 1);
    
    test(); test();
    if (L_sub(tmpSpecVar,L_mult_SL(1536, tmpMean)) < 0 && \
        L_sub(tmpSpecVar,L_mult_SL( 870, tmpMean)) > 0 )
    {
        vadFD->VarMean = L_add(L_mult_SL(819, tmpMean), L_mult_SL(205, L_shr(specVar,1)) );move32();
        // after multiplication; scale == 2^30, but for varMean scale == 2^20 is used
        tmpMean = L_shr(vadFD->VarMean,11);
    }
    test();
    if (L_sub(tmpSpecVar,L_mult_SL(256, tmpMean)) <= 0)    
    {
        vadFD->VarMean = L_add(L_mult_SL(993, tmpMean), L_mult_SL(31, L_shr(specVar,1)) );move32();
        // after multiplication; scale == 2^30, but for varMean scale == 2^20 is used
        tmpMean = L_shr(vadFD->VarMean,11);
    }
    
    vadFD->VarMean = L_shl(tmpMean ,1);move32();
    test();
    if (L_sub(tmpSpecVar,L_mult_SL(1690, tmpMean)) > 0)    
    {
        return 1;
    }
    else
    {
        return 0;
    }
} 
/*------------------------
* End Functions for VAD
*------------------------*/

/*----------------------------------------------------------------------------
 * FUNCTION NAME: UpDateDecal
 *
 * PURPOSE:       Compute global shift factor for a subframe
 *              
 * INPUT:         globalDecal: current global shift factor
 *                curDecal : shift factor for the subframe
 *                indMin : first index (relative to buffer32) for the subframe
 *                indMax : last index (relative to buffer32)for the subframe
 *                buffer32 : the buffer for the subFrame
 *
 * OUTPUT:       
 *
 *
 * RETURN VALUE:  
 *  the updated value for global shift
 *
 *---------------------------------------------------------------------------*/

static X_INT16 UpDateDecal(X_INT16 globalDecal,X_INT16 curDecal,X_INT16 indMin,X_INT16 indMax, X_INT32 *buffer32){
    /* Warning - there is a risk that the macro max() and the variable max
       are mixed up by the precompiler */
    X_INT16 max = shl(1,curDecal) ;
	X_INT16 i ;
    for(i=indMin;i<indMax;i++) {
        X_INT16 vabs = abs_s(round(buffer32 [i])) ;
        max = max |  vabs ; logic16();
    }
    {
        X_INT16 decal0;
        decal0 = curDecal; move16();
        test(); logic16();
        if(max & 0x4000) {
            decal0= sub(decal0,1) ;
        } else {
            test(); logic16();
            while( (max & (0xE000 )) == 0) {
                decal0 = add(decal0,1) ;
                max =shl(max,1) ;
                test(); logic16();
            }
        }
        test();
        if(sub(decal0,globalDecal) < 0 ) {
           globalDecal = decal0 ; move16();
        }
    }
	return globalDecal ;
}

/*----------------------------------------------------------------------------
 * FUNCTION NAME: ApplyDecal
 *
 * PURPOSE:       Apply the  global shift factor for a subframe
 *              
 * INPUT:         globalDecal: the global shift factor
 *                curDecal : shift factor for the subframe
 *                indMin : first index (relative to buffer32) for the subframe
 *                indMax : last index (relative to buffer32)for the subframe
 *                buffer32 : the buffer for the input subFrame
 *                bufferOut16 : the buffer for the output subFrame (with applied global shift)
 *
 * OUTPUT:       
 *
 *
 * RETURN VALUE:  
 *   none
 *
 *---------------------------------------------------------------------------*/

static void ApplyDecal(X_INT16 globalDecal,X_INT16 curDecal,X_INT16 indMin,X_INT16 indMax, X_INT32 *buffer32,X_INT16 *bufferOut16)
{
    X_INT16 shft16,i;
    shft16 = sub(16-SHFT_NOISE-1,sub(curDecal,globalDecal)) ;
    for(i=indMin;i<indMax;i++) {
        bufferOut16[i] =  round(L_shl(buffer32[i],shft16)); move16();
    }
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoNoiseSupAlloc_B
 *
 * PURPOSE:  Allocation of a noise suppression struct
 *
 * INPUT: 
 *
 * OUTPUT:
 *
 * RETURN VALUE: pointer to the new allocated struct
 *   
 *              
 *
 *---------------------------------------------------------------------------*/
extern NoiseSupStructX_F *DoNoiseSupAlloc_B (void)
{
    NoiseSupStructX_F *NSX = calloc (1, sizeof (NoiseSupStructX_F));
    
    if (NSX == NULL)
        return NULL;
    return NSX;
}




/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoNoiseSupInit_B 
 *
 * PURPOSE:  Initialisation of a noise suppression struct
 *
 * INPUT:
 *    - This : global FEParamsX_F context (for this->NSX)
 *
 * OUTPUT:
 *    - initialized this->NSX
 *
 * RETURN VALUE: none
 *   
 *              
 *---------------------------------------------------------------------------*/

extern void DoNoiseSupInit_B (FEParamsX_F *This)
{
    X_INT16 i;
    NoiseSupStructX_F *NSX = This->NSX;
    NSX->nsVar.SampFreq    = (X_INT16)This->SamplingFrequency;
#if defined(F_16KHZ)
    NSX->nsVar.Do16kHzProc = This->Do16kHzProc;
#endif  

    /*---------
    * buffers
    *---------*/
    for (i=0 ; i<NS_BUFFER_SIZE ; i++)
    {
        NSX->nsVar.buffers.FirstStageIn16Buffer [i]  = 0;
        NSX->nsVar.buffers.SecondStageInBuffer32 [i] = 0;
    }
    for(i=0;i<(NS_BUFFER_SIZE+2*NS_FRAME_SHIFT-1)/NS_FRAME_SHIFT;i++) {
        NSX->nsVar.buffers.SecondDecalSig[i]  = 0 ;
    }
    NSX->nsVar.buffers.nbFramesInFirstStage   = 0;
    NSX->nsVar.buffers.nbFramesInSecondStage  = 0;
    NSX->nsVar.buffers.nbFramesOutSecondStage = 0;
    
    /*-----------
    * dc_filter
    *-----------*/

    NSX->nsVar.prevSamples32.lastSampleIn32 = 0;
    NSX->nsVar.prevSamples32.lastDCOut32    = 0;

    /*-----------
    * gain_fact
    *-----------*/
    NSX->nsVar.gainFact.lowSNRtrack32 = 0 ;
    NSX->nsVar.gainFact.alfaGF16 = ROUND16(0.8 * GAIN_FACT_MUL) ;
    
    /*-------------
    * vad_data_ns
    *-------------*/
    NSX->nsVar.vadNS.nbFrame [0]    = NSX->nsVar.vadNS.nbFrame [1] = 0;
    NSX->nsVar.vadNS.hangOver       = 0;
    NSX->nsVar.vadNS.nbSpeechFrames = 0;
    NSX->nsVar.vadNS.meanEn32 = 0 ;
    NSX->nsVar.vadNS.flagVAD        = 0;
    
    NSX->nsVar.vadCA.hangOver       = 0;
    NSX->nsVar.vadCA.nbSpeechFrames = 0;
    NSX->nsVar.vadCA.meanCA         = 0;
    NSX->nsVar.vadCA.flagVAD        = 0;
    
    /*-------------
    * vad_data_fd
    *-------------*/
    NSX->nsVar.vadFD.MelMean = 0;
    NSX->nsVar.vadFD.VarMean = 0;
    NSX->nsVar.vadFD.AccTest = 0;
    NSX->nsVar.vadFD.AccTest2 = 0;
    NSX->nsVar.vadFD.SpecMean = 0;
    NSX->nsVar.vadFD.MelValues[0] = NSX->nsVar.vadFD.MelValues[1] = 0;
    NSX->nsVar.vadFD.SpecValues = 0;
    NSX->nsVar.vadFD.SpeechInVADQ = 0;
    NSX->nsVar.vadFD.SpeechInVADQ2 = 0;
    
    /*----------
    * spectrum
    *----------*/
    for (i=0 ; i<NS_SPEC_ORDER ; i++)
    {
        
        /* required as smallest shift needs to be taken */ 
        NSX->nsVar.spectrum.nSigSE1Ant_dec[i] = 8 ;
        NSX->nsVar.spectrum.PSDMeanAntBuffer1[i] = 0 ;
        NSX->nsVar.spectrum.PSDMeanAntBuffer2[i] = 0 ;

        NSX->nsVar.spectrum.denSigSE1_32[i] = 0 ;
        NSX->nsVar.spectrum.denSigSE1_dec[i] = 0 ;
    }
    
    NSX->nsVar.spectrum.indexBuffer1 = 0;
    NSX->nsVar.spectrum.indexBuffer2 = 0;
 }


/*----------------------------------------------------------------------------
* FUNCTION NAME: DoNoiseSupDelete_B
*
* PURPOSE: free noise suppression struct
*              
*
* INPUT:     - NSX : noise suppression struct
*
* OUTPUT: N/A
*
*
* RETURN VALUE: none1
*   
*
*---------------------------------------------------------------------------*/
extern void DoNoiseSupDelete_B (NoiseSupStructX_F *NSX)
{
    
    if (NSX != NULL)
    {
        free (NSX);
    }
}





/*----------------------------------------------------------------------------
 * FUNCTION NAME: DoNoiseSup_B
 *
 * PURPOSE: Noise suppression for a frame 
 *              
 *
 * INPUT:
 *      - InData16,   input signal frame
 *      - shiftIn,  Shift factor for DataIn[i]  DataIn[i] =Data[i] * 2**shiftIn
 *      - This  : global FEParamsX_F context (for this->NSX)                
 *
 * OUTPUT:
 *      - OutData32  :   denoised output signal
 *      - shftOut    :   shift factor for OutData32 : 0utData32[i] = OutData[i] *  2**shftOut 
 *
 * RETURN VALUE: TRUE if signal is output
 *               FALSE if no signal is output (first buffers initalisation)
 *   
 *
 *---------------------------------------------------------------------------*/

#if defined(XAFE)
extern BOOLEAN  DoNoiseSup_B (X_INT16 *InData16, X_INT32 *OutData32, FEParamsX_F *This, X_INT16 *shftOut, Shortword *pswStft, Shortword *pswSignal, X_INT16 *piStftShift, X_INT16 *pidecalSig, X_INT16 *piProcessFlag)
#else
extern BOOLEAN  DoNoiseSup_B (X_INT16 *InData16, X_INT32 *OutData32, FEParamsX_F *This, X_INT16 *shftOut)
#endif

{
    X_INT16 scratchDoNoise[512];
#if defined(F_16KHZ)
    X_INT16 bufData16kSize;          
	X_INT32 fb16k_B[5] ; 
    X_INT16 fb16k_dec;
    X_INT32 *bufferData16k_B;       
    X_INT32 *CodeForBands16k_B;    
    X_INT32 *BandsForCoding16k_B;
    DataFor16kProc_B * pData16k;       
#endif

    X_INT16 i; 
    X_INT16 fstage;                      // noise suppression filter stage
    
    NoiseSupStructX_F *NSX = This->NSX;
    
    X_INT16 *FirstStageIn16Buffer  = NSX->nsVar.buffers.FirstStageIn16Buffer;
    X_INT32 *SecondStageInBuffer32 = NSX->nsVar.buffers.SecondStageInBuffer32;
    
    X_INT32 *nbFramesInFirstStage     = &(NSX->nsVar.buffers.nbFramesInFirstStage);
    X_INT32 *nbFramesInSecondStage    = &(NSX->nsVar.buffers.nbFramesInSecondStage);
    X_INT32 *nbFramesOutSecondStage   = &(NSX->nsVar.buffers.nbFramesOutSecondStage);
    
    X_INT16 *indexBuffer = NULL;   
    
    X_INT16 *nSigSE_dec = NULL ;
    X_INT32 *PSDAnt32 = NULL ;
    X_INT16 *PSDAnt_dec = NULL ;
    
    X_INT32 *signalOut32 ;
    
    X_INT32 *signalIn32 = (X_INT32 *) scratchDoNoise ;
    
    X_INT32 *PSDMeaned32 = signalIn32 ;
    X_INT16 PSDMeaned_dec[NS_SPEC_ORDER] ;
    X_INT32 W_32[NS_HALF_FILTER_LENGTH+1] ;
    X_INT16 W_16[NS_SPEC_ORDER] ;
    
	X_INT32 nSigSE32[NS_SPEC_ORDER] ;
    BOOLEAN dataToProcess;           // data to process ?
    X_INT16 tmpCurFrame16[NS_FRAME_SHIFT+NS_FILTER_LENGTH-1] ;
    X_INT16 *curFrame16 = tmpCurFrame16 + (NS_FILTER_LENGTH / 2) ;
    
    X_INT16 decalSig = 0 ;
	X_INT16 *signalIn16 = (X_INT16 *) (signalIn32 + NS_FFT_LENGTH /2 );
    move16(); /* decalSig */
    /* I assume here that indexBuffer, nSigSE_dec, PSDant32 and 
       PSDant_dec do not really need to be initialized, otherwize, 
       move16();move16();move16();move16();move16(); would be needed.
       */

#if defined(F_16KHZ)
    /*------------------
    * 16kHz Processing
    *------------------*/
    if (NSX->nsVar.Do16kHzProc)
    {
        pData16k          = (DataFor16kProc_B *)This->pData16k;
        bufferData16k_B     = Get16k_p_bufferData16k_B     (pData16k);
        bufData16kSize    = Get16k_bufData16kSize_B      (pData16k);
        BandsForCoding16k_B = Get16k_p_BandsForCoding16k_B (pData16k);
        CodeForBands16k_B   = Get16k_p_CodeForBands16k_B   (pData16k);
    }
    else
    {
        pData16k          = NULL;
        bufferData16k_B     = NULL;
        bufData16kSize    = 0;
        BandsForCoding16k_B = NULL;
        CodeForBands16k_B   = NULL;
    }
#endif

#if defined(XAFE)
    *piProcessFlag = FALSE;
#endif

    test(); test(); sub(0,0); sub(0,NS_NB_FRAMES_LATENCY);
    if ((*nbFramesInFirstStage - *nbFramesInSecondStage) < NS_NB_FRAMES_LATENCY)
    {
        for (i=0 ; i<NS_DATA_IN_BUFFER; i++){
            FirstStageIn16Buffer [i] = FirstStageIn16Buffer [i + NS_FRAME_SHIFT]; move16();
        }
		/*---------------------------------------------------
		 * input next shift frame in FirstStageInFloatBuffer
		 *---------------------------------------------------*/
		for (i=0 ; i<NS_FRAME_SHIFT ; i++) {
			FirstStageIn16Buffer [NS_DATA_IN_BUFFER + i] = InData16[i]; move16();
		}
	}
    
    (*nbFramesInFirstStage)++; add(*nbFramesInFirstStage,1);

    /*------------------------------------
    * Two-Stage Noise Suppression Filter
    *------------------------------------*/
    for (fstage=0 ; fstage<2 ; fstage++) 
    {
        dataToProcess = FALSE; move16();
        /*----------------------------------------------------
        * 1st stage reads data from FirstStageInFloatBuffer
        * 2nd stage reads data from SecondStageInFloatBuffer
        *----------------------------------------------------*/
        test(); test(); sub(0,0);sub(0,NS_NB_FRAMES_LATENCY);
        if ((fstage == 0) && ((*nbFramesInFirstStage - *nbFramesInSecondStage) > NS_NB_FRAMES_LATENCY)) {
        /*-------------------
         * process 1st stage
         *-------------------*/
            dataToProcess = TRUE; move16();
            nSigSE_dec = NSX->nsVar.spectrum.nSigSE1Cur_dec ;move16();
            PSDAnt32 = NSX->nsVar.spectrum.PSDMeanAntBuffer1 ;move16();
            PSDAnt_dec = NSX->nsVar.spectrum.nSigSE1Ant_dec ;move16();
            indexBuffer = &(NSX->nsVar.spectrum.indexBuffer1);move16();
			signalOut32 = signalIn32 ;move16();
 		
			for (i=0 ; i<NS_FRAME_SHIFT ; i++) {
                signalIn16[i] = FirstStageIn16Buffer [i]; move16();
            }
			// copy the content of FirstStageInFloatBuffer [80..180[ to [0..100[
			for (i= 0 ; i<NS_DATA_IN_BUFFER; i++){
				FirstStageIn16Buffer [i] = FirstStageIn16Buffer [i + NS_FRAME_SHIFT]; move16();
			}
			/*---------------------------------------------------
			 * input next shift frame in FirstStageInFloatBuffer
			 *---------------------------------------------------*/
			for (i=0 ; i<NS_FRAME_SHIFT ; i++) {
				FirstStageIn16Buffer [NS_DATA_IN_BUFFER + i] = InData16[i]; move16();
			}
   
			for (i= 0; i<NS_FRAME_LENGTH-NS_FRAME_SHIFT ; i++) {
                signalIn16[i+NS_FRAME_SHIFT] = FirstStageIn16Buffer [i]; move16();
            }
        } else  if ((fstage == 1) && ((*nbFramesInSecondStage - *nbFramesOutSecondStage) > NS_NB_FRAMES_LATENCY)){
        /*-------------------
         * process 2nd stage
         *-------------------*/
            dataToProcess = TRUE;move16();
            
            indexBuffer = &(NSX->nsVar.spectrum.indexBuffer2);move16();
            nSigSE_dec = NSX->nsVar.spectrum.nSigSE2Cur_dec ;move16();
            PSDAnt32 = NSX->nsVar.spectrum.PSDMeanAntBuffer2 ;move16();
            PSDAnt_dec = NSX->nsVar.spectrum.nSigSE2Ant_dec ;move16();
            signalOut32 = OutData32 ;move16();
        }
        if (!dataToProcess) continue; // no processing required
        /*-----------------------------------
         * signal windowing and zero padding
         *-----------------------------------*/
        {
            int i ;
            /* saturation limits to 0x7fff */
            NSX->nsVar.vadNS.nbFrame [fstage] = add(NSX->nsVar.vadNS.nbFrame [fstage],1); move16();
            test();
            if(fstage == 0) {
              /* Warning - there is a risk that the macro max() 
                 and the variable max are mixed up by the precompiler */
                X_INT16 max;
                max = 1; move16();
                decalSig = 0 ; move16();
                for(i=0;i<NS_FRAME_LENGTH;i++) {
                    max = max | abs_s(signalIn16[i]) ; logic16();
                }
                test(); logic16();
                while ( ( max & 0xC000u ) == 0) {
                    decalSig = add(decalSig,1) ;
                    max =shl(max,1) ;
                    test(); logic16();
                }


#if defined(XAFE)

            test(); test(); sub(*nbFramesInFirstStage,NS_NB_FRAMES_LATENCY);
            if ((fstage == 0) && (*nbFramesInFirstStage > NS_NB_FRAMES_LATENCY))
            {

                /*
                * signalIn16[] contains the input signal used by the
                * pitch extractor. Copy this to the appropriate buffer.
                */

                for (i = 0; i < NS_FRAME_LENGTH; i++)
                {
                   pswSignal[i] = signalIn16[i];		move16();
                }

                *pidecalSig = decalSig;				move16();

            }

#endif

                VAD_F (&(NSX->nsVar.vadNS),signalIn16+NS_CUR_FRAME -NS_ANALYSIS_WINDOW_8K,decalSig);
                //# Output : curFrame16[i] = signalIn * 2**decalSig 
                for(i= -NS_FILTER_LENGTH/2 ;i<NS_FRAME_SHIFT+NS_FILTER_LENGTH/2 ; i++) {
                    curFrame16[i] = shl(signalIn16 [i- NS_ANALYSIS_WINDOW_8K +NS_FRAME_SHIFT],decalSig); move16();
                }
                //# Output : signalIn32[i] = sigWindow[i] * 2**(decalSig+8)
                DoSigWindowing16_F1 (signalIn16,signalIn32,decalSig, NS_FFT_LENGTH);
            } else {
                for(i= -NS_FILTER_LENGTH/2 ;i<NS_FRAME_SHIFT+NS_FILTER_LENGTH/2 ; i++) {
                    //# Output : curFrame16[i] = signalIn * 2**decalSig 
                    curFrame16[i] =  signalIn16 [i- NS_ANALYSIS_WINDOW_8K + NS_FRAME_SHIFT] ; move16();
                }
                //# Output : signalIn32[i] = sigWindow[i] * 2**(decalSig+9)
                DoSigWindowing16_F2 (signalIn16,signalIn32, NS_FFT_LENGTH);
            }
            ff4NRFix32_B(8,signalIn32);


#if defined(XAFE)
            test(); test(); sub(*nbFramesInFirstStage,NS_NB_FRAMES_LATENCY);
            if ((fstage == 0) && (*nbFramesInFirstStage > NS_NB_FRAMES_LATENCY))
            {
                X_INT16 iShift;
                Longword lwMax;

                /*
                * signalIn32[] contains the FFT scaled by 2**(decalSig+8)
                * in the order:
                *    DC, PI, Real(1), Imag(1), ..., Real(127), Imag(127).
                *
                * Copy the FFT into Stft in a form suitable for the
                * pitch extractor
                */

                lwMax = L_abs(signalIn32[1]);
                for (i = 2; i <= (NS_FFT_LENGTH-1); i++)
                {
                    lwMax = (lwMax | L_abs(signalIn32[i])); logic32();
                }
                iShift = norm_l(lwMax);
                iShift = sub(iShift,1);

                pswStft[0] = 0; move16();
                pswStft[1] = 0; move16();

                for (i = 2; i < (NS_FFT_LENGTH-1); i+= 2)
                {
                    pswStft[i] = round(L_shl(signalIn32[i],iShift));
								move16();
                    pswStft[i+1] = negate(round(L_shl(signalIn32[i+1],iShift))); move16();
                }

                pswStft[NS_FFT_LENGTH] = round(L_shl(signalIn32[1],iShift));
								move16();
                pswStft[NS_FFT_LENGTH+1] = 0;			move16();

                *piStftShift = add(sub(decalSig,8),iShift);	move16();
                *piProcessFlag = TRUE;				move16();

            }

#endif /* defined(XAFE) */
            

            /* represents the 1/2 shift of nSigSE32 with respect to floating point version */ 
            //#  Output : stage 0 : nSigSE32[i]= nSigSE[i] * 2**(2*nSigSE_dec[i]) , nSigSE_dec[i] is computed for each bin
            //#  Output : stage 1 : nSigSE32[i]= nSigSE[i] * 2**(2*(nSigSE_dec[i]+1)) , nSigSE_dec[i] is computed for each bin
            test();
            if(fstage == 0) {
                FFTtoPSD_F (signalIn32, nSigSE32,nSigSE_dec,decalSig);
            } else {
                sub(decalSig,1);
                FFTtoPSD_F (signalIn32, nSigSE32,nSigSE_dec,decalSig-1);
            }
        } 
#if defined(F_16KHZ)
       /*---------------------------------
        * coding for 16kHz Processing
        *---------------------------------*/
        test(); /* test to Do16kHwProc is for free */
        if (NSX->nsVar.Do16kHzProc && (fstage == 0)) { 
            int i ;
            X_INT32 aux[3];
            X_INT16 * decal = Get16k_BFC_dec_B(pData16k);
            GetBandsForCoding16k_B (nSigSE32, nSigSE_dec, aux, decal);
            for (i=0 ; i<6 ; i++) {
                BandsForCoding16k_B [i] = BandsForCoding16k_B [i+3]; move32();
            }
            
            for (i=0 ; i<3 ; i++)
            {
                test();
                if (aux [i] > 0) {  //NS_SPEC_FLOOR)
                    BandsForCoding16k_B [6 + i] = L_sub(Log_2 (aux[i]), shl(*decal,10)); move32();
                } else {
                    BandsForCoding16k_B [6 + i] = (NS_LOG_SPEC_FLOOR*1477.32);  move32();
                }
            }
        }
#endif      

        /*---------
        * PSDMean
        *---------*/
        //# stage Output : stage 0 PSDMeaned32[i] =  PSDMean[i] * 2**(2*PSDMeaned_Dec[i])
        //# stage Output : stage 1 PSDMeaned32[i] =  PSDMean[i] * 2**(2*(PSDMeaned_Dec[i]+1))
        PSDMean_F(nSigSE32,nSigSE_dec,PSDMeaned32,PSDMeaned_dec,PSDAnt32,PSDAnt_dec); 
        /*--------------------------
        * filter gains calculation
        *--------------------------*/
        { 
            X_INT16 PSDMeaned16[NS_SPEC_ORDER] ;
            test();
            if(fstage == 0) {
                //# Output :
                //# PSDMeaned16[i] =  PSDMean[i] * 2**PSDMeaned_Dec[i]
                //# noiseSE_32[i]  = noiseSE[i] * 2**(noiseSE_dec[i]+SHFT_NOISE)
                NoiseEstimation_F1(NSX,PSDMeaned32,PSDMeaned16,PSDMeaned_dec,nSigSE32) ;
            } else {
                //# Output :
                //# PSDMeaned16[i] =  PSDMean[i] * 2**(PSDMeaned_Dec[i]+1)
                //# noiseSE_32[i]  = noiseSE[i] * 2**(noiseSE_dec[i]+1+SHFT_NOISE)
                NoiseEstimation_F2(NSX,PSDMeaned32,PSDMeaned16,PSDMeaned_dec,nSigSE32) ;
            }
            //# Output : 
            //# W_16[i] = W[i] * 2**15 
            FilterCalc_F (fstage, NSX, W_16,PSDMeaned16,PSDMeaned_dec,nSigSE32);

            /*------------------------
            * VAD for Frame Dropping
            *------------------------*/
            test();
            if (fstage == 0) {
                This->SpeechFoundVar = SpeechQVar (This, W_16);
            }              
            /*-----------------
            * mel filter bank
            *-----------------*/
            //#  Output :
            //#   W_16[i] = W[i] * 2**15
            FilterBank16(W_16,W_16);
            /*------------------------
            * VAD for Frame Dropping
            *------------------------*/
            // This->SpeechFoundMel = SpeechQMel (This);
            test();
            if (fstage == 0)
            {
                X_INT16 TempEn;
                X_INT32 TempEn32 = L_mult(0,0);
                TempEn   = 0; move16();
                for(i=0 ; i<WF_MEL_ORDER ; i++) {
                    L_deposit_l(W_16[i]);
                    TempEn32 = L_add(TempEn32, (X_INT32)W_16[i]);  // max < ~2^5*2^14
                }
                test();
                if (L_sub(TempEn32,131071) > 0 ) { // 2^17
                    TempEn = extract_l(L_shr(TempEn32,7)); // max < 2^12
                    NSX->nsVar.vadFD.SpecValues = L_shr(L_sub(L_mult(TempEn, TempEn), 98304L),5);
                    // Scale down by 5 to avoid overflow inb SpeechQspec() new scale = 2^20 
                    // 98304L = 3 * 2^14 remark:  scale of TempEn = 2^7 --> Scale of TempEn^2 = 2^14
                }
                else
                {
                    // better scaling for low values
                    TempEn = extract_l(L_shr(TempEn32,3));
                    NSX->nsVar.vadFD.SpecValues = L_shr(L_sub(L_mult(TempEn, TempEn), 25165824L),13);
                    // 25165824L = 3 * 2^22
                }
                
                This->SpeechFoundSpec = SpeechQSpec (This); move16();
                
                // W(i) < 2^14 --> MelValues < k13
                NSX->nsVar.vadFD.MelValues[1] = round(L_mac(L_mac(L_mult(W_16[1],0x1000),0x1000,W_16[2]),0x1000,W_16[3])); move16();
                This->SpeechFoundMel = SpeechQMel (This); move16();
            }  

            /*--------------------
            * gain factorization
            *--------------------*/
            test();
            if(fstage == 0) {
                DoGainFact_F1 (NSX);
            } else {
                DoGainFact_F2 ( NSX, W_16);
            }
            /*-----------------
            * mel inverse DCT
            *-----------------*/
            //# Output :
            //#  W_32[i]  = W[i] * 2**30
            DoMelIDCT_F16 (W_16,W_32,idctMel16) ;
            //#  Output : signalOut32[i] = signalOut[i] * 2**(SHFT_NOISE+1+decalSig) 
            ApplyWF (curFrame16,W_32, signalOut32,NS_FRAME_SHIFT, NS_HALF_FILTER_LENGTH);
		}
#if defined(F_16KHZ)
        /*------------------
        * 16kHz Processing 
        *------------------*/
        test();
        if (NSX->nsVar.Do16kHzProc && (fstage > 0))
        {
            X_INT16 tmpdec, diffdec;
			X_INT16 *nSig16kSE_dec = (X_INT16 *) scratchDoNoise;
			X_INT32 * tmpIn32 = (X_INT32 *) scratchDoNoise ;
            X_INT16 wnd16k_dec1 = Get16k_dec1(pData16k);
            X_INT16 wnd16k_dec2 = Get16k_dec2(pData16k);
            X_INT16 wnd16k_dec3 = Get16k_dec3(pData16k);
            
            X_INT32 intFilter32[5];
            
            X_INT16 wnd_dec = wnd16k_dec1; move16();
            test();
            if(sub(wnd_dec,wnd16k_dec2) > 0) {
                wnd_dec = wnd16k_dec2;          move16();
            }
            test();
            if(sub(wnd_dec,wnd16k_dec3) > 0) {
                wnd_dec = wnd16k_dec3;          move16();
            }
            diffdec = sub(wnd16k_dec3,wnd_dec); 
            test();
            if(diffdec > 0 ) {
                for(i=0;i<NS_FRAME_SHIFT-40;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-NS_FRAME_SHIFT] =L_shr(bufferData16k_B[i+NS_FRAME_LENGTH+40-NS_FRAME_SHIFT],diffdec) ; move32();
                }
            } else {
                for(i=0;i<NS_FRAME_SHIFT-40;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-NS_FRAME_SHIFT] =bufferData16k_B[i+NS_FRAME_LENGTH+40-NS_FRAME_SHIFT] ; move32();
                }
			}
            diffdec = sub(wnd16k_dec2,wnd_dec); 
            test();
            if(diffdec > 0 ) {
                for(i=0;i<NS_FRAME_SHIFT;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-2*NS_FRAME_SHIFT] =L_shr(bufferData16k_B[i+NS_FRAME_LENGTH+40-2*NS_FRAME_SHIFT],diffdec) ; move32();
                }
            } else {
                for(i=0;i<NS_FRAME_SHIFT;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-2*NS_FRAME_SHIFT] =bufferData16k_B[i+NS_FRAME_LENGTH+40-2*NS_FRAME_SHIFT] ; move32();
                }
			}
            diffdec = sub(wnd16k_dec1,wnd_dec); 
            test();
            if(diffdec > 0 ) {
                for(i=0;i<NS_FRAME_SHIFT;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-3*NS_FRAME_SHIFT] =L_shr(bufferData16k_B[i+NS_FRAME_LENGTH+40-3*NS_FRAME_SHIFT],diffdec) ; move32() ;
                }
            } else {
                for(i=0;i<NS_FRAME_SHIFT;i++) {
                    tmpIn32[i+NS_FRAME_LENGTH+40-3*NS_FRAME_SHIFT] =bufferData16k_B[i+NS_FRAME_LENGTH+40-3*NS_FRAME_SHIFT] ; move32() ;
                }
			}
            DoSigWindowing16_F3(tmpIn32, tmpIn32, NS_FFT_LENGTH);
            
            ff4NRFix32_B(8,tmpIn32);
            
            FFTtoPSD_F (tmpIn32, nSigSE32,nSig16kSE_dec,wnd_dec+10-8);
            
            tmpdec = nSig16kSE_dec[0]; move16();
            for(i=1; i<NS_SPEC_ORDER; i++){
                test();
                if(sub(tmpdec,nSig16kSE_dec[i]) > 0) {
		  tmpdec = nSig16kSE_dec[i]; move16();
                }
            }
            for(i=0; i<NS_SPEC_ORDER; i++){
                diffdec = sub(nSig16kSE_dec[i], tmpdec);
                nSigSE32[i] = L_shr_r(nSigSE32[i], shl(diffdec,1)); move32();
                nSig16kSE_dec[i] = tmpdec; move16();
            }
            
            DoMelFB_B (nSigSE32, intFilter32);
            
            tmpdec = sub(shl(tmpdec,1), 5);
            
            for(i=0; i<5; i++) {
                fb16k_B[i] = intFilter32[i]; move32();
            }
            
            for (i=0 ; i<HP16k_MEL_USED ; i++){
                intFilter32 [i] = intFilter32 [i+1]; move32();
                fb16k_B[i] = fb16k_B[i+1]; move32();
            }
            
            for (i=0 ; i<HP16k_MEL_USED ; i++)
            {
                test();
                if (L_sub(intFilter32 [i],NS_SPEC_FLOOR)>0) {
                    intFilter32 [i] = L_sub(Log_2 (intFilter32 [i]), shl(tmpdec,10)); move32();
                } else {
                    intFilter32 [i] = (NS_LOG_SPEC_FLOOR*1477.32); move32();
                }
            }
            
            CodeBands16k_B (intFilter32, BandsForCoding16k_B, CodeForBands16k_B);
            
	    DoSpecSub16k_B (fb16k_B, &tmpdec, pData16k, NSX->nsVar.vadNS.nbFrame [1], HP16k_MEL_USED);
	    
	    fb16k_dec = tmpdec; move16();
	    
        }
#endif  /* defined(F_16KHZ) */    

 
        /*---------------------------------------------------
         * update buffer with filtered signal for next stage
         *---------------------------------------------------*/
        test();
        if (fstage == 0)
        {
            //# Output :    SecondStageInBuffer32[i] = signalOut[i] * 2**(SHFT_NOISE+1+decalSig)
            //#             SecondDecalSig = decalSig 
            NSX->nsVar.buffers.SecondDecalSig[(NS_DATA_IN_BUFFER+2*NS_FRAME_SHIFT-1)/NS_FRAME_SHIFT] = decalSig ; move16();
            NSX->nsVar.buffers.nbFramesInSecondStage++; add(0,1); move16();
			{
				decalSig = 14 ; move16();
				decalSig = UpDateDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[0]
					,0,NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32) ;
				decalSig = UpDateDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[1]
					,NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,2*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32) ;
				decalSig = UpDateDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[2]
					,2*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,3*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32) ;
				decalSig = UpDateDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[3]
					,0,NS_FRAME_LENGTH-3*NS_FRAME_SHIFT+NS_ANALYSIS_WINDOW_8K,signalOut32) ;
				//# Output : signalIn16[i] = SecondStageInBuffer32[j] * 2**(-curDecal + decalSig)
                //#                        = signalOut[i] * 2**(SHFT_NOISE+1+curDecal - curDecal + decalSig-SHFT_NOISE-1)
                //#                        = signalOut[i] * 2**(decalSig)
				ApplyDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[0]
					,0,NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32,signalIn16) ;
				ApplyDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[1]
					,NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,2*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32,signalIn16) ;
				ApplyDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[2]
					,2*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,3*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,SecondStageInBuffer32,signalIn16) ; 
				ApplyDecal(decalSig,NSX->nsVar.buffers.SecondDecalSig[3]
					,3*NS_FRAME_SHIFT-NS_ANALYSIS_WINDOW_8K,NS_FRAME_LENGTH,signalOut32-3*NS_FRAME_SHIFT+NS_ANALYSIS_WINDOW_8K,signalIn16) ;
			}
			for (i=0 ; i<NS_DATA_IN_BUFFER ; i++) {
					SecondStageInBuffer32 [i] = SecondStageInBuffer32 [i + NS_FRAME_SHIFT]; move32();
			}
			for (i=0;i<NS_FRAME_SHIFT ; i++) {
					SecondStageInBuffer32 [i+NS_DATA_IN_BUFFER] = signalOut32 [i]; move32();
			}

			for (i=0 ; i<(NS_DATA_IN_BUFFER+2*NS_FRAME_SHIFT-1)/NS_FRAME_SHIFT ; i++) {
				NSX->nsVar.buffers.SecondDecalSig[i] = NSX->nsVar.buffers.SecondDecalSig[i+1] ; move16();
			}
        

        } else  {
            NSX->nsVar.buffers.nbFramesOutSecondStage++;add(0,1); move16();
        }
            
        /*------------------------
        * VAD for Frame Dropping
        *------------------------*/
        test();
        if (fstage == 0)
        {
            test();
            if (sub(NSX->nsVar.vadNS.nbSpeechFrames,4)> 0) {
                This->SpeechFoundVADNS = 1; move16();
            } else {
                This->SpeechFoundVADNS = 0; move16();
            }
        }
        This->FrameCounter = This->NSX->nsVar.vadNS.nbFrame[0];  move16();
    } // loop fstage 0 1
    
    /*--------------------------------------------
    * shift data in NSX->FirstStageInFloatBuffer
    *--------------------------------------------*/
    test();
    if (NSX->nsVar.buffers.nbFramesInFirstStage)
    {
#if defined(F_16KHZ)        
        if (NSX->nsVar.Do16kHzProc)
        {
            for (i=0 ; i<(bufData16kSize-NS_FRAME_SHIFT) ; i++)
                bufferData16k_B [i] = bufferData16k_B [i + NS_FRAME_SHIFT]; move32();
        }
#endif  
    }
    
    test();
    if (NSX->nsVar.buffers.nbFramesOutSecondStage > 0)
    {
        //#  Output : signalOut32[i] = signalOut[i] * 2**(SHFT_NOISE+1+decalSig) 
        //          DCOffsetFil (OutData, &(NSX->nsVar.prevSamples), NS_FRAME_SHIFT);
        DCOffsetFil_F(signalOut32,&(NSX->nsVar.prevSamples32),NS_FRAME_SHIFT,decalSig);
       *shftOut = SHFT_NOISE-1 + decalSig ; add(SHFT_NOISE-1,decalSig);
#if defined(F_16KHZ)
        /*------------------
        * 16kHz Processing 
        *------------------*/
        if (This->Do16kHzProc)
        {
            X_INT16 i;
            X_INT16 hpBandsSize;
            X_INT32 *hpBands_B ;
            X_INT32 *CodeForBands16k_B, *bufferCodeForBands16k_B;
            X_INT16 *bufferCodeWeights_B;
            X_INT16 *codeWeights_B;
            DataFor16kProc_B * pData16k =  (DataFor16kProc_B *)This->pData16k ;
            hpBandsSize           = Get16k_hpBandsSize_B (pData16k);
            hpBands_B               = Get16k_p_hpBands_B (pData16k);
            bufferCodeForBands16k_B = Get16k_p_bufferCodeForBands16k_B (pData16k);
            CodeForBands16k_B       = Get16k_p_CodeForBands16k_B (pData16k);
            bufferCodeWeights_B     = Get16k_p_bufferCodeWeights_B (pData16k);
            codeWeights_B           = Get16k_p_codeWeights_B (pData16k);
            
            //                          WARNING: 16kHz Processing DM                          
            // CurrentFrame [FrameShift : (FrameShift + HP16k_MEL_USED - 1)] contains HP bands
            for (i=HP16k_MEL_USED ; i<hpBandsSize ; i++) {
                hpBands_B [i - HP16k_MEL_USED] = hpBands_B [i]; move32();
            } 
            
            for (i=0 ; i<HP16k_MEL_USED ; i++) {
                hpBands_B [hpBandsSize - HP16k_MEL_USED + i] =fb16k_B [i] ; move32();
            }
            Set16k_hpBands_dec_B(pData16k, fb16k_dec);
    
            for (i=0 ; i<(2*HP16k_MEL_USED*NB_LP_BANDS_CODING) ; i++) {
                bufferCodeForBands16k_B [i] = bufferCodeForBands16k_B [i + HP16k_MEL_USED * NB_LP_BANDS_CODING]; move32();
            }
            
            for (i=0 ; i<(HP16k_MEL_USED*NB_LP_BANDS_CODING) ; i++) {
                bufferCodeForBands16k_B [(2 * HP16k_MEL_USED * NB_LP_BANDS_CODING) + i] = CodeForBands16k_B [i]; move32();
            }
            
            for (i=0 ; i<(2*NB_LP_BANDS_CODING) ; i++) {
                bufferCodeWeights_B [i] = bufferCodeWeights_B [i + NB_LP_BANDS_CODING]; move16();
            }
            
            for (i=0 ; i<NB_LP_BANDS_CODING ; i++) {
                bufferCodeWeights_B [(2 * NB_LP_BANDS_CODING) + i] = codeWeights_B [i]; move16();
            }
        }
#endif          
        return TRUE;
    }
    else {
        return FALSE;
    }
}

 

/*
 * For XAFE
 */

#if defined(XAFE)

/*
 * Static variables
 */

static const Shortword pswPePower[NS_FFT_LENGTH/2+1] = {
8,
13,
27,
52,
86,
131,
185,
249,
322,
406,
498,
601,
713,
834,
965,
1105,
1254,
1413,
1580,
1757,
1942,
2136,
2338,
2549,
2768,
2996,
3231,
3475,
3726,
3985,
4251,
4524,
4805,
5093,
5388,
5689,
5996,
6310,
6630,
6956,
7287,
7624,
7967,
8314,
8666,
9023,
9384,
9750,
10119,
10493,
10869,
11250,
11633,
12019,
12408,
12799,
13192,
13587,
13984,
14383,
14782,
15183,
15584,
15986,
16388,
16790,
17192,
17593,
17993,
18393,
18791,
19188,
19583,
19977,
20368,
20757,
21143,
21526,
21906,
22283,
22656,
23026,
23391,
23753,
24109,
24462,
24809,
25151,
25488,
25820,
26145,
26465,
26779,
27087,
27388,
27683,
27970,
28251,
28525,
28791,
29050,
29301,
29544,
29780,
30007,
30227,
30438,
30640,
30834,
31019,
31195,
31363,
31521,
31670,
31810,
31941,
32063,
32175,
32277,
32370,
32453,
32527,
32591,
32645,
32689,
32724,
32748,
32763,
32767 };

static X_INT16 iFirstFrameFlag = TRUE;
static Shortword pswUBSpeech[FRAME_LEN];
static Shortword pswDownSampledProcSpeech[(FRAME_LEN+HISTORY_LEN)/4];

#define		NUM_CHAN	23


/*
 * API function
 */

extern BOOLEAN DoPitchExtract_B(Shortword *pswStft,
                                Shortword *pswSignal,
                                X_INT16 iStftShift,
                                X_INT16 decalSig,
                                X_INT16 iProcessFlag,
                                FEParamsX_F *This)
{

  RVC_PITCH_METER_be *pstMeter;
  Longword *plwPowerSpectrum;
  Longword *plwPowerSpectrumTemp;
  Longword *plwPEPowerSpectrum;
  Longword *plwMFBOut;
  Shortword *pswTempBuf;
  Shortword *pswMFBOutArray;

  Longword lwSpecAverage;
  Longword lwMax;
  Longword lwSumSq;
  Longword lwTemp;
  Longword lwPitchPeriod;

  Shortword swSum;
  Shortword swTemp1;
  Shortword swTemp2;

  X_INT16 scratchDoPitch[SIZE_SCRATCH_DOPITCH] ;
  X_INT16 i;
  X_INT16 iMFBOutShift;
  X_INT16 iHangOverFlag;
  X_INT16 iLowBandNoiseFlag;
  X_INT16 iVad;
  X_INT16 iClass;
  X_INT16 iShift;
  X_INT16 iShift1;
  X_INT16 iShift2;
  X_INT16 iRC;
  X_INT16 iZcm;

  NormSw nsFrameEnergy;

  float fPitchPeriod;
  X_CHAR cVoicingClass;

  static BOOLEAN IsLowBandNoise(Longword *plwPEPowerSpectrum,
                                BOOLEAN iVad,
                                NormSw nsFrameEnergy);


  /*
   * Return if no processing is required
   */

  if (iProcessFlag == FALSE)
  {
    return TRUE;
  }


  /*
   * Use scratch memory for local arrays - Pointer ops.
   * (A total of 3*129*2 + 23*2 + 23*1 = 843 INT16 words are used here)
   */

  pstMeter = This->pRvcPitchMeter;
  pstMeter->pvScratchMemory = scratchDoPitch;
  
  plwPowerSpectrum = (Longword *)(pstMeter->pvScratchMemory);
  plwPowerSpectrumTemp = plwPowerSpectrum + (NS_FFT_LENGTH/2+1);
  plwPEPowerSpectrum = plwPowerSpectrumTemp + (NS_FFT_LENGTH/2+1);
  plwMFBOut = plwPowerSpectrumTemp + (NS_FFT_LENGTH/2+1);
  pswMFBOutArray = (Shortword *)(plwMFBOut + NUM_CHAN);


  /*
   * Set the pointers for holding spectral peaks using the scratch
   * memory from DoAdvProcess() - Pointer ops.
   */

  pstMeter->pstSpectralPeaks = (POINT_be *) pswStft;move16();
  pstMeter->pstPoints = pstMeter->pstSpectralPeaks + MAX_PEAKS_FINAL;move16();


  /*
   * Compute the spectral average
   */

  lwTemp = L_deposit_l(shr(pswStft[NS_FFT_LENGTH],1));
  for (i = 2; i < NS_FFT_LENGTH; i += 2)
  {
    lwTemp = L_add(lwTemp,L_deposit_l(pswStft[i]));
  }
  lwSpecAverage = L_shl(lwTemp,9); 


  /*
   * Compute the power spectrum
   */

  plwPowerSpectrum[0] = 0;				move32();
  for (i = 1; i < NS_FFT_LENGTH/2; i++)
  {
    plwPowerSpectrum[i] = L_mac(L_mult(pswStft[2*i],pswStft[2*i]),
                                       pswStft[2*i+1],pswStft[2*i+1]);
							move32();
  }
  plwPowerSpectrum[NS_FFT_LENGTH/2] = 
                L_mult(pswStft[NS_FFT_LENGTH],pswStft[NS_FFT_LENGTH]);
							move32();


  /*
   * Compute the frame energy
   */

  iShift1 = sub(decalSig,4);
  iShift2 = sub(decalSig,8);
  swTemp1 = shl(pswSignal[0],iShift1);
  lwSumSq = L_mult(swTemp1,swTemp1);
  swSum = shl(pswSignal[0],iShift2);
  for (i = 1; i < NS_FRAME_LENGTH; i++)
  {
    swTemp1 = shl(pswSignal[i],iShift1);
    lwSumSq = L_mac(lwSumSq,swTemp1,swTemp1);
    swTemp2 = shl(pswSignal[i],iShift2);
    swSum = add(swSum,swTemp2);
  }

  lwTemp = L_mult(164,round(L_mult(swSum,swSum)));
  lwTemp = L_shl(lwTemp,8);
  lwSumSq = L_sub(lwSumSq,lwTemp);

  iShift = norm_l(lwSumSq);
  nsFrameEnergy.swMantissa = round(L_shl(lwSumSq,iShift));
  nsFrameEnergy.iShift = add(add(shl(iShift1,1),iShift),8);


  /*
   * Voice Activity Detection
   */

  for (i = 0; i <= NS_FFT_LENGTH/2; i++)
  {
    plwPowerSpectrumTemp[i] = L_shr(plwPowerSpectrum[i],3);	move32();
  }

  FilterBank(plwPowerSpectrumTemp,plwMFBOut);

  lwMax = plwMFBOut[0]; move32();
  for (i = 1; i < 23; i++)
  {
    lwMax = (lwMax | plwMFBOut[i]); logic32();
  }
  iShift = norm_l(lwMax);
  for (i = 0; i < 23; i++)
  {
    pswMFBOutArray[i] = round(L_shl(plwMFBOut[i],iShift));	move16();
  }
  iMFBOutShift = add(add(shl(iStftShift,1),iShift),4);

  iVad = dsr_afe_vad(pswMFBOutArray,iMFBOutShift,&iHangOverFlag);


  /*
   * Low-band noise detection
   */

  plwPEPowerSpectrum[0] = 0; move32();
  for (i = 1; i <= NS_FFT_LENGTH/2; i++)
  {
    plwPEPowerSpectrum[i] = L_add(L_mult(pswPePower[i],
          extract_h(plwPowerSpectrum[i])),L_shr(L_mult(pswPePower[i],
          extract_l(L_shr(plwPowerSpectrum[i],1) & 0x7FFF)),15));  move32();
  }

  iLowBandNoiseFlag = IsLowBandNoise(plwPEPowerSpectrum,iVad,nsFrameEnergy);  


  /*
   * Get the zero crossing measure from the input signal (to be used
   * in classification later)
   */

  iZcm = get_zcm(pswSignal);


  /*
   * Pre-processing
   */

	pswTempBuf = ((X_INT16 *)pstMeter->pvScratchMemory) + NS_FFT_LENGTH+2;
  test(); logic16();
  if (iFirstFrameFlag == TRUE)
  {
    pre_process(pswSignal,iFirstFrameFlag,iLowBandNoiseFlag,pswUBSpeech,
                pswDownSampledProcSpeech,pswTempBuf);
    iFirstFrameFlag = FALSE; move16();
  }

  pre_process(pswSignal,iFirstFrameFlag,iLowBandNoiseFlag,pswUBSpeech,
                pswDownSampledProcSpeech,pswTempBuf);


  /*
   * Pitch extraction
   * (Note: The scratch memory from the local function uses only 258
   * INT16 words corresponding to plwPowerSpectrum at this point. The
   * scratch memory from DoAdvProcess() also uses 258 INT16 words
   * corresponding to pswStft at this point).
   */

  iRC = RVC_MeasurePitch_be(This->pRvcPitchRom,
                This->pRvcPitchMeter,lwSpecAverage,
                pswDownSampledProcSpeech+(HISTORY_LEN/DOWN_SAMP_FACTOR),
                pswStft,iStftShift+16,plwPowerSpectrum,0,
                nsFrameEnergy,iVad,iLowBandNoiseFlag,&lwPitchPeriod);

  if (iRC != RVC_OK)
  {
    fprintf (stderr,"ERROR: pitch estimation error RC = %d\r\n",iRC);
    return FALSE;
  }


  /*
   * Class extraction
   */

  iClass = classify_frame(iVad,iHangOverFlag,lwPitchPeriod,nsFrameEnergy,
                          iZcm,pswUBSpeech);


  /*
   * Output pitch
   */

  /* store output to file  */
  fPitchPeriod = (float)lwPitchPeriod / 65536.0;
  if (1 != fwrite(&fPitchPeriod,sizeof(float),1,This->pitch_fid))
  {
    fprintf (stderr,"ERROR: can't write pitch to file \r\n");
    return FALSE;
  }


  /*
   * Output class
   */
  /* transfer the classing voice to the VQ module */
  test();test();test();test();move16();
  switch (iClass)
  {
    case 0: // non-speech
      cVoicingClass = '0';
      break;
    case 1: // unvoiced speech
      cVoicingClass = '1';
      break;
    case 2: // mixed-voiced speech
      cVoicingClass = '2'; 
      break;
    case 3: // voiced speech
      cVoicingClass = '3';
      break;
    default:
      fprintf (stderr,"ERROR: illegal voicing class \r\n");
      return FALSE;
  }

  if (1 != fprintf(This->class_fid,"%c",cVoicingClass))
  {
    fprintf (stderr,"ERROR: can't write voicing class to file \r\n");
    return FALSE;
  }


  return TRUE;

}


/*
 * Local function
 */

static BOOLEAN IsLowBandNoise(Longword *plwPEPowerSpectrum,
     BOOLEAN iVad,
     NormSw nsFrameEnergy)
 {

#define         NO_OF_DFT_POINTS        128
#define         BREAK_POINT             12

#define         LBN_HIST_WEIGHT         32440
#define         LBN_CURR_WEIGHT         (32768-LBN_HIST_WEIGHT)

#define         LBN_MAX_THR             124518
#define         LBN_LOW_ENR_LEVEL_MANT  32000
#define         LBN_LOW_ENR_LEVEL_SHFT  22

     X_INT16 i;
     static Longword lwCritMax = LBN_MAX_THR;
     Longword lwLowMax;
     Longword lwHighMax;
     Longword lwMaxRatio;
     Longword lwTemp1;
     Longword lwTemp2;

     X_INT16 iHigh;
     X_INT16 iShift;
     Shortword swLow;
     Shortword swNumerator;
     Shortword swDenominator;

     test(); logic16();
     if (iVad == TRUE)
     {
         // Speech frame - nothing changed
         test();
         return (L_sub(lwCritMax,LBN_MAX_THR) > 0); 
     }

     test();test();test();
     if ((sub(nsFrameEnergy.iShift,LBN_LOW_ENR_LEVEL_SHFT) > 0) ||
         ((sub(nsFrameEnergy.iShift,LBN_LOW_ENR_LEVEL_SHFT) == 0) &&
         (sub(nsFrameEnergy.swMantissa,LBN_LOW_ENR_LEVEL_MANT) <= 0)))
     {
         // -------------------------------------
         //      Silence - Assume neutral values
         // -------------------------------------

         lwMaxRatio = 0; move32();

     }
     else
     {
         // --------------------------------------------
         //       Background Noise - Compute the values
         // --------------------------------------------

         // Maxima Ratio
         lwLowMax = 0; move32();
         for (i = 1; i <= BREAK_POINT; i++)
         {
             test();
             if (L_sub(plwPEPowerSpectrum[i],lwLowMax) > 0)
             {
                 lwLowMax = plwPEPowerSpectrum[i]; move32();
             }
         }

         lwHighMax = 0; move32();
         for (i=BREAK_POINT+1 ; i <= NO_OF_DFT_POINTS; i++)
         {
             test();
             if (L_sub(plwPEPowerSpectrum[i],lwHighMax) > 0)
             {
                 lwHighMax = plwPEPowerSpectrum[i]; move32();
             }
         }

         test();
         if (lwHighMax <= 0)
         {
             lwMaxRatio = L_deposit_h(10);
         }
         else
         {
             test();
             if (L_sub(lwLowMax,lwHighMax) > 0)
             {
                 iShift = norm_l(lwLowMax);
             }
             else
             {
                 iShift = norm_l(lwHighMax);
             }
             swNumerator = round(L_shl(lwLowMax,iShift));
             swDenominator = round(L_shl(lwHighMax,iShift));
             if (swDenominator <= 0)
             {
                 lwMaxRatio = L_deposit_h(10);
             }
             else
             {
                 iHigh = 0; move16();
                 test();
                 while (sub(swNumerator,swDenominator) >= 0)
                 {
                     iHigh = add(iHigh,1);
                     swNumerator = sub(swNumerator,swDenominator);
                     test();
                 }
                 swLow = div_s(swNumerator,swDenominator);
                 lwMaxRatio = L_add(L_deposit_h(iHigh),L_shl(L_deposit_l(swLow),1));
             }
         }

     }

     // Update Criteria
     lwTemp1 = L_add(L_mult(extract_h(lwCritMax),LBN_HIST_WEIGHT),
         L_shr(L_mult(extract_l(L_shr(lwCritMax,1)&0x7FFF),
         LBN_HIST_WEIGHT),15)); logic32();
     lwTemp2 = L_add(L_mult(extract_h(lwMaxRatio),LBN_CURR_WEIGHT),
         L_shr(L_mult(extract_l(L_shr(lwMaxRatio,1)&0x7FFF),
         LBN_CURR_WEIGHT),15)); logic32();
     lwCritMax = L_add(lwTemp1,lwTemp2);

     // Make Decision
     test();
     return (L_sub(lwCritMax,LBN_MAX_THR) > 0);

}

#endif	/* defined(XAFE) */

